'''
File: BinnedLikelihood.py
Author: Duan Kai-Kai
Email: duankk@pmo.ac.cn
'''

from pprint import pprint
import numpy as np

from .LikelihoodBase import LikelihoodBase


class BinnedLikelihood(LikelihoodBase):
    '''
    This class is Binned Likelihood class.
    It stores the counts cube class, exposure class and/or srcmaps.
    for calculating the model counts cube and the likelihood.
    '''
    def __init__(self, CubeType, ccube, exposure, emin, emax, srcmaps=None, fitRegion=None, phi_dependence=False):
        assert isinstance(CubeType, str)
        assert isinstance(emin, (int, float))
        assert isinstance(emax, (int, float))
        assert isinstance(phi_dependence, bool)

        if not CubeType in ['Map', 'Hpx', 'MaskedHpx']:
            raise RuntimeError('The CubeType must be one of Map or Hpx, please check it.')

        if not emax > emin:
            raise RuntimeError('The emax must be larger than emin, please check it.')

        self.CubeType = CubeType
        self.ccube = ccube
        self.exposure = exposure
        emins = self.ccube.emins
        emaxs = self.ccube.emaxs
        self.emin = emins[abs(emins - emin) == min(abs(emins - emin))][0]
        self.emax = emaxs[abs(emaxs - emax) == min(abs(emaxs - emax))][0]
        self.imin = list(emins).index(self.emin)
        self.imax = list(emaxs).index(self.emax)

        if CubeType == 'Map':
            self.N = self.ccube.data[:, :, self.imin:self.imax+1]
        if CubeType in ['Hpx', 'MaskedHpx']:
            self.N = self.ccube.data[:, self.imin:self.imax+1]
        self.emins = emins[self.imin:self.imax+1]
        self.emaxs = emaxs[self.imin:self.imax+1]
        self.ebins = self.N.shape[-1]
        print('the energy range is [{}]:{}MeV to [{}]:{}MeV'.format(self.imin, self.emin, self.imax, self.emax))

        if srcmaps:
            self.srcmaps = srcmaps
        else:
            self.srcmaps = {}

        self.fitRegion = fitRegion
        self.phi_dependence = phi_dependence

        self.mdlcubes = {}

        self.modelDict = {}
        self.SrcDict = {}
        self.ParsDict = {}
        self.FixSrcDict = {}
        self.FreeSrcDict = {}
        self.FixParsDict = {}
        self.FreeParsDict = {}

    def __str__(self):
        print('Fix Source:', self.FixSrcDict.keys())
        print('Free Source:', self.FreeSrcDict.keys())
        print('Fix Parameter:', self.FixParsDict.keys())
        print('Free Parameter:', self.FreeParsDict.keys())

    def __call__(self, *args):
        kwargs = dict(list(zip(self.__code__.co_varnames, args)))
        # pprint(kwargs)
        R = self.FreeModelCube(**kwargs) + self.Fixmdlcube
        if np.sum(R) < 1e-10:
            likelihood = 0
        else:
            R[R < 1e-10] = 1e-10

        if self.fitRegion:
            if self.CubeType == 'Map':
                fitRegion = np.zeros(self.ccube.nxpix * self.ccube.nypix, dtype=bool)
            if self.CubeType in ['Hpx', 'MaskedHpx']:
                fitRegion = np.zeros(self.ccube.npix, dtype=bool)

            for region in self.fitRegion.inRegionList:
                # print(region)
                fitRegion += region.skycrd.separation(self.ccube.skycrd) < region.rad
            for region in self.fitRegion.outRegionList:
                # print(region)
                fitRegion *= region.skycrd.separation(self.ccube.skycrd) > region.rad

            if self.CubeType == 'Map':
                fitRegion = fitRegion.reshape(self.ccube.nxpix, self.ccube.nypix)
            likelihood = -2 * np.sum(self.N[fitRegion] * np.log(R[fitRegion]) - R[fitRegion])
        else:
            likelihood = -2 * np.sum(self.N * np.log(R) - R)

        #print('ccube:', self.N.sum(axis=(0, 1)), self.N.sum())
        #print('mdlcube:', R.sum(axis=(0, 1)), R.sum())
        #print('likelihood = ', likelihood)
        return likelihood
