'''
File: DmpIRFs.py
Author: Duan Kai-Kai
Email: duankk@pmo.ac.cn
'''

import os
import numpy as np
import numpy.ma as ma
from astropy.io import fits
from scipy.special import gamma
from scipy.interpolate import interp2d, interpn
from scipy.integrate import simps, quad
from scipy.optimize import brentq
import matplotlib.pyplot as plt

class AEFF:
    '''
    This class stores the effective area information,
    effective area is a function of energy and incident angle.
    '''
    def __init__(self, irf='DmpIRFs', AeffFile='DAMPE_AEFF.fits'):
        assert isinstance(irf, str)
        if not os.path.exists(os.path.join(irf, AeffFile)):
            raise FileNotFoundError('AEFF File not found, please check it.')

        hdulist = fits.open(os.path.join(irf, AeffFile))

        try:
            aeffdata = hdulist['AEFF'].data
            self.logE = aeffdata.field('LOGE')[0]          # unit:log(MeV)
            self.cosTheta = aeffdata.field('COSTHETA')[0]  # unit:cos(rad)
            self.aeff_LE = aeffdata.field('AEFF_LE')[0]    # unit:cm^2
            self.aeff_HE = aeffdata.field('AEFF_HE')[0]    # unit:cm^2

            self.aeff_LE = np.vstack([np.zeros(len(self.logE)), self.aeff_LE])
            self.aeff_HE = np.vstack([np.zeros(len(self.logE)), self.aeff_HE])
            self.cosTheta = np.hstack([0.5, self.cosTheta])

        except:
            aeffdata = hdulist['AEFF_LE'].data
            self.logE_LE = aeffdata.field('LOGE')[0]       # unit:log(MeV)
            self.cosTheta_LE = aeffdata.field('COSTHETA')[0]  # unit:cos(rad)
            self.aeff_LE = aeffdata.field('AEFF_LE')[0]    # unit:cm^2

            aeffdata = hdulist['AEFF_HE'].data
            self.logE_HE = aeffdata.field('LOGE')[0]       # unit:log(MeV)
            self.cosTheta_HE = aeffdata.field('COSTHETA')[0]  # unit:cos(rad)
            self.aeff_HE = aeffdata.field('AEFF_HE')[0]    # unit:cm^2

            self.logE_HE = np.hstack([3, self.logE_HE])
            self.cosTheta_HE = np.hstack([0.5, self.cosTheta_HE])
            self.aeff_HE = np.vstack([np.zeros(len(self.logE_HE)-1), self.aeff_HE])
            self.aeff_HE = np.hstack([np.zeros(len(self.cosTheta_HE)).reshape(-1, 1), self.aeff_HE])
        
        try:
            phi_aeffdata = hdulist['AEFF_PHI'].data
            self.phi_logE = phi_aeffdata.field('LOGE')[0]          # unit:log(MeV)
            self.phi_cosTheta = phi_aeffdata.field('COSTHETA')[0]  # unit:cos(rad)
            self.phi_phi = phi_aeffdata.field('PHI')[0]            # unit:deg
            self.phi_aeff_LE = phi_aeffdata.field('AEFF_LE')[0]    # unit:cm^2
            self.phi_aeff_HE = phi_aeffdata.field('AEFF_HE')[0]    # unit:cm^2

            self.phi_aeff_LE = np.hstack([np.zeros((len(self.phi_phi), 1, len(self.phi_logE))), self.phi_aeff_LE])
            self.phi_aeff_HE = np.hstack([np.zeros((len(self.phi_phi), 1, len(self.phi_logE))), self.phi_aeff_HE])
            self.phi_cosTheta = np.hstack([0.5, self.phi_cosTheta])

        except:
            phi_aeffdata = hdulist['AEFF_PHI_LE'].data
            self.phi_logE_LE = phi_aeffdata.field('LOGE')[0]       # unit:log(MeV)
            self.phi_cosTheta_LE = phi_aeffdata.field('COSTHETA')[0]  # unit:cos(rad)
            self.phi_phi = phi_aeffdata.field('PHI')[0]            # unit:deg
            self.phi_aeff_LE = phi_aeffdata.field('AEFF_LE')[0]    # unit:cm^2

            phi_aeffdata = hdulist['AEFF_PHI_HE'].data
            self.phi_logE_HE = phi_aeffdata.field('LOGE')[0]       # unit:log(MeV)
            self.phi_cosTheta_HE = phi_aeffdata.field('COSTHETA')[0]  # unit:cos(rad)
            self.phi_phi = phi_aeffdata.field('PHI')[0]            # unit:deg
            self.phi_aeff_HE = phi_aeffdata.field('AEFF_HE')[0]    # unit:cm^2

        try:
            phidata = hdulist['PHI_DEPENDENCE_PARAMETER'].data
            self.phi_logE = phidata.field('LOGE')[0]
            self.phi_cosTheta = phidata.field('COSTHETA')[0]
            self.p0_LE = phidata.field('P0_LE')[0][:, 0:2]
            self.p1_LE = phidata.field('p1_LE')[0][:, 0:2]
            self.p2_LE = phidata.field('p2_LE')[0][:, 0:2]
            self.p3_LE = phidata.field('p3_LE')[0][:, 0:2]
            self.p4_LE = phidata.field('p4_LE')[0][:, 0:2]
            self.p5_LE = phidata.field('p5_LE')[0][:, 0:2]
            self.p0_HE = phidata.field('p0_HE')[0]
            self.p1_HE = phidata.field('p1_HE')[0]
            self.p2_HE = phidata.field('p2_HE')[0]
            self.p3_HE = phidata.field('p3_HE')[0]
            self.p4_HE = phidata.field('p4_HE')[0]
            self.p5_HE = phidata.field('p5_HE')[0]
        except:
            pass
        hdulist.close()

    def GetAeff(self, evtype, logE, cosTheta):
        '''
        evtype: 0 for 'LE' or 1 for 'HE'
        logE: MeV, 3-5 for 'LE', 3-7 for 'HE'
        cosTheta: 0.5 - 1.0
        aeff: cm^2
        '''
        assert isinstance(evtype, int)
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(cosTheta, (int, float, list, tuple, np.ndarray))

        if not evtype in [0, 1]:
            raise RuntimeError('The evtype must be one of [0, 1], please check it.')

        logE = np.array([logE], dtype='float64').ravel()
        cosTheta = np.array([cosTheta], dtype='float64').ravel()
        if evtype == 0:
            try:
                aeff = interp2d(self.logE, self.cosTheta, self.aeff_LE, kind='cubic', fill_value=None)(logE, cosTheta)
            except:
                aeff = interp2d(self.logE_LE, self.cosTheta_LE, self.aeff_LE, kind='cubic', fill_value=None)(logE, cosTheta)
        if evtype == 1:
            try:
                aeff = interp2d(self.logE, self.cosTheta, self.aeff_HE, kind='cubic', fill_value=None)(logE, cosTheta)
            except:
                aeff = interp2d(self.logE_HE, self.cosTheta_HE, self.aeff_HE, kind='cubic', fill_value=None)(logE, cosTheta)

        aeff = aeff.reshape(cosTheta.size, logE.size)
        aeff[cosTheta < 0.5, :] = 0
        aeff[aeff < 0] = 0
        return aeff

    def GetAeff_phi(self, evtype, logE, cosTheta, phi):
        '''
        evtype: 0 for 'LE' or 1 for 'HE'
        logE: MeV, 3-5 for 'LE', 3-7 for 'HE'
        cosTheta: 0.5 - 1.0
        aeff: cm^2
        '''
        assert isinstance(evtype, int)
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(cosTheta, (int, float, list, tuple, np.ndarray))
        assert isinstance(phi, (int, float, list, tuple, np.ndarray))

        if not evtype in [0, 1]:
            raise RuntimeError('The evtype must be one of [0, 1], please check it.')

        logE = np.array([logE], dtype='float64').ravel()
        cosTheta = np.array([cosTheta], dtype='float64').ravel()
        phi = np.array([phi], dtype='float64').ravel()
        logE = logE.reshape(1, 1, logE.size)
        cosTheta = cosTheta.reshape(1, cosTheta.size, 1)
        phi = phi.reshape(phi.size, 1, 1)

        if evtype == 0:
            try:
                aeff_phi = interpn((self.phi_phi, self.phi_cosTheta, self.phi_logE), self.phi_aeff_LE, (phi, cosTheta, logE), bounds_error=False, fill_value=None)
            except:
                aeff_phi = interpn((self.phi_phi, self.phi_cosTheta_LE, self.phi_logE_LE), self.phi_aeff_LE, (phi, cosTheta, logE), bounds_error=False, fill_value=None)
        if evtype == 1:
            try:
                aeff_phi = interpn((self.phi_phi, self.phi_cosTheta, self.phi_logE), self.phi_aeff_HE, (phi, cosTheta, logE), bounds_error=False, fill_value=None)
            except:
                aeff_phi = interpn((self.phi_phi, self.phi_cosTheta_HE, self.phi_logE_HE), self.phi_aeff_HE, (phi, cosTheta, logE), bounds_error=False, fill_value=None)

        aeff_phi = aeff_phi.reshape(phi.size, cosTheta.size, logE.size)
        aeff_phi[:, cosTheta.ravel() < 0.5, :] = 0
        aeff_phi[aeff_phi < 0] = 0
        return aeff_phi


    def PlotAeff_E(self, evtype, ebins, cosTheta):
        '''
        Plot effective area over energy.
        '''
        assert isinstance(evtype, int)
        assert isinstance(ebins, int)
        assert isinstance(cosTheta, (int, float))

        if not evtype in [0, 1]:
            raise RuntimeError('The evtype must be one of [0, 1], please check it.')
        
        logEmin = np.log10(1e3) # MeV
        logEmax = np.log10(1e6) # MeV
        logE = np.linspace(logEmin, logEmax, ebins)

        aeff = self.GetAeff(evtype, logE, cosTheta)

        plt.figure()
        plt.plot(10**logE, aeff.ravel(), '-')
        plt.xscale('log')
        plt.xlim(1e3, 1e6)
        plt.ylim(0, 1500)
        plt.xlabel('Energy (MeV)')
        plt.ylabel(r'Effective area ($cm^{2}$)')
        plt.title(r'Aeff_logE_{}_{}'.format(cosTheta, evtype))
        plt.grid()
        plt.savefig('Aeff_E_{}_{}.png'.format(cosTheta, evtype))
        return logE, aeff.ravel()

    def PlotAeff_Theta(self, evtype, logE, Thetabins):
        '''
        Plot effective area over Theta.
        '''
        assert isinstance(evtype, int)
        assert isinstance(Thetabins, int)
        assert isinstance(logE, (int, float))

        if not evtype in [0, 1]:
            raise RuntimeError('The evtype must be one of [0, 1], please check it.')
        
        cosTheta = np.linspace(0.5, 1, Thetabins)
        aeff = self.GetAeff(evtype, logE, cosTheta)

        plt.figure()
        plt.plot(cosTheta, aeff.ravel(), '-')
        plt.xlim(0.5, 1)
        plt.ylim(0, 1500)
        plt.xlabel(r'$cos\theta$')
        plt.ylabel(r'Effective area ($cm^{2}$)')
        plt.title(r'Aeff_cosTheta_{}_{}'.format(logE, evtype))
        plt.grid()
        plt.savefig('Aeff_Theta_{}_{}.png'.format(logE, evtype))
        return cosTheta, aeff.ravel()

    def PlotAeff_E_phi(self, evtype, ebins, cosTheta):
        assert isinstance(evtype, int)
        assert isinstance(ebins, int)
        assert isinstance(cosTheta, (int, float))

        if not evtype in [0, 1]:
            raise RuntimeError('The evtype must be one of [0, 1], please check it.')
        
        logEmin = np.log10(1e3) # MeV
        logEmax = np.log10(1e6) # MeV
        logE = np.linspace(logEmin, logEmax, ebins)

        phi = np.linspace(0, 360, len(self.phi_phi))
        aeff_phi = self.GetAeff_phi(evtype, logE, cosTheta, phi)
        aeff = aeff_phi.sum(axis=0)
        print(aeff)

        plt.figure()
        plt.plot(10**logE, aeff.ravel(), '-')
        plt.xscale('log')
        plt.xlim(1e3, 1e6)
        plt.ylim(0, 1500)
        plt.xlabel('Energy (MeV)')
        plt.ylabel(r'Effective area ($cm^{2}$)')
        plt.title(r'Aeff_logE_{}_{}'.format(cosTheta, evtype))
        plt.grid()
        plt.savefig('Aeff_E_{}_{}.png'.format(cosTheta, evtype))
        return logE, aeff.ravel()


    def PlotAeff_Theta_phi(self, evtype, logE, Thetabins):
        assert isinstance(evtype, int)
        assert isinstance(logE, (int, float))
        assert isinstance(Thetabins, int)

        if not evtype in [0, 1]:
            raise RuntimeError('The evtype must be one of [0, 1], please check it.')

        cosTheta = np.linspace(0.5, 1, Thetabins)
        aeff = self.GetAeff(evtype, logE, cosTheta)

        phi = np.linspace(0, 360, len(self.phi_phi))
        aeff_phi = self.GetAeff_phi(evtype, logE, cosTheta, phi)
        aeff = aeff_phi.sum(axis=0)

        plt.figure()
        plt.plot(cosTheta, aeff.ravel(), '-')
        plt.xlim(0.5, 1)
        plt.ylim(0, 1500)
        plt.xlabel(r'$cos\theta$')
        plt.ylabel(r'Effective area ($cm^{2}$)')
        plt.title(r'Aeff_cosTheta_{}_{}'.format(logE, evtype))
        plt.grid()
        plt.savefig('Aeff_Theta_{}_{}.png'.format(logE, evtype))
        return cosTheta, aeff.ravel()

    def PlotAeff_Phi_phi(self, evtype, logE, cosTheta, Phibins):
        '''
        Plot effective area over Phi.
        '''
        assert isinstance(evtype, int)
        assert isinstance(logE, (int, float))
        assert isinstance(cosTheta, (int, float))
        assert isinstance(Phibins, int)

        if not evtype in [0, 1]:
            raise RuntimeError('The evtype must be one of [0, 1], please check it.')
        
        phi = np.linspace(0, 360, Phibins)
        #aeff = self.GetAeff(evtype, logE, cosTheta)
        aeff_phi = self.GetAeff_phi(evtype, logE, cosTheta, phi)

        plt.figure()
        plt.plot(phi, aeff_phi.ravel(), '-')
        plt.xlim(0, 360)
        plt.ylim(0, 1500)
        plt.xlabel(r'$\phi$')
        plt.ylabel(r'Effective area ($cm^{2}$)')
        plt.title(r'Aeff_phi_{}_{}_{}'.format(logE, cosTheta, evtype))
        plt.grid()
        plt.savefig('Aeff_phi_{}_{}_{}.png'.format(logE, cosTheta, evtype))
        return phi, aeff_phi.ravel()


class PSF:
    '''
    This class stores the information of point-spread function (PSF),
    PSF is a function of energy, incident angle and dTheta.
    '''
    def __init__(self, irf='DmpIRFs', PsfFile='DAMPE_PSF.fits'):
        assert isinstance(irf, str)
        if not os.path.exists(os.path.join(irf, PsfFile)):
            raise FileNotFoundError('PSF PARS File not Found, please check it.')

        hdulist = fits.open(os.path.join(irf, PsfFile))
        Psfdata_LE = hdulist['PSF_LE'].data

        self.logE_LE = Psfdata_LE.field('LOGE')[0]
        self.cosTheta_LE = Psfdata_LE.field('COSTHETA')[0]

        self.Sp_LE = Psfdata_LE.field('SP')[0]
        self.fcore_LE = Psfdata_LE.field('FCORE')[0]

        self.score_LE = Psfdata_LE.field('SCORE')[0]
        self.stail_LE = Psfdata_LE.field('STAIL')[0]

        self.gcore_LE = Psfdata_LE.field('GCORE')[0]
        self.gtail_LE = Psfdata_LE.field('GTAIL')[0]

        Psfdata_HE = hdulist['PSF_HE'].data

        self.logE_HE = Psfdata_HE.field('LOGE')[0]
        self.cosTheta_HE = Psfdata_HE.field('COSTHETA')[0]

        self.Sp_HE = Psfdata_HE.field('SP')[0]
        self.fcore_HE = Psfdata_HE.field('FCORE')[0]

        self.score_HE = Psfdata_HE.field('SCORE')[0]
        self.stail_HE = Psfdata_HE.field('STAIL')[0]

        self.gcore_HE = Psfdata_HE.field('GCORE')[0]
        self.gtail_HE = Psfdata_HE.field('GTAIL')[0]

        try:
            scaledata = hdulist['SCALE_PARS'].data
            self.c0_LE = scaledata.field('LE')[0][0]
            self.c1_LE = scaledata.field('LE')[0][1]
            self.beta_LE = scaledata.field('LE')[0][2]

            self.c0_HE = scaledata.field('HE')[0][0]
            self.c1_HE = scaledata.field('HE')[0][1]
            self.beta_HE = scaledata.field('HE')[0][2]
        except:
            pass
        hdulist.close()

    @staticmethod
    def __Sp(e, c0, c1, beta):
        return np.sqrt((c0 * (e/1e3)**beta)**2 + c1**2)

    @staticmethod
    def __king(x, s, g):
        return (1. / (2. * np.pi * s**2)) * (1. - 1. / g)\
            * (1. + (x**2) / (2. * g * s**2))**(-g)


    def GetPsf_dTheta(self, evtype, logE, cosTheta, dTheta):
        '''
        evtype: 0 for 'LE' or 1 for 'HE'
        logE: MeV, 3-5 for 'LE', 3-7 for 'HE'
        cosTheta: 0.5 - 1.0
        dTheta: deg
        '''
        assert isinstance(evtype, int)
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(cosTheta, (int, float, list, tuple, np.ndarray))
        assert isinstance(dTheta, (int, float, list, tuple, np.ndarray))

        if not evtype in [0, 1]:
            raise RuntimeError('The evtype must be one of [0, 1], please check it.')

        logE = np.array([logE], dtype='float64').ravel()
        cosTheta = np.array([cosTheta], dtype='float64').ravel()
        if not isinstance(dTheta, np.ndarray):
            dTheta = np.array([dTheta], dtype='float64')

        if evtype == 0:
            logE[logE < min(self.logE_LE)] = min(self.logE_LE)
            logE[logE > max(self.logE_LE)] = max(self.logE_LE)
            cosTheta[cosTheta < min(self.cosTheta_LE)] = min(self.cosTheta_LE)
            cosTheta[cosTheta > max(self.cosTheta_LE)] = max(self.cosTheta_LE)
            Sp = interp2d(self.logE_LE, self.cosTheta_LE, self.Sp_LE)(logE, cosTheta)
            fcore = interp2d(self.logE_LE, self.cosTheta_LE, self.fcore_LE)(logE, cosTheta)
            score = interp2d(self.logE_LE, self.cosTheta_LE, self.score_LE)(logE, cosTheta)
            stail = interp2d(self.logE_LE, self.cosTheta_LE, self.stail_LE)(logE, cosTheta)
            gcore = interp2d(self.logE_LE, self.cosTheta_LE, self.gcore_LE)(logE, cosTheta)
            gtail = interp2d(self.logE_LE, self.cosTheta_LE, self.gtail_LE)(logE, cosTheta)

        if evtype == 1:
            logE[logE < min(self.logE_HE)] = min(self.logE_HE)
            logE[logE > max(self.logE_HE)] = max(self.logE_HE)
            cosTheta[cosTheta < min(self.cosTheta_HE)] = min(self.cosTheta_HE)
            cosTheta[cosTheta > max(self.cosTheta_HE)] = max(self.cosTheta_HE)
            Sp = interp2d(self.logE_HE, self.cosTheta_HE, self.Sp_HE)(logE, cosTheta)
            fcore = interp2d(self.logE_HE, self.cosTheta_HE, self.fcore_HE)(logE, cosTheta)
            score = interp2d(self.logE_HE, self.cosTheta_HE, self.score_HE)(logE, cosTheta)
            stail = interp2d(self.logE_HE, self.cosTheta_HE, self.stail_HE)(logE, cosTheta)
            gcore = interp2d(self.logE_HE, self.cosTheta_HE, self.gcore_HE)(logE, cosTheta)
            gtail = interp2d(self.logE_HE, self.cosTheta_HE, self.gtail_HE)(logE, cosTheta)

        Sp = Sp.reshape(cosTheta.size, logE.size)
        if dTheta.ndim == 1:
            x = dTheta[:, np.newaxis, np.newaxis] / Sp[np.newaxis, :, :]
        if dTheta.ndim == 2:
            x = dTheta[:, :, np.newaxis, np.newaxis] / Sp[np.newaxis, np.newaxis, :, :]
        P = fcore * self.__king(x, score, gcore)\
            + (1 - fcore) * self.__king(x, stail, gtail)
        return Sp, x, P

    def GetPsf_dTheta2(self, evtype, logE, cosTheta, dTheta):
        '''
        evtype: 0 for 'LE' or 1 for'HE'
        logE: MeV, 3-5 for 'LE', 3-7 for 'HE'
        cosTheta: 0.5 - 1.0
        dTheta: deg
        '''
        assert isinstance(evtype, int)
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(cosTheta, (int, float, list, tuple, np.ndarray))
        assert isinstance(dTheta, (int, float, list, tuple, np.ndarray))

        if not evtype in [0, 1]:
            raise RuntimeError('The evtype must be one of [0, 1], please check it.')

        logE0 = np.array([logE], dtype='float64').ravel()
        cosTheta0 = np.array([cosTheta], dtype='float64').ravel()
        dTheta = np.array([dTheta], dtype='float64').ravel()

        P = np.zeros((dTheta.size, cosTheta0.size, logE0.size))
        if evtype == 0:
            logE0[logE0 < min(self.logE_LE)] = min(self.logE_LE)
            logE0[logE0 > max(self.logE_LE)] = max(self.logE_LE)
            cosTheta0[cosTheta0 < min(self.cosTheta_LE)] = min(self.cosTheta_LE)
            cosTheta0[cosTheta0 > max(self.cosTheta_LE)] = max(self.cosTheta_LE)
            Sp = interp2d(self.logE_LE, self.cosTheta_LE, self.Sp_LE)(logE0, cosTheta0)
            xs = dTheta.reshape(dTheta.size, 1, 1) / self.Sp_LE
            profile = self.fcore_LE * self.__king(xs, self.score_LE, self.gcore_LE) \
              + (1 - self.fcore_LE) * self.__king(xs, self.stail_LE, self.gtail_LE)
            for i in range(len(dTheta)):
                P[i, :, :] = interp2d(self.logE_LE, self.cosTheta_LE, profile[i, :, :])(logE0, cosTheta0)

        if evtype == 1:
            logE0[logE0 < min(self.logE_HE)] = min(self.logE_HE)
            logE0[logE0 > max(self.logE_HE)] = max(self.logE_HE)
            cosTheta0[cosTheta0 < min(self.cosTheta_HE)] = min(self.cosTheta_HE)
            cosTheta0[cosTheta0 > max(self.cosTheta_HE)] = max(self.cosTheta_HE)
            Sp = interp2d(self.logE_HE, self.cosTheta_HE, self.Sp_HE)(logE0, cosTheta0)
            xs = dTheta.reshape(dTheta.size, 1, 1) / self.Sp_HE
            profile = self.fcore_HE * self.__king(xs, self.score_HE, self.gcore_HE) \
              + (1 - self.fcore_HE) * self.__king(xs, self.stail_HE, self.gtail_HE)
            for i in range(len(dTheta)):
                P[i, :, :] = interp2d(self.logE_HE, self.cosTheta_HE, profile[i, :, :])(logE0, cosTheta0)

        x = dTheta.reshape(dTheta.size, 1, 1) / Sp
        return Sp, x, P

    def PlotPsf_dTheta(self, evtype, logE, cosTheta, dThetamax=10, dThetabins=100):
        '''
        Plot PSF over dTheta.
        '''
        assert isinstance(evtype, int)
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(cosTheta, (int, float, list, tuple, np.ndarray))
        assert isinstance(dThetamax, (int, float))
        assert isinstance(dThetabins, int)

        if not evtype in [0, 1, 2]:
            raise RuntimeError('The evtype must be one of [0, 1, 2], please check it.')

        dTheta = np.logspace(-3, np.log10(dThetamax), dThetabins)
        Sp, x, P = self.GetPsf_dTheta(evtype, logE, cosTheta, dTheta)
        print(2 * np.pi * simps(P.ravel() * x.ravel(), x.ravel(), axis=0))
        print(2 * np.pi * simps(P.ravel() / Sp.ravel()**2 * dTheta.ravel(), dTheta.ravel(), axis=0))

        R68 = self.GetPsf(evtype, logE, cosTheta, ratiolimit=0.68)
        R95 = self.GetPsf(evtype, logE, cosTheta, ratiolimit=0.95)

        plt.figure()
        plt.plot(dTheta, P.ravel(), 'r')
        plt.axvline(R68, c='g', label='R68')
        plt.axvline(R95, c='b', label='R95')
        plt.xlim(1e-3, 10)
        plt.ylim(1e-5, 10)
        plt.xscale('log')
        plt.yscale('log')
        plt.xlabel('dTheta')
        plt.ylabel('Psf')
        plt.title('PSF_dTheta_{}_{}_{}'.format(logE, cosTheta, evtype))
        plt.legend()
        plt.grid()
        plt.savefig('Psf_dTheta_{}_{}_{}.png'.format(logE, cosTheta, evtype))
        return dTheta, P.ravel()

    def cdf(self, evtype, logE, cosTheta, dThetamax):
        '''
        Calculate the integral of PSF from 0 to dThetamax.
        '''
        assert isinstance(evtype, int)
        assert isinstance(logE, (int, float))
        assert isinstance(cosTheta, (int, float))
        assert isinstance(dThetamax, (int, float))

        if not evtype in [0, 1, 2]:
            raise RuntimeError('The evtype must be one of [0, 1, 2], please check it.')

        dTheta = np.logspace(-5, np.log10(dThetamax), 100)
        Sp, x, P = self.GetPsf_dTheta(evtype, logE, cosTheta, dTheta)
        cdf = 2 * np.pi * simps(P * x, x, axis=0)
        return cdf

    def GetPsf(self, evtype, logE, cosTheta, ratiolimit=0.68):
        '''
        Calculate the PSF at logE, cosTheta, with ratiolimit.
        '''
        assert isinstance(evtype, int)
        assert isinstance(logE, (int, float))
        assert isinstance(cosTheta, (int, float))
        assert isinstance(ratiolimit, float)

        if not evtype in [0, 1, 2]:
            raise RuntimeError('The evtype must be one of [0, 1, 2], please check it.')
        
        Psf = brentq(lambda dThetamax: self.cdf(evtype, logE, cosTheta, dThetamax) - ratiolimit, 1e-3, 10)
        return Psf

    def PlotPsf_E(self, evtype, ebins, cosTheta0, ratiolimit=0.68):
        '''
        Plot PSF over energy at cosTheta with ratiolimit.
        '''
        assert isinstance(evtype, int)
        assert isinstance(ebins, int)
        assert isinstance(cosTheta0, (int, float))
        assert isinstance(ratiolimit, float)

        if not evtype in [0, 1, 2]:
            raise RuntimeError('The evtype must be one of [0, 1, 2], please check it.')
        
        logE = np.linspace(3, 6, ebins)
        Psf = np.zeros(ebins)
        for i, logE0 in enumerate(logE):
            try:
                Psf[i] = self.GetPsf(evtype, logE0, cosTheta0, ratiolimit)
            except:
                Psf[i] = 0

        plt.figure()
        plt.plot(10**logE, Psf, 'o-')
        if evtype == 0:
            plt.xlim(1e3, 1e5)
        if evtype == 1:
            plt.xlim(1e3, 1e6)
        plt.ylim(1e-2, 1e1)
        plt.xscale('log')
        plt.yscale('log')
        plt.xlabel('Energy (MeV)')
        plt.ylabel('Containment angle (deg)')
        plt.title('PSF_logE_{}_{}'.format(cosTheta0, evtype))
        plt.grid()
        plt.savefig('Psf_E_{}_{}.png'.format(cosTheta0, evtype))
        return logE, Psf

    def PlotPsf_Theta(self, evtype, logE0, thetabins, ratiolimit=0.68):
        '''
        Plot PSF over theta at logE with ratiolimit.
        '''
        assert isinstance(evtype, int)
        assert isinstance(logE0, (int, float))
        assert isinstance(thetabins, int)
        assert isinstance(ratiolimit, float)

        if not evtype in [0, 1, 2]:
            raise RuntimeError('The evtype must be one of [0, 1, 2], please check it.')
        
        Theta = np.linspace(0, 60, thetabins)
        cosTheta = np.cos(Theta / 180. * np.pi)
        Psf = np.zeros(thetabins)
        for j, cosTheta0 in enumerate(cosTheta):
            try:
                Psf[j] = self.GetPsf(evtype, logE0, cosTheta0, ratiolimit)
            except ValueError:
                Psf[j] = 0

        plt.figure()
        plt.plot(cosTheta, Psf, 'o-')
        plt.xlim(0.5, 1)
        plt.ylim(1e-2, 1e1)
        plt.yscale('log')
        plt.xlabel(r'cos$\theta$')
        plt.ylabel(r'Containment angle (deg)')
        plt.title('PSF_cosTheta_{}_{}'.format(logE0, evtype))
        plt.grid()
        plt.savefig('Psf_Theta_{}_{}.png'.format(logE0, evtype))
        return cosTheta, Psf


class EDISP:
    '''
    This class stores the information of energy dispersion (Edisp),
    Edisp is function of energy, incident angle and dE/energy.
    '''
    def __init__(self, irf='DmpIRFs', EdispFile='DAMPE_EDISP.fits'):
        assert isinstance(irf, str)
        if not os.path.exists(os.path.join(irf, EdispFile)):
            raise FileNotFoundError('EDISP PARS File not Found, please check it.')

        hdulist = fits.open(os.path.join(irf, EdispFile))

        Edispdata_LE = hdulist['EDISP_LE'].data

        self.logE_LE = Edispdata_LE.field('LOGE')[0]
        self.cosTheta_LE = Edispdata_LE.field('COSTHETA')[0]

        self.SD_LE = Edispdata_LE.field('SD')[0]
        self.norm_LE = Edispdata_LE.field('NORM')[0]
        self.x0_LE = Edispdata_LE.field('X0')[0]
        self.xbar_LE = Edispdata_LE.field('XBAR')[0]
        self.sigmaR_LE = Edispdata_LE.field('SIGMAR')[0]
        self.sigmaL1_LE = Edispdata_LE.field('SIGMAL1')[0]
        self.sigmaL2_LE = Edispdata_LE.field('SIGMAL2')[0]
        self.gammaR_LE = Edispdata_LE.field('GAMMAR')[0]
        self.gammaL1_LE = Edispdata_LE.field('GAMMAL1')[0]
        self.gammaL2_LE = Edispdata_LE.field('GAMMAL2')[0]

        Edispdata_HE = hdulist['EDISP_HE'].data

        self.logE_HE = Edispdata_HE.field('LOGE')[0]
        self.cosTheta_HE = Edispdata_HE.field('COSTHETA')[0]

        self.SD_HE = Edispdata_HE.field('SD')[0]
        self.norm_HE = Edispdata_HE.field('NORM')[0]
        self.x0_HE = Edispdata_HE.field('X0')[0]
        self.xbar_HE = Edispdata_HE.field('XBAR')[0]
        self.sigmaR_HE = Edispdata_HE.field('SIGMAR')[0]
        self.sigmaL1_HE = Edispdata_HE.field('SIGMAL1')[0]
        self.sigmaL2_HE = Edispdata_HE.field('SIGMAL2')[0]
        self.gammaR_HE = Edispdata_HE.field('GAMMAR')[0]
        self.gammaL1_HE = Edispdata_HE.field('GAMMAL1')[0]
        self.gammaL2_HE = Edispdata_HE.field('GAMMAL2')[0]
        hdulist.close()

    def __R(self, x, x0, sigma, gamma):
        R = np.exp(-0.5 * abs((x - x0) / sigma)**gamma)
        return R

    def __D(self, x, pars):
        N, x0, xbar, sigma_r, sigma_l, sigma_L, gamma_r, gamma_l, gamma_L = pars
        D = np.zeros_like(x)

        mx = ma.masked_array(x, mask=~((x-x0)>0))
        mD = N * self.__R(mx, x0, sigma_r, gamma_r)
        D += mD.filled(0)

        mx = ma.masked_array(x, mask=~(((x - x0) >= -xbar) * ((x - x0) <= 0)))
        mD = N * self.__R(mx, x0, sigma_l, gamma_l)
        D += mD.filled(0)

        Nl = self.__R(x0 - xbar, x0, sigma_l, gamma_l)
        NL = self.__R(x0 - xbar, x0, sigma_L, gamma_L)

        mx = ma.masked_array(x, mask=~((x - x0) < -xbar))
        '''
        if NL > 1e-99:
            mD = N * Nl / NL * self.__R(mx, x0, sigma_L, gamma_L)
        else:
            mD = N * self.__R(mx, x0, sigma_l, gamma_l)
        '''
        condition = NL > 1e-99
        mD[:, condition] = N[condition] * Nl[condition] / NL[condition] * self.__R(mx[:, condition], x0[condition], sigma_L[condition], gamma_L[condition])
        condition = NL < 1e-99
        mD[:, condition] = N[condition] * self.__R(mx[:, condition], x0[condition], sigma_l[condition], gamma_l[condition])
        D += mD.filled(0)
        return D


    def GetEdisp_dEratio1(self, evtype, logE, cosTheta, dEratio):
        '''
        evtype: 0 for 'LE' or 1 for 'HE'
        logE: MeV, 3-5 for 'LE', 3-7 for 'HE'
        cosTheta: 0.5 - 1.0
        dEratio: (E - E0) / E0
        '''
        assert isinstance(evtype, int)
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(cosTheta, (int, float, list, tuple, np.ndarray))
        assert isinstance(dEratio, (int, float, list, tuple, np.ndarray))

        if not evtype in [0, 1]:
            raise RuntimeError('The evtype must be one of [0, 1], please check it.')

        logE0 = np.array([logE], dtype='float64').ravel()
        cosTheta0 = np.array([cosTheta], dtype='float64').ravel()
        dEratio = np.array([dEratio], dtype='float64').ravel()

        if evtype == 0:
            logE0[logE0 < min(self.logE_LE)] = min(self.logE_LE)
            logE0[logE0 > max(self.logE_LE)] = max(self.logE_LE)
            cosTheta0[cosTheta0 < min(self.cosTheta_LE)] = min(self.cosTheta_LE)
            cosTheta0[cosTheta0 > max(self.cosTheta_LE)] = max(self.cosTheta_LE)
            SD = interp2d(self.logE_LE, self.cosTheta_LE, self.SD_LE)(logE0, cosTheta0)
            norm = interp2d(self.logE_LE, self.cosTheta_LE, self.norm_LE)(logE0, cosTheta0)
            x0 = interp2d(self.logE_LE, self.cosTheta_LE, self.x0_LE)(logE0, cosTheta0)
            xbar = interp2d(self.logE_LE, self.cosTheta_LE, self.xbar_LE)(logE0, cosTheta0)
            sigmaR = interp2d(self.logE_LE, self.cosTheta_LE, self.sigmaR_LE)(logE0, cosTheta0)
            sigmaL1 = interp2d(self.logE_LE, self.cosTheta_LE, self.sigmaL1_LE)(logE0, cosTheta0)
            sigmaL2 = interp2d(self.logE_LE, self.cosTheta_LE, self.sigmaL2_LE)(logE0, cosTheta0)
            gammaR = interp2d(self.logE_LE, self.cosTheta_LE, self.gammaR_LE)(logE0, cosTheta0)
            gammaL1 = interp2d(self.logE_LE, self.cosTheta_LE, self.gammaL1_LE)(logE0, cosTheta0)
            gammaL2 = interp2d(self.logE_LE, self.cosTheta_LE, self.gammaL2_LE)(logE0, cosTheta0)
        if evtype == 1:
            logE0[logE0 < min(self.logE_HE)] = min(self.logE_HE)
            logE0[logE0 > max(self.logE_HE)] = max(self.logE_HE)
            cosTheta0[cosTheta0 < min(self.cosTheta_LE)] = min(self.cosTheta_LE)
            cosTheta0[cosTheta0 > max(self.cosTheta_HE)] = max(self.cosTheta_HE)
            SD = interp2d(self.logE_HE, self.cosTheta_HE, self.SD_HE)(logE0, cosTheta0)
            norm = interp2d(self.logE_HE, self.cosTheta_HE, self.norm_HE)(logE0, cosTheta0)
            x0 = interp2d(self.logE_HE, self.cosTheta_HE, self.x0_HE)(logE0, cosTheta0)
            xbar = interp2d(self.logE_HE, self.cosTheta_HE, self.xbar_HE)(logE0, cosTheta0)
            sigmaR = interp2d(self.logE_HE, self.cosTheta_HE, self.sigmaR_HE)(logE0, cosTheta0)
            sigmaL1 = interp2d(self.logE_HE, self.cosTheta_HE, self.sigmaL1_HE)(logE0, cosTheta0)
            sigmaL2 = interp2d(self.logE_HE, self.cosTheta_HE, self.sigmaL2_HE)(logE0, cosTheta0)
            gammaR = interp2d(self.logE_HE, self.cosTheta_HE, self.gammaR_HE)(logE0, cosTheta0)
            gammaL1 = interp2d(self.logE_HE, self.cosTheta_HE, self.gammaL1_HE)(logE0, cosTheta0)
            gammaL2 = interp2d(self.logE_HE, self.cosTheta_HE, self.gammaL2_HE)(logE0, cosTheta0)

        x = dEratio[:, np.newaxis, np.newaxis] / SD
        D = self.__D(x, (norm, x0, xbar, sigmaR, sigmaL1, sigmaL2, gammaR, gammaL1, gammaL2))

        return SD, x, D

    def GetEdisp_dEratio(self, evtype, logE, cosTheta, dEratio):
        '''
        evtype: 0 for 'LE' or 1 for 'HE'
        logE: MeV, 3-5 for 'LE', 3-7 for 'HE'
        cosTheta: 0.5 - 1.0
        dEratio: (E - E0) / E0
        '''
        assert isinstance(evtype, int)
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(cosTheta, (int, float, list, tuple, np.ndarray))
        assert isinstance(dEratio, (int, float, list, tuple, np.ndarray))

        if not evtype in [0, 1]:
            raise RuntimeError('The evtype must be one of [0, 1], please check it.')

        logE0 = np.array([logE], dtype='float64').ravel()
        cosTheta0 = np.array([cosTheta], dtype='float64').ravel()
        dEratio = np.array([dEratio], dtype='float64').ravel()

        D = np.zeros((dEratio.size, cosTheta0.size, logE0.size))
        if evtype == 0:
            logE0[logE0 < min(self.logE_LE)] = min(self.logE_LE)
            logE0[logE0 > max(self.logE_LE)] = max(self.logE_LE)
            cosTheta0[cosTheta0 < min(self.cosTheta_LE)] = min(self.cosTheta_LE)
            cosTheta0[cosTheta0 > max(self.cosTheta_LE)] = max(self.cosTheta_LE)
            SD = interp2d(self.logE_LE, self.cosTheta_LE, self.SD_LE)(logE0, cosTheta0)
            xs = dEratio.reshape(dEratio.size, 1, 1) / self.SD_LE
            profile = self.__D(xs, (self.norm_LE, self.x0_LE, self.xbar_LE, self.sigmaR_LE, self.sigmaL1_LE, self.sigmaL2_LE, self.gammaR_LE, self.gammaL1_LE, self.gammaL2_LE))
            for i in range(len(dEratio)):
                D[i, :, :] = interp2d(self.logE_LE, self.cosTheta_LE, profile[i, :, :])(logE0, cosTheta0)

        if evtype == 1:
            logE0[logE0 < min(self.logE_HE)] = min(self.logE_HE)
            logE0[logE0 > max(self.logE_HE)] = max(self.logE_HE)
            cosTheta0[cosTheta0 < min(self.cosTheta_HE)] = min(self.cosTheta_HE)
            cosTheta0[cosTheta0 > max(self.cosTheta_HE)] = max(self.cosTheta_HE)
            SD = interp2d(self.logE_HE, self.cosTheta_HE, self.SD_HE)(logE0, cosTheta0)
            xs = dEratio.reshape(dEratio.size, 1, 1) / self.SD_HE
            profile = self.__D(xs, (self.norm_HE, self.x0_HE, self.xbar_HE, self.sigmaR_HE, self.sigmaL1_HE, self.sigmaL2_HE, self.gammaR_HE, self.gammaL1_HE, self.gammaL2_HE))
            for i in range(len(dEratio)):
                D[i, :, :] = interp2d(self.logE_HE, self.cosTheta_HE, profile[i, :, :])(logE0, cosTheta0)

        x = dEratio.reshape(dEratio.size, 1, 1) / SD
        return SD, x, D

    def PlotEdisp_dEratio(self, evtype, logE, cosTheta, dEratiomax=1, dEratiobins=1000):
        '''
        Plot EDISP over dEratio.
        '''
        assert isinstance(evtype, int)
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(cosTheta, (int, float, list, tuple, np.ndarray))
        assert isinstance(dEratiomax, (int, float))
        assert isinstance(dEratiobins, int)

        if not evtype in [0, 1]:
            raise RuntimeError('The evtype must be one of [0, 1], please check it.')
        
        dEratio = np.linspace(-1, dEratiomax, dEratiobins)
        E = (dEratio + 1) * (10 ** logE)

        SD, x, D = self.GetEdisp_dEratio(evtype, logE, cosTheta, dEratio)
        print(simps(D.ravel(), x.ravel(), axis=0))
        print(simps(D.ravel() / SD, dEratio, axis=0))

        x0 = self.get_highest_point(evtype, logE, cosTheta)
        R68, R68_left, R68_right = self.GetEdisp(evtype, logE, cosTheta, ratiolimit=0.68)
        R95, R95_left, R95_right = self.GetEdisp(evtype, logE, cosTheta, ratiolimit=0.95)

        plt.figure()
        plt.plot(x.ravel(), D.ravel())
        # plt.axvline(x0/SD, c='r')
        plt.axvline(R68_left/SD, c='g')
        plt.axvline(R68_right/SD, c='g', label='R68')
        plt.axvline(R95_left/SD, c='b')
        plt.axvline(R95_right/SD, c='b', label='R95')
        plt.xlim(-15, 5)
        plt.ylim(1e-5, 1e0)
        #plt.xscale('log')
        plt.yscale('log')
        plt.xlabel('scaled dEratio')
        plt.ylabel('Edisp')
        plt.title('EDISP_dEratio_{}_{}_{}'.format(logE, cosTheta, evtype))
        plt.grid()
        plt.legend()
        plt.savefig('Edisp_dEratio_{}_{}_{}.png'.format(logE, cosTheta, evtype))
        return dEratio, D.ravel()

    def get_highest_point(self, evtype, logE, cosTheta, dEratiomax=1, dEratiobins=100):
        dEratio = np.linspace(-1, dEratiomax, dEratiobins)
        SD, x, D = self.GetEdisp_dEratio(evtype, logE, cosTheta, dEratio)
        D = D.ravel()
        return float(dEratio[D == max(D)])

    def cdf(self, evtype, logE, cosTheta, dEratiomax):
        '''
        Calculate the integral at logE, cosTheta from 0 to dEratiomax.
        '''
        assert isinstance(evtype, int)
        assert isinstance(logE, (int, float))
        assert isinstance(cosTheta, (int, float))
        assert isinstance(dEratiomax, (int, float))

        if not evtype in [0, 1, 2]:
            raise RuntimeError('The evtype must be one of [0, 1, 2], please check it.')
        
        dEratio = np.linspace(-1, dEratiomax, 100)
        SD, x, D = self.GetEdisp_dEratio(evtype, logE, cosTheta, dEratio)
        cdf = simps(D.ravel(), x.ravel(), axis=0)
        return cdf

    def GetEdisp1(self, evtype, logE, cosTheta, ratiolimit=0.68):
        '''
        Calculate the EDISP at logE, cosTheta with ratiolimit.
        '''
        assert isinstance(evtype, int)
        assert isinstance(logE, (int, float))
        assert isinstance(cosTheta, (int, float))
        assert isinstance(ratiolimit, float)

        if not evtype in [0, 1, 2]:
            raise RuntimeError('The evtype must be one of [0, 1, 2], please check it.')

        highest_point = self.get_highest_point(evtype, logE, cosTheta)
        highest_cdf = self.cdf(evtype, logE, cosTheta, highest_point)

        Edisp_left = brentq(lambda x: self.cdf(evtype, logE, cosTheta, highest_point) - self.cdf(evtype, logE, cosTheta, x) - ratiolimit / 2, -1+1e-2, highest_point)
        left_cdf = self.cdf(evtype, logE, cosTheta, highest_point) - self.cdf(evtype, logE, cosTheta, Edisp_left)

        Edisp_right = brentq(lambda x: self.cdf(evtype, logE, cosTheta, x) - self.cdf(evtype, logE, cosTheta, highest_point) - ratiolimit / 2, highest_point, 1)
        right_cdf = self.cdf(evtype, logE, cosTheta, Edisp_right) - self.cdf(evtype, logE, cosTheta, highest_point)

        Edisp = (Edisp_right - Edisp_left) / 2
        return Edisp, Edisp_left, Edisp_right

    def GetEdisp(self, evtype, logE, cosTheta, ratiolimit=0.68):
        '''
        Calculate the EDISP at logE, cosTheta with ratiolimit.
        '''
        assert isinstance(evtype, int)
        assert isinstance(logE, (int, float))
        assert isinstance(cosTheta, (int, float))
        assert isinstance(ratiolimit, float)

        if not evtype in [0, 1, 2]:
            raise RuntimeError('The evtype must be one of [0, 1, 2], please check it.')

        cdf = 0
        ymin = 0
        highest_point = self.get_highest_point(evtype, logE, cosTheta)

        ymax = float(self.GetEdisp_dEratio(evtype, logE, cosTheta, highest_point)[2])
        while abs(cdf - ratiolimit) > 1e-3:
            y = (ymin + ymax) / 2
            x1 = brentq(lambda x: float(abs(self.GetEdisp_dEratio(evtype, logE, cosTheta, x)[2]) - y), -1, highest_point)
            x2 = brentq(lambda x: float(abs(self.GetEdisp_dEratio(evtype, logE, cosTheta, x)[2]) - y), highest_point, 1)
            cdf = self.cdf(evtype, logE, cosTheta, x2) - self.cdf(evtype, logE, cosTheta, x1)
            if cdf > ratiolimit:
                ymin = y
            if cdf < ratiolimit:
                ymax = y
        Edisp = (x2 - x1) / 2
        return Edisp, x1, x2

    def PlotEdisp_E(self, evtype, ebins, cosTheta0, ratiolimit=0.68):
        '''
        Plot EDISP over energy at cosTheta with ratiolimit.
        '''
        assert isinstance(evtype, int)
        assert isinstance(ebins, int)
        assert isinstance(cosTheta0, (int, float))
        assert isinstance(ratiolimit, float)

        if not evtype in [0, 1, 2]:
            raise RuntimeError('The evtype must be one of [0, 1, 2], please check it.')

        logE = np.linspace(3, 6, ebins)
        Edisp = np.zeros(ebins)
        for i, logE0 in enumerate(logE):
            try:
                Edisp[i] = self.GetEdisp(evtype, logE0, cosTheta0, ratiolimit)[0]
            except ValueError:
                Edisp[i] = 0

        plt.figure()
        plt.plot(10**logE, Edisp, 'o-')
        plt.xscale('log')
        if evtype == 0:
            plt.xlim(1e3, 1e5)
        if evtype == 1:
            plt.xlim(1e3, 1e6)
        plt.ylim(0, 0.05)
        plt.xlabel('Energy (MeV)')
        plt.ylabel(r'Edisp ($\delta$E/E)')
        plt.title('EDISP_logE_{}_{}'.format(cosTheta0, evtype))
        plt.grid()
        plt.savefig('Edisp_E_{}_{}.png'.format(cosTheta0, evtype))
        return logE, Edisp.ravel()

    def PlotEdisp_Theta(self, evtype, logE0, thetabins, ratiolimit=0.68):
        '''
        Plot EDISP over theta at logE with ratiolimit.
        '''
        assert isinstance(evtype, int)
        assert isinstance(logE0, (int, float))
        assert isinstance(thetabins, int)
        assert isinstance(ratiolimit, float)

        if not evtype in [0, 1, 2]:
            raise RuntimeError('The evtype must be one of [0, 1, 2], please check it.')
        
        Theta = np.linspace(0, 60, thetabins)
        cosTheta = np.cos(Theta / 180. * np.pi)
        Edisp = np.zeros(thetabins)
        for j, cosTheta0 in enumerate(cosTheta):
            try:
                Edisp[j] = self.GetEdisp(evtype, logE0, cosTheta0, ratiolimit)[0]
            except ValueError:
                Edisp[j] = 0

        plt.figure()
        plt.plot(cosTheta, Edisp, 'o-')
        plt.xlim(0.5, 1)
        plt.ylim(0, 0.05)
        plt.xlabel(r'cos$\theta$')
        plt.ylabel(r'Edisp ($\delta$E/E)')
        plt.title('EDISP_cosTheta_{}_{}'.format(logE0, evtype))
        plt.grid()
        plt.savefig('Edisp_Theta_{}_{}.png'.format(logE0, evtype))
        return cosTheta, Edisp
