'''
File: Events.py
Author: Duan Kai-Kai
Email: duankk@pmo.ac.cn
'''

import os
from astropy.coordinates import SkyCoord
from astropy.io import fits
import astropy.units as u
import healpy as hp
import numpy as np
from . import maptools
from . import SkyMap
from . import DmpTime


class Event:
    '''
    The Event class stores the photon information for every photon
    including time, energy, skycrd_C, skycrd_G, theta, phi, evtype
    '''
    def __init__(self, time, energy, ra, dec, l, b, theta, phi, evtype):
        '''
        time: MET
        energy: MeV
        ra: deg
        dec: deg
        l: deg
        b: deg
        theta: deg
        phi: deg
        evtype: [0, 1]
        '''
        self.time = time
        self.energy = energy
        self.skycrd_C = (ra, dec)
        self.skycrd_G = (l, b)
        # self.skycrd = SkyCoord(ra=ra*u.degree, dec=dec*u.degree, frame='fk5')
        self.theta = theta
        self.phi = phi
        self.evtype = evtype


class Events:
    '''
    The Events Class stores information of the photon events and GTI,
    Method: AllSkySelect, CircleSelect, BoxSelect, Cmap, Ccube, Hpxmap, Hpxcube. TimeBin, EnergyBin, PhaseBin.
    evfile: str
    '''
    def __init__(self, evfile):
        if not os.path.exists(evfile):
            raise FileNotFoundError(f'The evfile "{evfile}" not be found, please check it.')

        self.evfile = evfile
        hdulist = fits.open(evfile)
        prihdu = hdulist['PRIMARY']
        self.priheader = prihdu.header
        self.pridata = prihdu.data

        evhdu = hdulist['EVENTS']
        self.evheader = evhdu.header.copy()
        self.evdata = evhdu.data

        self.Time = self.evdata.field('TIME')
        self.Energy = self.evdata.field('ENERGY')

        self.RA = self.evdata.field('RA')
        self.DEC = self.evdata.field('DEC')
        self.skycrd_C = np.array(list(zip(self.RA, self.DEC)))

        self.L = self.evdata.field('L')
        self.B = self.evdata.field('B')
        self.skycrd_G = np.array(list(zip(self.L, self.B)))

        # self.skycrd = SkyCoord(ra=self.RA*u.degree, dec=self.DEC*u.degree, frame='fk5')

        self.Theta = self.evdata.field('THETA')
        self.cosTheta = np.cos(np.deg2rad(self.Theta))
        self.Phi = self.evdata.field('PHI')
        self.LETrigger = self.evdata.field('LETrigger')
        self.HETrigger = self.evdata.field('HETrigger')
        self.Evtype = self.HETrigger

        self.events = []
        for time, energy, ra, dec, l, b, theta, phi, evtype in zip(self.Time, self.Energy, self.RA, self.DEC, self.L, self.B, self.Theta, self.Phi, self.Evtype):
            self.events.append(Event(time, energy, ra, dec, l, b, theta, phi, evtype))

        try:
            self.PulsePhase = self.evdata.field('PULSE_PHASE')
        except:
            pass

        gtihdu = hdulist['GTI']
        self.gtiheader = gtihdu.header
        self.gtidata = gtihdu.data

        self.GTImin = self.gtidata.field('START')
        self.GTImax = self.gtidata.field('STOP')
        self.FD = self.gtidata.field('FD')
        hdulist.close()

    def __repr__(self):
        num = len(self.evdata)
        emin = min(self.Energy)
        emax = max(self.Energy)
        tmin = min(self.Time)
        tmax = max(self.Time)
        HETnum = sum(self.HETrigger)
        LETnum = sum(1-self.HETrigger)

        GTInum = len(self.gtidata)
        GTImin = min(self.GTImin)
        GTImax = max(self.GTImax)
        UTCmin = DmpTime.MET2UTC(GTImin)
        UTCmax = DmpTime.MET2UTC(GTImax)

        summary = f"""
        {num} photons: 
        LET = {LETnum}, HET = {HETnum} 
        Emin = {emin} MeV, Emax = {emax} MeV
        tmin = {tmin}, tmax = {tmax}
        {GTInum} GTI:
        GTI: from {GTImin} to {GTImax}
        UTC: from {UTCmin} to {UTCmax}
        """
        return summary

    def __GTISelect(self, tmin, tmax):
        '''
        Select GTI based on time.
        tmin: MET
        tmax: MET
        '''
        gtidata = self.gtidata[(self.GTImax >= tmin)
                             * (self.GTImin <= tmax)]
        if gtidata.field('START')[0] < tmin:
            gtidata.field('START')[0] = tmin
        if gtidata.field('STOP')[-1] > tmax:
            gtidata.field('STOP')[-1] = tmax
        return gtidata

    def __EventSelect(self, tmin, tmax, emin, emax, cosThetamin, cosThetamax, evtype):
        '''
        Select photons from the events class based on time(tmin, tmax), energy(emin, emax), cosTheta(cosThetamin, cosThetamax) and evtype.
        tmin: MET
        tmax: MET
        emin: MeV
        emax: MeV
        evtype: [0, 1, 2]
        '''
        if evtype == 0:
            evtypefilter = (self.Evtype == 0)
        if evtype == 1:
            evtypefilter = (self.Evtype == 1)
        if evtype == 2:
            evtypefilter = 1
        self.evheader['HISTORY'] = 'evtype = {}'.format(evtype)

        cosThetafilter = (self.cosTheta >= cosThetamin) * (self.cosTheta <= cosThetamax)
        self.evheader['HISTORY'] = '{} < cosTheta < {}'.format(cosThetamin, cosThetamax)

        timefilter = (self.Time >= tmin) * (self.Time <= tmax)
        self.evheader['HISTORY'] = '{} < time < {}'.format(tmin, tmax)

        energyfilter = (self.Energy >= emin) * (self.Energy <= emax)
        self.evheader['HISTORY'] = '{} < energy < {}'.format(emin, emax)

        allfilter = np.array(timefilter * energyfilter * cosThetafilter * evtypefilter, dtype=bool)
        evdata = self.evdata[allfilter]
        return evdata

    def AllSkySelect(self, tmin, tmax, emin, emax, cosThetamin=0.5, cosThetamax=1, evtype=2, outfile='AllSkySelect_events.fits'):
        '''
        Select photons from the events class dependiing on position(GorC, skycrd, rad), time(tmin, tmax), energy(emin, emax) cosTheta(cosThetamin, cosThetamax) and evtype for AllSky. The selected photons will be stored in the outfile.
        tmin: MET
        tmax: MET
        emin: MeV
        emax: MeV
        evtype: [0, 1, 2]
        '''
        assert isinstance(tmin, (int, float))
        assert isinstance(tmax, (int, float))
        assert isinstance(emin, (int, float))
        assert isinstance(emax, (int, float))
        assert isinstance(cosThetamin, (int, float))
        assert isinstance(cosThetamax, (int, float))
        assert isinstance(evtype, int)
        assert isinstance(outfile, str)

        if not tmax > tmin:
            raise RuntimeError('The tmax must be larger than tmin, please check it.')
        if not emax > emin:
            raise RuntimeError('The emax must be larger than emin, please check it.')
        if not cosThetamax > cosThetamin:
            raise RuntimeError('The cosThetamax must be larger than cosThetamin, please check it.')
        if not evtype in [0, 1, 2]:
            raise RuntimeError('The evtype must be one of [0, 1, 2], please check it.')

        evdata = self.__EventSelect(tmin, tmax, emin, emax, cosThetamin, cosThetamax, evtype)
        evhdu = fits.BinTableHDU(data=evdata, header=self.evheader, name='EVENTS')

        gtidata = self.__GTISelect(tmin, tmax)
        gtihdu = fits.BinTableHDU(data=gtidata, header=self.gtiheader, name='GTI')

        prihdu = fits.PrimaryHDU(data=self.pridata, header=self.priheader)
        hdulist = fits.HDUList([prihdu, evhdu, gtihdu])
        hdulist.writeto(outfile, overwrite=True)
        return Events(outfile)

    def CircleSelect(self, GorC, xref, yref, rad, tmin, tmax, emin, emax, cosThetamin=0.5, cosThetamax=1, evtype=2, outfile='selectCircle_events.fits'):
        '''
        Select photons from the events class based on position(GorC, xref, yref, rad), time(tmin, tmax), energy(emin, emax), cosTheta(cosThetamin, cosThetamax) and evtype for a circle shape ROI. The selected photons will be stored in the outfile.
        GorC: ['G', 'C']
        xref: deg
        yref: deg
        rad : deg
        tmin: MET
        tmax: MET
        emin: MeV
        emax: MeV
        evtype: [0, 1, 2]
        '''
        assert isinstance(GorC, str)
        assert isinstance(xref, (int, float))
        assert isinstance(yref, (int, float))
        assert isinstance(rad, (int, float))
        assert isinstance(tmin, (int, float))
        assert isinstance(tmax, (int, float))
        assert isinstance(emin, (int, float))
        assert isinstance(emax, (int, float))
        assert isinstance(cosThetamin, (int, float))
        assert isinstance(cosThetamax, (int, float))
        assert isinstance(evtype, int)
        assert isinstance(outfile, str)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')
        if not 0 <= xref <= 360:
            raise RuntimeError('The xref must be in the range of from 0 to 360, please check it.')
        if not -90 <= yref <= 90:
            raise RuntimeError('The yref must be in the range of from -90 to 90, please check it.')
        if not 0 <= rad <= 90:
            raise RuntimeError('The rad must be in the range of from 0 to 90, please check it.')
        if not tmax > tmin:
            raise RuntimeError('The tmax must be larger than tmin, please check it.')
        if not emax > emin:
            raise RuntimeError('The emax must be larger than emin, please check it.')
        if not cosThetamax > cosThetamin:
            raise RuntimeError('The cosThetamax must be larger than cosThetamin, please check it.')
        if not evtype in [0, 1, 2]:
            raise RuntimeError('The evtype must be one of [0, 1, 2], please check it.')

        evdata = self.__EventSelect(tmin, tmax, emin, emax, cosThetamin, cosThetamax, evtype)

        if GorC == 'G':
            sep = maptools.Sep(np.column_stack([evdata.L, evdata.B]), [xref, yref])
        if GorC == 'C':
            sep = maptools.Sep(np.column_stack([evdata.RA, evdata.DEC]), [xref, yref])
        posfilter = np.array((sep <= rad), dtype=bool)
        self.evheader['HISTORY'] = 'Circle Select distance from ({}, {}, {}) < {}'.format(GorC, xref, yref, rad)

        evdata = evdata[posfilter]
        evhdu = fits.BinTableHDU(data=evdata, header=self.evheader, name='EVENTS')

        gtidata = self.__GTISelect(tmin, tmax)
        gtihdu = fits.BinTableHDU(data=gtidata, header=self.gtiheader, name='GTI')

        prihdu = fits.PrimaryHDU(data=self.pridata, header=self.priheader)
        hdulist = fits.HDUList([prihdu, evhdu, gtihdu])
        hdulist.writeto(outfile, overwrite=True)
        return Events(outfile)

    def BoxSelect(self, GorC, xmin, xmax, ymin, ymax, tmin, tmax, emin, emax, cosThetamin=0.5, cosThetamax=1, evtype=2, outfile='selectedBox_events.fits'):
        '''
        Select photons from the events class based on position(GorC, xmin, xmax, ymin, ymax), time(tmin, tmax), energy(emin, emax), cosTheta(cosThetamin, cosThetamax) and evtype for a box shape ROI. The selected photons will be stored in the outfile.
        GorC: ['G', 'C']
        xmin: deg
        xmax: deg
        ymin: deg
        ymax: deg
        tmin: MET
        tmax: MET
        emin: MeV
        emax: MeV
        evtype: [0, 1, 2]
        '''
        assert isinstance(GorC, str)
        assert isinstance(xmin, (int, float))
        assert isinstance(xmax, (int, float))
        assert isinstance(ymin, (int, float))
        assert isinstance(ymax, (int, float))
        assert isinstance(tmin, (int, float))
        assert isinstance(tmax, (int, float))
        assert isinstance(emin, (int, float))
        assert isinstance(emax, (int, float))
        assert isinstance(cosThetamin, (int, float))
        assert isinstance(cosThetamax, (int, float))
        assert isinstance(evtype, int)
        assert isinstance(outfile, str)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')
        if not 0 <= xmin <= 360:
            raise RuntimeError('The xmin must be in the range of from 0 to 360, please check it.')
        if not 0 <= xmax <= 360:
            raise RuntimeError('The xmax must be in the range of from 0 to 360, please check it.')
        if not -90 <= ymin <= 90:
            raise RuntimeError('The ymin must be in the range of from -90 to 90, please check it.')
        if not -90 <= ymax <= 90:
            raise RuntimeError('The ymax must be in the range of from -90 to 90, please check it.')
        if not ymax > ymin:
            raise RuntimeError('The ymax must be larger than ymin, please check it.')
        if not tmax > tmin:
            raise RuntimeError('The tmax must be larger than tmin, please check it.')
        if not emax > emin:
            raise RuntimeError('The emax must be larger than emin, please check it.')
        if not cosThetamax > cosThetamin:
            raise RuntimeError('The cosThetamax must be larger than cosThetamin, please check it.')
        if not evtype in [0, 1, 2]:
            raise RuntimeError('The evtype must be one of [0, 1, 2], please check it.')

        evdata = self.__EventSelect(tmin, tmax, emin, emax, cosThetamin, cosThetamax, evtype)

        if GorC == 'G':
            if xmax > xmin:
                posfilter = np.array((evdata.L >= xmin) * (evdata.L <= xmax) * (evdata.B >= ymin) * (evdata.B <= ymax), dtype=bool)
            if xmax < xmin:
                posfilter = np.array(((evdata.L >= xmin) + (evdata.L <= xmax)) * (evdata.B >= ymin) * (evdata.B <= ymax), dtype=bool)
        if GorC == 'C':
            if xmax > xmin:
                posfilter = np.array((evdata.RA >= xmin) * (evdata.RA <= xmax) * (evdata.DEC >= ymin) * (evdata.DEC <= ymax), dtype=bool)
            if xmax < xmin:
                posfilter = np.array(((evdata.RA >= xmin) + (evdata.RA <= xmax)) * (evdata.DEC >= ymin) * (evdata.DEC <= ymax), dtype=bool)

        self.evheader['HISTORY'] = '{} in the box {} to {} and {} to {}'.format(GorC, xmin, xmax, ymin, ymax)

        evdata = evdata[posfilter]
        evhdu = fits.BinTableHDU(data=evdata, header=self.evheader, name='EVENTS')

        gtidata = self.__GTISelect(tmin, tmax)
        gtihdu = fits.BinTableHDU(data=gtidata, header=self.gtiheader, name='GTI')

        prihdu = fits.PrimaryHDU(data=self.pridata, header=self.priheader)
        hdulist = fits.HDUList([prihdu, evhdu, gtihdu])
        hdulist.writeto(outfile, overwrite=True)
        return Events(outfile)

    def Cmap(self, outfile, nxpix, nypix, binsz, GorC, xref, yref, proj):
        '''
        Bin the photons into counts map based on the map size(nxpix, nypix, binsz) and wcs(GorC, xref, yref, proj).
        The counts map will be stored in the outfile.
        outfile: str
        nxpix: int
        nypix: int
        binsz: deg
        GorC: ['G', 'C']
        xref: deg
        yref: deg
        proj: ['CAR', 'AIT']
        '''
        assert isinstance(outfile, str)
        assert isinstance(nxpix, int)
        assert isinstance(nypix, int)
        assert isinstance(binsz, (int, float))
        assert isinstance(GorC, str)
        assert isinstance(xref, (int, float))
        assert isinstance(yref, (int, float))
        assert isinstance(proj, str)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')
        if not 0 <= xref <= 360:
            raise RuntimeError('The xref must be in the range of from 0 to 360, please check it.')
        if not -90 <= yref <= 90:
            raise RuntimeError('The yref must be in the range of from -90 to 90, please check it.')
        if not proj in ['CAR', 'AIT']:
            raise RuntimeError('The proj must be one of CAR or AIT, please check it.')

        print('make cmap for the photons.................................')
        wcs = maptools.make_wcs(nxpix, nypix, binsz, binsz, GorC, xref, yref, proj)

        if GorC == 'C':
            skycrd = self.skycrd_C
        if GorC == 'G':
            skycrd = self.skycrd_G

        if skycrd.size == 0:
            cmap = np.zeros((nxpix, nypix), dtype='int')
        else:
            pixcrd = wcs.wcs_world2pix(skycrd, 0)

            xyrange = [[-0.5, nxpix - 0.5], [-0.5, nypix - 0.5]]
            cmap, xedges, yedges = np.histogram2d(
                pixcrd[:, 0], pixcrd[:, 1], [nxpix, nypix], xyrange)
        cmapobj = SkyMap.NewCmap(outfile, cmap, wcs, self.gtidata, self.gtiheader)
        cmapobj.WriteFits()
        return cmapobj

    def Ccube(self, outfile, nxpix, nypix, binsz, GorC, xref, yref, proj, emin, emax, ebins):
        '''
        Bin the photons into counts cube based on the map size(nxpix, nypix, binsz), wcs(GorC. xref, yref, proj) and energy bin(emin, emax, ebins). The counts cube will be stored in the outfile.
        outfile: str
        nxpix: int
        nypix: int
        binsz: deg
        GorC: ['G', 'C']
        xref: deg
        yref: deg
        proj: ['CAR', 'AIT']
        emin: MeV
        emax: MeV
        ebins: int
        '''
        assert isinstance(outfile, str)
        assert isinstance(nxpix, int)
        assert isinstance(nypix, int)
        assert isinstance(binsz, (int, float))
        assert isinstance(GorC, str)
        assert isinstance(xref, (int, float))
        assert isinstance(yref, (int, float))
        assert isinstance(proj, str)
        assert isinstance(emin, (int, float))
        assert isinstance(emax, (int, float))
        assert isinstance(ebins, int)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')
        if not 0 <= xref <= 360:
            raise RuntimeError('The xref must be in the range of from 0 to 360, please check it.')
        if not -90 <= yref <= 90:
            raise RuntimeError('The yref must be in the range of from -90 to 90, please check it.')
        if not proj in ['CAR', 'AIT']:
            raise RuntimeError('The proj must be one of CAR or AIT, please check it.')
        if not emax > emin:
            raise RuntimeError('The emax must be larger than emin, please check it.')
        if not ebins > 1:
            raise RuntimeError('The ebins must be larger than 1, please check it.')

        ebounds = np.logspace(np.log10(emin), np.log10(emax), ebins + 1)
        emins = ebounds[:-1]
        emaxs = ebounds[1:]

        print('make ccube for the photons.................................')
        wcs = maptools.make_wcs(nxpix, nypix, binsz, binsz, GorC, xref, yref, proj)

        if GorC == 'C':
            skycrd = self.skycrd_C
        if GorC == 'G':
            skycrd = self.skycrd_G
        ccube = np.zeros((nxpix, nypix, ebins), dtype='int')
        for i, (emin, emax) in enumerate(zip(emins, emaxs)):
            efilter = (self.Energy >= emin) * (self.Energy < emax)

            skycrd_i = skycrd[efilter]
            if skycrd_i.size == 0:
                continue

            pixcrd = wcs.wcs_world2pix(skycrd_i, 0)

            xyrange = [[-0.5, nxpix - 0.5], [-0.5, nypix - 0.5]]
            cmap, xedges, yedges = np.histogram2d(
                pixcrd[:, 0], pixcrd[:, 1], [nxpix, nypix], xyrange)
            ccube[:, :, i] = cmap
        ccubeobj = SkyMap.NewCcube(outfile, ccube, wcs, emins, emaxs, self.gtidata,self.gtiheader)
        ccubeobj.WriteFits()
        return ccubeobj

    def HpxMap(self, outfile, order, ordering, GorC):
        '''
        Bin the photons into healpix map based on the order, ordering, and reference point(GorC, xref, yref).
        The healpix map will be stored in the outfile.
        outfile: str
        order: int
        ordering: ['RING', 'NESTED']
        GorC: ['G', 'C']
        '''
        assert isinstance(outfile, str)
        assert isinstance(order, int)
        assert isinstance(ordering, str)
        assert isinstance(GorC, str)

        if not ordering in ['RING', 'NESTED']:
            raise RuntimeError('The ordering must be one of RING or NESTED, please check it.')
        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')

        nside = hp.order2nside(order)
        npix = hp.nside2npix(nside)
        print('make hpxmap for the photons................................')
        if GorC == 'C':
            skycrd = self.skycrd_C
        if GorC == 'G':
            skycrd = self.skycrd_G
        if skycrd.size == 0:
            hpxmap = np.zeros(npix, dtype='int')
        else:
            ipix = maptools.skycrd2ipix(ordering, nside, skycrd)

            hpxmap, deges = np.histogram(ipix, npix, [-0.5, npix - 0.5])
        hpxmapobj = SkyMap.NewHpxmap(outfile, hpxmap, nside, ordering, GorC, self.gtidata, self.gtiheader)
        hpxmapobj.WriteFits()
        return hpxmapobj

    def HpxCube(self, outfile, order, ordering, GorC, emin, emax, ebins):
        '''
        Bin the photons into healpix cube based on the order, ordering, reference point(GorC, xref, yref) and energy bin(emin, emax, ebins). The healpix cube will be stored in the outfile.
        outfile: str
        order: int
        ordering: ['RING', 'NESTED']
        GorC: ['G', 'C']
        emin: MeV
        emax: MeV
        ebins: int
        '''
        assert isinstance(outfile, str)
        assert isinstance(order, int)
        assert isinstance(ordering, str)
        assert isinstance(GorC, str)
        assert isinstance(emin, (int, float))
        assert isinstance(emax, (int, float))
        assert isinstance(ebins, int)

        if not ordering in ['RING', 'NESTED']:
            raise RuntimeError('The ordering must be one of RING or NESTED, please check it.')
        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')
        if not emax > emin:
            raise RuntimeError('The emax must be larger than emin, please check it.')
        if not ebins > 1:
            raise RuntimeError('The ebins must be larger than 1, please check it.')

        ebounds = np.logspace(np.log10(emin), np.log10(emax), ebins + 1)
        emins = ebounds[:-1]
        emaxs = ebounds[1:]

        nside = hp.order2nside(order)
        npix = hp.nside2npix(nside)
        print('make hpxcube for the photons................................')
        if GorC == 'C':
            skycrd = self.skycrd_C
        if GorC == 'G':
            skycrd = self.skycrd_G
        hpxcube = np.zeros((npix, ebins), dtype='int')
        for i, (emin, emax) in enumerate(zip(emins, emaxs)):
            efilter = (self.Energy >= emin) * (self.Energy < emax)

            skycrd_i = skycrd[efilter]
            if skycrd_i.size == 0:
                continue

            ipix = maptools.skycrd2ipix(ordering, nside, skycrd_i)

            hpxmap, deges = np.histogram(ipix, npix, [-0.5, npix - 0.5])
            hpxcube[:, i] = hpxmap
        hpxcubeobj = SkyMap.NewHpxcube(outfile, hpxcube, nside, ordering, GorC, emins, emaxs, self.gtidata, self.gtiheader)
        hpxcubeobj.WriteFits()
        return hpxcubeobj

    def TimeBin(self, outfile, tmin, tmax, Tperiod):
        '''
        Binned the photons based on the time.
        outfile: str
        tmin: MET
        tmax: MET
        Tperiod: s
        '''
        assert isinstance(outfile, str)
        assert isinstance(tmin, (int, float))
        assert isinstance(tmax, (int, float))
        assert isinstance(Tperiod, (int, float))

        if not tmax > tmin:
            raise RuntimeError('The tmax must be larger than tmin, please check it.')

        n = int(round((tmax - tmin) / Tperiod))
        
        tmax = tmin + n * Tperiod
        lc, tedges = np.histogram(self.Time, n, [tmin, tmax])
        tmins = tedges[:-1]
        tmaxs = tedges[1:]
        np.savetxt(outfile, np.array([tmins, tmaxs, lc]).T, fmt='%d', header='light curve of counts, tmin, tmax, counts')

    def EnergyBin(self, outfile, emin, emax, ebins):
        '''
        Binned the photons based in the energy.
        outfile: str
        emin: MeV
        emax: MeV
        ebins: int
        '''
        assert isinstance(outfile, str)
        assert isinstance(emin, (int, float))
        assert isinstance(emax, (int, float))
        assert isinstance(ebins, int)

        if not emax > emin:
            raise RuntimeError('The emax must be larger than emin, please check it.')

        logE = np.log10(self.Energy)
        hist, edges = np.histogram(logE, bins=ebins, range=[np.log10(emin), np.log10(emax)])
        emins = 10**edges[:-1]
        emaxs = 10**edges[1:]
        np.savetxt(outfile, np.array([emins, emaxs, hist]).T, fmt='%d', header='energy bins of counts, emin, emax, counts')

    def PhaseBin(self, outfile, phasebins):
        '''
        Binned the photons based on the phase.
        outfile: str
        phasebins: int
        '''
        assert isinstance(outfile, str)
        assert isinstance(phasebins, int)

        phase, pedges = np.histogram(self.PulsePhase, bins=phasebins, range=[0, 1])
        pmins = pedges[:-1]
        pmaxs = pedges[1:]
        np.savetxt(outfile, np.array([pmins, pmaxs, phase]).T, fmt='%f', header='pulse phase of counts, pmin, pmax, phase')

    def cosThetaBin(self, outfile, cosThetamin, cosThetamax, Thetabins):
        assert isinstance(outfile, str)
        assert isinstance(cosThetamin, (int, float))
        assert isinstance(cosThetamax, (int, float))
        assert isinstance(Thetabins, int)

        if not cosThetamin < cosThetamax:
            raise RuntimeError('The cosThetamax must be larger than cosThetamin, please check it.')

        hist, edges = np.histogram(self.cosTheta, bins=Thetabins, range=[cosThetamin, cosThetamax])
        cosThetamins = edges[:-1]
        cosThetamaxs = edges[1:]
        np.savetxt(outfile, np.array([cosThetamins, cosThetamaxs, hist]).T, fmt='%f', header='cosTheta bins of counts, cosThetamin, cosThetamax, counts')

    def PhiBin(self, outfile, phimin, phimax, phibins):
        assert isinstance(outfile, str)
        assert isinstance(phimin, (int, float))
        assert isinstance(phimax, (int, float))
        assert isinstance(phibins, int)

        if not phimin < phimax:
            raise RuntimeError('The phimin must be larger than phimax, please check it.')

        hist, edges = np.histogram(self.Phi, bins=phibins, range=[phimin, phimax])
        phimins = edges[:-1]
        phimaxs = edges[1:]
        np.savetxt(outfile, np.array([phimins, phimaxs, hist]).T, fmt='%d', header='Phi bins of counts, phimin, phimax, counts')


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('evfile', help='The name of EVENT file')
    args = parser.parse_args()
    DmpEvents = Events(args.evfile)

    nxpix = 648
    nypix = 324
    binsz = 0.5
    GorC = 'G'
    xref = 0
    yref = 0
    proj = 'AIT'
    order = 6
    ordering = 'RING'
    emin = np.amin(DmpEvents.Energy)
    emax = np.amax(DmpEvents.Energy)
    ebins = 6
    tmin = np.amin(DmpEvents.Time)
    tmax = np.amax(DmpEvents.Time)
    Tperiod = 24 * 3600

    DmpEvents.Cmap('SkyMap.fits', nxpix, nypix, binsz, GorC, xref, yref, proj)
    DmpEvents.Ccube('SkyCube.fits', nxpix, nypix, binsz, GorC, xref, yref, proj, emin, emax, ebins)
    DmpEvents.HpxMap('SkyMap_hpx.fits', order, ordering, GorC)
    DmpEvents.HpxCube('SkyCube_hpx.fits', order, ordering, GorC, emin, emax, ebins)
    DmpEvents.TimeBin('TimeBin.dat', tmin, tmax, Tperiod)
    DmpEvents.EnergyBin('EnergyBin.dat', emin, emax, ebins)
