'''
File: Exposure.py
Author: Duan Kaikai
Email: duankk@pmo.ac.cn
'''

import os
import numpy as np
import healpy as hp
import matplotlib.pyplot as plt
from DmpST import maptools
from DmpST import DmpIRFs
from astropy.io import fits
from scipy.optimize import brentq, root_scalar
from scipy.integrate import simps
from scipy.interpolate import interp1d


class Exposure:
    '''
    This class stores the livetime cube,
    and allows users to calculate the exposure of the sky,
    the exposure-weighted psf and exposure-weihted deisp
    '''
    def __init__(self, ltcube, irf, evtype, cosThetamin=0.5, cosThetamax=1.0, AeffFile='DAMPE_AEFF.fits', PsfFile='DAMPE_PSF.fits', EdispFile='DAMPE_EDISP.fits'):
        assert isinstance(ltcube, str)
        assert isinstance(irf, str)
        assert isinstance(evtype, int)
        assert isinstance(cosThetamin, (int, float))
        assert isinstance(cosThetamax, (int, float))

        if not os.path.exists(ltcube):
            raise FileNotFoundError(f'The ltcube "{ltcube}" not be found, please check it.')
        if not evtype in [0, 1, 2]:
            raise RuntimeError('The evtype must be one of [0, 1, 2], please check it.')
        if not cosThetamax > cosThetamin:
            raise RuntimeError('The cosThetamax must be larger than cosThetamin, please check it.')

        hdulist = fits.open(ltcube)
        self.ltcube_LL = hdulist['EXPOSURE'].data.field('LL')
        self.ltcube_HL = hdulist['EXPOSURE'].data.field('HL')
        self.ordering = hdulist['EXPOSURE'].header['ORDERING']
        self.nside = hdulist['EXPOSURE'].header['NSIDE']

        xref = hdulist['EXPOSURE'].data.field('RA')
        yref = hdulist['EXPOSURE'].data.field('DEC')
        self.skycrd_C = np.column_stack([xref, yref])
        self.skycrd_G = maptools.skycrdC2G(self.skycrd_C)

        self.CThetamin = hdulist['CTHETABOUNDS'].data.field('CTHETA_MIN')
        self.CThetamax = hdulist['CTHETABOUNDS'].data.field('CTHETA_MAX')
        self.cosTheta = (self.CThetamin + self.CThetamax) / 2
        self.cosTheta = np.sort(self.cosTheta)

        try:
            self.phimin = hdulist['PHIBOUNDS'].data.field('PHI_MIN')
            self.phimax = hdulist['PHIBOUNDS'].data.field('PHI_MAX')
            phi = (self.phimin + self.phimax) / 2
            self.phi = np.sort(phi)
        except:
            pass
        hdulist.close()

        if cosThetamin not in self.CThetamin or cosThetamax not in self.CThetamax:
            raise RuntimeError('cosThetamin and cosThetamax must be in the bounds of ltcube')

        self.irf = irf
        self.evtype = evtype
        self.cosThetafilter = (self.cosTheta >= cosThetamin)\
                            * (self.cosTheta <= cosThetamax)
        self.aeff = DmpIRFs.AEFF(irf, AeffFile=AeffFile)
        self.psf = DmpIRFs.PSF(irf, PsfFile=PsfFile)
        self.edisp = DmpIRFs.EDISP(irf, EdispFile=EdispFile)

    def livetime(self, skycrd, GorC='C', phi_dependence=False):
        '''
        Calculate the livetime for skycrd.
        '''
        assert isinstance(skycrd, (list, tuple, np.ndarray))
        assert isinstance(GorC, str)
        assert isinstance(phi_dependence, bool)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')

        if not isinstance(skycrd, np.ndarray):
            skycrd = np.array(skycrd)
        if GorC == 'G':
            skycrd_G = skycrd.T
            skycrd_C = maptools.skycrdG2C(skycrd).T
        if GorC == 'C':
            skycrd_C = skycrd.T
            skycrd_G = maptools.skycrdG2C(skycrd).T

        ipix = hp.ang2pix(self.nside, *skycrd_C, nest=(self.ordering=='NESTED'), lonlat=True)
        if phi_dependence:
            livetime_LL = self.ltcube_LL[ipix].reshape(ipix.size, self.phi.size, self.cosTheta.size)
            livetime_HL = self.ltcube_HL[ipix].reshape(ipix.size, self.phi.size, self.cosTheta.size)
        else:
            livetime_LL = self.ltcube_LL[ipix, :].reshape(ipix.size, self.cosTheta.size)
            livetime_HL = self.ltcube_HL[ipix, :].reshape(ipix.size, self.cosTheta.size)
        return livetime_LL, livetime_HL


    def livetime_unbinned(self, skycrd, cosTheta, phi=None, GorC='C', phi_dependence=False):
        '''
        Calculate the livetime for skycrd.
        '''
        assert isinstance(skycrd, (list, tuple, np.ndarray))
        assert isinstance(cosTheta, (int, float, list, tuple, np.ndarray))
        assert isinstance(GorC, str)
        assert isinstance(phi_dependence, bool)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')

        if not isinstance(skycrd, np.ndarray):
            skycrd = np.array(skycrd)
        if GorC == 'G':
            skycrd_G = skycrd.T
            skycrd_C = maptools.skycrdG2C(skycrd).T
        if GorC == 'C':
            skycrd_C = skycrd.T
            skycrd_G = maptools.skycrdG2C(skycrd).T

        ipix = hp.ang2pix(self.nside, *skycrd_C, nest=(self.ordering=='NESTED'), lonlat=True)
        if phi_dependence:
            assert isinstance(phi, (int, float, list, tuple, np.ndarray))
            livetime_LL = self.ltcube_LL[ipix, (cosTheta > self.CThetamin)*(cosTheta <= self.CThetamax), (phi>=self.phimin)*(phi<self.phimax)].reshape(ipix.size)
            livetime_HL = self.ltcube_HL[ipix, (cosTheta > self.CThetamin)*(cosTheta <= self.CThetamax), (phi>=self.phimin)*(phi<self.phimax)].reshape(ipix.size)
        else:
            livetime_LL = self.ltcube_LL[ipix, (cosTheta > self.CThetamin)*(cosTheta <= self.CThetamax)].reshape(ipix.size)
            livetime_HL = self.ltcube_HL[ipix, (cosTheta > self.CThetamin)*(cosTheta <= self.CThetamax)].reshape(ipix.size)
        return livetime_LL, livetime_HL

    def pixexp_LE_unbinned(self, skycrd, logE, cosTheta, phi=None, GorC='C', phi_dependence=False):
        assert isinstance(skycrd, (list, tuple, np.ndarray))
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(cosTheta, (int, float, list, tuple, np.ndarray))
        assert isinstance(GorC, str)
        assert isinstance(phi_dependence, bool)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')

        livetime_LL, livetime_HL = self.livetime_unbinned(skycrd, cosTheta, phi, GorC, phi_dependence)
        # livetime = livetime_LL / 8 + livetime_HL / 64
        livetime = livetime_LL * 0.1288 + livetime_HL * 0.0159

        if phi_dependence:
            assert isinstance(phi, (int, float, list, tuple, np.ndarray))
            aeff_LE = self.aeff.GetAeff_phi(0, logE, cosTheta, phi)
            pixexp = aeff_LE * livetime[:, :, :, np.newaxis]
        else:
            aeff_LE = self.aeff.GetAeff(0, logE, cosTheta)
            pixexp = aeff_LE * livetime[:, np.newaxis]
        return pixexp


    def pixexp_HE_unbinned(self, skycrd, logE, cosTheta, phi=None, GorC='C', phi_dependence=False):
        '''
        Calculate the exposure for skycrd at logE of High Energy.
        '''
        assert isinstance(skycrd, (list, tuple, np.ndarray))
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(cosTheta, (int, float, list, tuple, np.ndarray))
        assert isinstance(GorC, str)
        assert isinstance(phi_dependence, bool)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')

        livetime_LL, livetime_HL = self.livetime_unbinned(skycrd, cosTheta, phi, GorC, phi_dependence)
        livetime = livetime_LL + livetime_HL

        if phi_dependence:
            assert isinstance(phi, (int, float, list, tuple, np.ndarray))
            aeff_HE = self.aeff.GetAeff_phi(1, logE, cosTheta, phi)
            pixexp = aeff_HE * livetime[:, :, :, np.newaxis]
        else:
            aeff_HE = self.aeff.GetAeff(1, logE, cosTheta)
            pixexp = aeff_HE * livetime[:, np.newaxis]
        return pixexp


    def expcube_LE(self, skycrd, logE, GorC='C', phi_dependence=False):
        '''
        Calculate the exposure for skycrd at logE of Low Energy.
        '''
        assert isinstance(skycrd, (list, tuple, np.ndarray))
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(GorC, str)
        assert isinstance(phi_dependence, bool)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')

        livetime_LL, livetime_HL = self.livetime(skycrd, GorC, phi_dependence)
        # livetime = livetime_LL / 8 + livetime_HL / 64
        livetime = livetime_LL * 0.1288 + livetime_HL * 0.0159

        if phi_dependence:
            aeff_LE = self.aeff.GetAeff_phi(0, logE, self.cosTheta, self.phi)
            aeff_LE[:, ~self.cosThetafilter, :] = 0
            pixexp = aeff_LE * livetime[:, :, :, np.newaxis]
        else:
            aeff_LE = self.aeff.GetAeff(0, logE, self.cosTheta)
            aeff_LE[~self.cosThetafilter, :] = 0
            pixexp = aeff_LE * livetime[:, :, np.newaxis]
        return pixexp

    def expcube_HE(self, skycrd, logE, GorC='C', phi_dependence=False):
        '''
        Calculate the exposure for skycrd at logE of High Energy.
        '''
        assert isinstance(skycrd, (list, tuple, np.ndarray))
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(GorC, str)
        assert isinstance(phi_dependence, bool)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')

        livetime_LL, livetime_HL = self.livetime(skycrd, GorC, phi_dependence)
        livetime = livetime_LL + livetime_HL

        if phi_dependence:
            aeff_HE = self.aeff.GetAeff_phi(1, logE, self.cosTheta, self.phi)
            aeff_HE[:, ~self.cosThetafilter, :] = 0
            pixexp = aeff_HE * livetime[:, :, :, np.newaxis]
        else:
            aeff_HE = self.aeff.GetAeff(1, logE, self.cosTheta)
            aeff_HE[~self.cosThetafilter, :] = 0
            pixexp = aeff_HE * livetime[:, :, np.newaxis]
        return pixexp

    def pixexp_LE(self, skycrd, logE, GorC='C', phi_dependence=False):
        '''
        Calculate the exposure for skycrd at logE of Low Energy summed over phi.
        '''
        assert isinstance(skycrd, (list, tuple, np.ndarray))
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(GorC, str)
        assert isinstance(phi_dependence, bool)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')

        pixexp = self.expcube_LE(skycrd, logE, GorC, phi_dependence)
        if phi_dependence:
            pixexp = pixexp.sum(axis=1)#sum over phi
        return pixexp

    def pixexp_HE(self, skycrd, logE, GorC='C', phi_dependence=False):
        '''
        Calculate the exposure for skycrd at logE of High Energy summed over phi.
        '''
        assert isinstance(skycrd, (list, tuple, np.ndarray))
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(GorC, str)
        assert isinstance(phi_dependence, bool)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')

        pixexp = self.expcube_HE(skycrd, logE, GorC, phi_dependence)
        if phi_dependence:
            pixexp = pixexp.sum(axis=1)#sum over phi
        return pixexp

    def pixexp(self, skycrd, logE, GorC='C', phi_dependence=False):
        '''
        Calculate the exposure for skycrd at logE summed over theta and phi.
        '''
        assert isinstance(skycrd, (list, tuple, np.ndarray))
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(GorC, str)
        assert isinstance(phi_dependence, bool)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')

        if not isinstance(skycrd, np.ndarray):
            skycrd = np.array(skycrd)
        if not isinstance(logE, np.ndarray):
            logE = np.array(logE, dtype='float64')

        if self.evtype == 0:
            pixexp = self.pixexp_LE(skycrd, logE, GorC, phi_dependence)
        if self.evtype == 1:
            pixexp = self.pixexp_HE(skycrd, logE, GorC, phi_dependence)
        if self.evtype == 2:
            pixexp = self.pixexp_LE(skycrd, logE, GorC, phi_dependence)\
                   + self.pixexp_HE(skycrd, logE, GorC, phi_dependence)
        return pixexp.sum(axis=1)


    def psf_LE(self, logE, dTheta):
        '''
        Calculate PSF profile for Low Energy of dTheta at logE.
        '''
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(dTheta, (int, float, list, tuple, np.ndarray))

        if not isinstance(logE, np.ndarray):
            logE = np.array(logE, dtype='float64')
        if not isinstance(dTheta, np.ndarray):
            dTheta = np.array(dTheta, dtype='float64')

        Sp, x, P = self.psf.GetPsf_dTheta(0, logE, self.cosTheta, dTheta)
        return P / np.deg2rad(Sp)**2

    def psf_HE(self, logE, dTheta):
        '''
        Calculate PSF profile for High Energy of dTheta at logE.
        '''
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(dTheta, (int, float, list, tuple, np.ndarray))

        if not isinstance(logE, np.ndarray):
            logE = np.array(logE, dtype='float64')
        if not isinstance(dTheta, np.ndarray):
            dTheta = np.array(dTheta, dtype='float64')

        Sp, x, P = self.psf.GetPsf_dTheta(1, logE, self.cosTheta, dTheta)
        return P / np.deg2rad(Sp)**2

    def wpsf(self, skycrd, logE, dTheta, GorC='C', phi_dependence=False):
        '''
        Calculate exposure-weighted PSF for skycrd at logE of dTheta.
        '''
        assert isinstance(skycrd, (list, tuple, np.ndarray))
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(dTheta, (int, float, list, tuple, np.ndarray))
        assert isinstance(GorC, str)
        assert isinstance(phi_dependence, bool)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')

        if not isinstance(skycrd, np.ndarray):
            skycrd = np.array(skycrd)
        if not isinstance(logE, np.ndarray):
            logE = np.array(logE, dtype='float64')
        if not isinstance(dTheta, np.ndarray):
            dTheta = np.array(dTheta, dtype='float64')

        #ipix = maptools.skycrd2ipix(self.ordering, self.nside, skycrd_C)
        #wpsf = np.zeros((ipix.size, dTheta.size, self.cosTheta.size, logE.size))

        if self.evtype == 0:
            pixexp = self.pixexp_LE(skycrd, logE, GorC, phi_dependence)[:, np.newaxis, :, :]
            Sp_LE, x_LE, P_LE = self.psf.GetPsf_dTheta(0, logE, self.cosTheta, dTheta)
            wpsf = P_LE[np.newaxis, :, :, :] / np.deg2rad(Sp_LE[np.newaxis, np.newaxis, :, :])**2 * pixexp

        if self.evtype == 1:
            pixexp = self.pixexp_HE(skycrd, logE, GorC, phi_dependence)[:, np.newaxis, :, :]
            Sp_HE, x_HE, P_HE = self.psf.GetPsf_dTheta(1, logE, self.cosTheta, dTheta)
            wpsf = P_HE[np.newaxis, :, :, :] / np.deg2rad(Sp_HE[np.newaxis, np.newaxis, :, :])**2 * pixexp

        if self.evtype == 2:
            pixexp_LE = self.pixexp_LE(skycrd, logE, GorC, phi_dependence)[:, np.newaxis, :, :]
            pixexp_HE = self.pixexp_HE(skycrd, logE, GorC, phi_dependence)[:, np.newaxis, :, :]
            pixexp = pixexp_LE + pixexp_HE
            Sp_LE, x_LE, P_LE = self.psf.GetPsf_dTheta(0, logE, self.cosTheta, dTheta)
            Sp_HE, x_HE, P_HE = self.psf.GetPsf_dTheta(1, logE, self.cosTheta, dTheta)
            wpsf = P_LE[np.newaxis, :, :, :] / np.deg2rad(Sp_LE[np.newaxis, np.newaxis, :, :])**2 * pixexp_LE\
                 + P_HE[np.newaxis, :, :, :] / np.deg2rad(Sp_HE[np.newaxis, np.newaxis, :, :])**2 * pixexp_HE

        wpsf = wpsf.sum(axis=2) / pixexp.sum(axis=2)
        wpsf[np.isnan(wpsf)] = 0
        return wpsf

    def edisp_LE(self, logE, dEratio):
        '''
        Calculate EDISP profile for Low Energy of dEratio at logE.
        '''
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(dEratio, (int, float, list, tuple, np.ndarray))

        if not isinstance(logE, np.ndarray):
            logE = np.array(logE, dtype='float64')
        if not isinstance(dEratio, np.ndarray):
            dEratio = np.array(dEratio, dtype='float64')

        SD, x, D = self.psf.GetEdisp_dEratio(0, logE, self.cosTheta, dEratio)
        return D / np.deg2rad(SD)**2

    def edisp_HE(self, logE, dEratio):
        '''
        Calculate EDISP profile for High Energy of dEratio at logE.
        '''
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(dEratio, (int, float, list, tuple, np.ndarray))

        if not isinstance(logE, np.ndarray):
            logE = np.array(logE, dtype='float64')
        if not isinstance(dEratio, np.ndarray):
            dEratio = np.array(dEratio, dtype='float64')

        SD, x, D = self.psf.GetEdisp_dEratio(1, logE, self.cosTheta, dEratio)
        return D / np.deg2rad(SD)**2


    def wedisp(self, skycrd, logE, dEratio, GorC='C', phi_dependence=False):
        '''
        Calculate exposure-weighted EDISP for skycrd at logE of dEratio.
        '''
        assert isinstance(skycrd, (list, tuple, np.ndarray))
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(dEratio, (int, float, list, tuple, np.ndarray))
        assert isinstance(GorC, str)
        assert isinstance(phi_dependence, bool)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')

        if not isinstance(skycrd, np.ndarray):
            skycrd = np.array(skycrd)
        if not isinstance(logE, np.ndarray):
            logE = np.array(logE, dtype='float64')
        if not isinstance(dEratio, np.ndarray):
            dEratio = np.array(dEratio, dtype='float64')

        #ipix = maptools.skycrd2ipix(self.ordering, self.nside, skycrd_C)
        #wedisp = np.zeros((dEratio.size, self.cosTheta.size, logE.size))

        if self.evtype == 0:
            pixexp = self.pixexp_LE(skycrd, logE, GorC, phi_dependence)[:, np.newaxis, :, :]
            SD_LE, x_LE, D_LE = self.edisp.GetEdisp_dEratio(0, logE, self.cosTheta, dEratio)
            wedisp = D_LE[np.newaxis, :, :, :] / SD_LE[np.newaxis, np.newaxis, :, :] * pixexp

        if self.evtype == 1:
            pixexp = self.pixexp_HE(skycrd, logE, GorC, phi_dependence)[:, np.newaxis, :, :]
            SD_HE, x_HE, D_HE = self.edisp.GetEdisp_dEratio(1, logE, self.cosTheta, dEratio)
            wedisp = D_HE[np.newaxis, :, :, :] / SD_HE[np.newaxis, np.newaxis, :, :] * pixexp

        if self.evtype == 2:
            pixexp_LE = self.pixexp_LE(skycrd, logE, GorC, phi_dependence)[:, np.newaxis, :, :]
            pixexp_HE = self.pixexp_HE(skycrd, logE, GorC, phi_dependence)[:, np.newaxis, :, :]
            pixexp = pixexp_LE + pixexp_HE
            SD_LE, x_LE, D_LE = self.edisp.GetEdisp_dEratio(0, logE, self.cosTheta, dEratio)
            SD_HE, x_HE, D_HE = self.edisp.GetEdisp_dEratio(1, logE, self.cosTheta, dEratio)
            wedisp = D_LE[np.newaxis, :, :, :] / SD_LE[np.newaxis, np.newaxis, :, :] * pixexp_LE\
                   + D_HE[np.newaxis, :, :, :] / SD_HE[np.newaxis, np.newaxis, :, :] * pixexp_HE

        wedisp = wedisp.sum(axis=2) / pixexp.sum(axis=2)
        wedisp[np.isnan(wedisp)] = 0
        return wedisp


    def PlotWpsf_dTheta(self, skycrd, logE, dThetamax, dThetabins, GorC='C', phi_dependence=False):
        '''
        Plot exposure-weighted PSF over dTheta for skycrd at logE.
        '''
        assert isinstance(skycrd, (list, tuple, np.ndarray))
        assert isinstance(logE, (int, float))
        assert isinstance(dThetamax, (int, float))
        assert isinstance(dThetabins, int)
        assert isinstance(GorC, str)
        assert isinstance(phi_dependence, bool)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')

        dTheta = np.linspace(0, dThetamax, dThetabins)
        wpsf = self.wpsf(skycrd, logE, dTheta, GorC, phi_dependence)
        print(2 * np.pi * simps(wpsf.flatten() * np.deg2rad(dTheta), np.deg2rad(dTheta)))

        plt.figure()
        plt.plot(dTheta, wpsf.flatten())
        # plt.xscale('log')
        plt.yscale('log')
        plt.xlabel('dTheta [deg]')
        plt.ylabel('wpsf')
        plt.grid()
        plt.savefig('wpsf_dTheta.pdf')

    def psfcdf(self, skycrd, logE, dThetamax, GorC='C', phi_dependence=False):
        '''
        Integral the exposure-weighted PSF over dTheta for skycrd at logE.
        '''
        assert isinstance(skycrd, (list, tuple, np.ndarray))
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(dThetamax, (int, float))
        assert isinstance(GorC, str)
        assert isinstance(phi_dependence, bool)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')

        dTheta = np.linspace(0, dThetamax, 10000)
        wpsf = self.wpsf(skycrd, logE, dTheta, GorC, phi_dependence)
        cdf = 2 * np.pi * simps(wpsf.flatten() * np.deg2rad(dTheta), np.deg2rad(dTheta))
        return cdf

    def GetWpsf(self, skycrd, logE, ratiolimit=0.68, GorC='C', phi_dependence=False):
        '''
        Calculate the exposure-weighted PSF for skycrd at logE with ratiolimit.
        '''
        assert isinstance(skycrd, (list, tuple, np.ndarray))
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(ratiolimit, float)
        assert isinstance(GorC, str)
        assert isinstance(phi_dependence, bool)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')
        if not 0 < ratiolimit < 1:
            raise RuntimeError('The ratiolimit must be in the range from 0 to 1.')

        psf = brentq(lambda dThetamax: self.psfcdf(skycrd, logE, dThetamax, GorC, phi_dependence) - ratiolimit, 1e-30, 10)
        cdf = self.psfcdf(skycrd, logE, psf, GorC, phi_dependence)
        return psf

    def PlotWpsf_E(self, skycrd, ebins, ratiolimit=0.68, GorC='C', phi_dependence=False):
        '''
        Plot exposure-weighted PSF over energy for skycrd with ratiolimit.
        '''
        assert isinstance(skycrd, (list, tuple, np.ndarray))
        assert isinstance(ebins, int)
        assert isinstance(ratiolimit, float)
        assert isinstance(GorC, str)
        assert isinstance(phi_dependence, bool)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')
        if not 0 < ratiolimit < 1:
            raise RuntimeError('The ratiolimit must be in the range from 0 to 1.')

        logE = np.linspace(3, 6, ebins)
        wpsf = np.zeros(ebins)
        for i in range(ebins):
            wpsf[i] = self.GetWpsf(skycrd, logE[i], ratiolimit, GorC, phi_dependence)

        np.savetxt('wpsf_E.dat', np.array([logE, wpsf]).T)
        plt.figure()
        plt.plot(10**logE, wpsf, 'o-')
        plt.xscale('log')
        # plt.yscale('log')
        plt.xlim(3, 6)
        plt.ylim(bottom=0)
        plt.xlabel('Energy [MeV]')
        plt.ylabel('Angular resolution')
        plt.savefig('wpsf_E.pdf')


    def PlotWedisp_dEratio(self, skycrd, logE, dEratiomax, Eratiobins, GorC='C', phi_dependence=False):
        '''
        Plot exposure-weighted EDISP over dEratio for skycrd at logE.
        '''
        assert isinstance(skycrd, (list, tuple, np.ndarray))
        assert isinstance(logE, (int, float))
        assert isinstance(dEratiomax, (int, float))
        assert isinstance(Eratiobins, int)
        assert isinstance(GorC, str)
        assert isinstance(phi_dependence, bool)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')

        dEratio = np.linspace(-1+1e-5, dEratiomax, Eratiobins)

        wedisp = self.wedisp(skycrd, logE, dEratio, GorC, phi_dependence)
        print(simps(wedisp.flatten(), dEratio))

        plt.plot(dEratio, wedisp.flatten())
        plt.xlim(-1, dEratiomax)
        plt.xlabel('dEratio')
        plt.ylabel('wedisp')
        plt.grid()
        plt.savefig('wedisp_dEratio.pdf')

    def get_highest_point(self, skycrd, logE, dEratiomax, Eratiobins, GorC='C', phi_dependence=False):
        '''
        Get the highest point of the exposure-weighted EDISP profile for skycrd at logE.
        '''
        assert isinstance(skycrd, (list, tuple, np.ndarray))
        assert isinstance(logE, (int, float))
        assert isinstance(dEratiomax, (int, float))
        assert isinstance(Eratiobins, int)
        assert isinstance(GorC, str)
        assert isinstance(phi_dependence, bool)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')

        dEratio = np.linspace(-1, dEratiomax, Eratiobins)
        wedisp = self.wedisp(skycrd, logE, dEratio, GorC, phi_dependence)
        wedisp = wedisp.flatten()
        return float(dEratio[wedisp == max(wedisp)])

    def edispcdf(self, skycrd, logE, dEratiomax, GorC='C', phi_dependence=False):
        '''
        Integral the exposure-weighted PSF over dTheta for skycrd at logE.
        '''
        assert isinstance(skycrd, (list, tuple, np.ndarray))
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(dEratiomax, (int, float))
        assert isinstance(GorC, str)
        assert isinstance(phi_dependence, bool)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')

        dEratio = np.linspace(-1, dEratiomax, 1000)
        wedisp = self.wedisp(skycrd, logE, dEratio, GorC, phi_dependence)
        cdf = simps(wedisp.flatten(), dEratio, axis=0)
        return cdf

    def GetWedisp1(self, skycrd, logE, ratiolimit, GorC='C', phi_dependence=False):
        '''
        Calculate the exposure-weighted EDISP for skycrd at logE with ratiolimit.
        '''
        assert isinstance(skycrd, (list, tuple, np.ndarray))
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(ratiolimit, float)
        assert isinstance(GorC, str)
        assert isinstance(phi_dependence, bool)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')
        if not 0 < ratiolimit < 1:
            raise RuntimeError('The ratiolimit must be in the range from 0 to 1.')

        highest_point = self.get_highest_point(skycrd, logE, 1, 1000, GorC, phi_dependence)
        highest_cdf = self.edispcdf(skycrd, logE, highest_point, GorC, phi_dependence)

        Edisp_left = brentq(lambda x: self.edispcdf(skycrd, logE, highest_point, GorC, phi_dependence) - self.edispcdf(skycrd, logE, x, GorC, phi_dependence) - ratiolimit/2, -1+1e-5, highest_point)
        left_cdf = self.edispcdf(skycrd, logE, highest_point, GorC, phi_dependence) - self.edispcdf(skycrd, logE, Edisp_left, GorC, phi_dependence)

        Edisp_right = brentq(lambda x: self.edispcdf(skycrd, logE, x, GorC, phi_dependence) - self.edispcdf(skycrd, logE, highest_point, GorC, phi_dependence) - ratiolimit/2, highest_point, 1-1e-5)
        right_cdf = self.edispcdf(skycrd, logE, Edisp_right, GorC, phi_dependence) - self.edispcdf(skycrd, logE, highest_point, GorC, phi_dependence)

        Edisp = (Edisp_right - Edisp_left) / 2
        return Edisp

    def GetWedisp(self, skycrd, logE, ratiolimit=0.68, GorC='C', phi_dependence=False):
        '''
        Calculate the exposure-weighted EDISP for skycrd at logE with ratiolimit.
        '''
        assert isinstance(skycrd, (list, tuple, np.ndarray))
        assert isinstance(logE, (int, float, list, tuple, np.ndarray))
        assert isinstance(ratiolimit, float)
        assert isinstance(GorC, str)
        assert isinstance(phi_dependence, bool)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')
        if not 0 < ratiolimit < 1:
            raise RuntimeError('The ratiolimit must be in the range from 0 to 1.')

        cdf = 0
        ymin = 0
        highest_point = self.get_highest_point(skycrd, logE, 1, 1000, GorC, phi_dependence)
        ymax = float(self.wedisp(skycrd, logE, highest_point, GorC, phi_dependence))
        while abs(cdf - ratiolimit) > 1e-3:
            y = (ymin + ymax) / 2
            x1 = brentq(lambda x: float(abs(self.wedisp(skycrd, logE, x, GorC, phi_dependence)) - y), -1, highest_point)
            x2 = brentq(lambda x: float(abs(self.wedisp(skycrd, logE, x, GorC, phi_dependence)) - y), highest_point, 1)
            cdf = self.edispcdf(skycrd, logE, x2, GorC, phi_dependence) - self.edispcdf(skycrd, logE, x1, GorC, phi_dependence)
            if cdf > ratiolimit:
                ymin = y
            if cdf < ratiolimit:
                ymax = y
        Edisp = (x2 - x1) / 2
        return Edisp#, x1, x2

    def PlotWedisp_E(self, skycrd, ebins, ratiolimit=0.68, GorC='C', phi_dependence=False):
        '''
        Plot exposure-weighted EDISP over energy for skycrd with ratiolimit.
        '''
        assert isinstance(skycrd, (list, tuple, np.ndarray))
        assert isinstance(ebins, int)
        assert isinstance(ratiolimit, float)
        assert isinstance(GorC, str)
        assert isinstance(phi_dependence, bool)

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')
        if not 0 < ratiolimit < 1:
            raise RuntimeError('The ratiolimit must be in the range from 0 to 1.')

        logE = np.linspace(3, 6, ebins)
        wedisp = np.zeros(ebins)
        for i in range(ebins):
            wedisp[i] = self.GetWedisp(skycrd, logE[i], ratiolimit, GorC, phi_dependence)

        np.savetxt('wedisp_E.dat', np.array([logE, wedisp]).T)
        plt.figure()
        plt.plot(10**logE, wedisp, 'o-')
        plt.xscale('log')
        plt.xlim(3, 6)
        plt.ylim(bottom=0)
        plt.xlabel(r'Energy (MeV)')
        plt.ylabel(r'Energy resolution ($\delta$E/E)')
        plt.grid()
        plt.savefig('wedisp_E.pdf')


def expCube(ltcube, outfile, irf, evtype,
            nxpix, nypix, binsz, emin, emax, ebins,
            GorC, xref, yref, proj, bincal='EDGE',
            cosThetamin=0.5, cosThetamax=1.0, phi_dependence=False):
    '''
    This function make expcube file depending the ltcube and irf,
    the spatial and energy bin is nxpix, nypix, binsz, emin, emax, ebins,
    the GorC, xref, yref, proj is the wcs infomation.
    '''
    assert isinstance(ltcube, str)
    assert isinstance(outfile, str)
    assert isinstance(irf, str)
    assert isinstance(evtype, int)
    assert isinstance(nxpix, int)
    assert isinstance(nypix, int)
    assert isinstance(binsz, (int, float))
    assert isinstance(emin, (int, float))
    assert isinstance(emax, (int, float))
    assert isinstance(ebins, int)
    assert isinstance(GorC, str)
    assert isinstance(xref, (int, float))
    assert isinstance(yref, (int, float))
    assert isinstance(proj, str)
    assert isinstance(bincal, str)
    assert isinstance(cosThetamin, (int, float))
    assert isinstance(cosThetamax, (int, float))
    assert isinstance(phi_dependence, bool)
    
    if not os.path.exists(ltcube):
        raise FileNotFoundError('The ltcube not found, please check it.')
    if not evtype in [0, 1, 2]:
        raise RuntimeError('The evtype must be one of [0, 1, 2], please check it.')
    if not binsz > 0:
        raise RuntimeError('The binsz must be larger than 0, please check it.')
    if not emax > emin:
        raise RuntimeError('The emax must be larger than emin, please check it.')
    if not ebins > 0:
        raise RuntimeError('The ebins must be larger than 0, please check it.')
    if not GorC in ['G', 'C']:
        raise RuntimeError('The GorC must be one of G or C, please check it.')
    if not 0 <= xref <= 360:
        raise RuntimeError('The xref must be in the range of from 0 to 360, please check it.')
    if not -90 <= yref <= 90:
        raise RuntimeError('The yref must be in the range of from -90 to 90, please check it.')
    if not proj in ['CAR', 'AIT']:
        raise RuntimeError('The proj must be one of CAR or AIT, please check it.')
    if not bincal in ['CENTER', 'EDGE']:
        raise RuntimeError('The bincal must be one of CENTER or EDGE, please check it.')
    if not cosThetamax > cosThetamin:
        raise RuntimeError('The cosThetamax must be larger than cosThetamin, please check it.')

    exposure = Exposure(ltcube, irf, evtype, cosThetamin, cosThetamax)
    wcs = maptools.make_wcs(nxpix, nypix, binsz, binsz, GorC, xref, yref, proj)
    header = wcs.to_header()

    npix = nxpix * nypix
    pixcrd = np.mgrid[0:nxpix, 0:nypix].reshape(2, npix).T

    if GorC == 'G':
        skycrd_G = wcs.all_pix2world(pixcrd, 0)
        skycrd_C = maptools.skycrdG2C(skycrd_G) 
    if GorC == 'C':
        skycrd_C = wcs.all_pix2world(pixcrd, 0)
        skycrd_G = maptools.skycrdC2G(skycrd_C)

    if bincal == 'CENTER':
        logEl = np.linspace(np.log10(emin), np.log10(emax), ebins + 1)[:-1]
        logEh = np.linspace(np.log10(emin), np.log10(emax), ebins + 1)[1:]
        logE = (logEl + logEh) / 2
    if bincal == 'EDGE':
        logE = np.linspace(np.log10(emin), np.log10(emax), ebins + 1)

    try:
        expcube = exposure.pixexp(skycrd_C, logE, 'C', phi_dependence)
    except:
        expcube = np.zeros((npix, logE.size))
        expcube[~np.isnan(skycrd_C[:, 0])] = exposure.pixexp(skycrd_C[~np.isnan(skycrd_C[:, 0])], logE, 'C', phi_dependence)

    expcube = expcube.reshape(nxpix, nypix, logE.size)

    prihdu = fits.PrimaryHDU(expcube.T, header=header)
    ebshdu = fits.BinTableHDU.from_columns([\
        fits.Column(name='Energy', format='E', unit='MeV', array=10**logE)],\
        name='ENERGIES')
    hdulist = fits.HDUList([prihdu, ebshdu])
    hdulist.writeto(outfile, overwrite=True)
    hdulist.close()

def expHpx(ltcube, outfile, irf, evtype, order, ordering,
           emin, emax, ebins, GorC, bincal='EDGE',
           cosThetamin=0.5, cosThetamax=1.0, phi_dependence=False):
    '''
    This function make healpix exposure cube depending the ltcube and irf,
    energy bin is emin, emax and ebins, order is for the pix area,
    ordering is 'RING' or 'NESTED', GorC is 'G' or 'C'.
    '''
    assert isinstance(ltcube, str)
    assert isinstance(outfile, str)
    assert isinstance(irf, str)
    assert isinstance(evtype, int)
    assert isinstance(order, int)
    assert isinstance(ordering, str)
    assert isinstance(emin, (int, float))
    assert isinstance(emax, (int, float))
    assert isinstance(ebins, int)
    assert isinstance(GorC, str)
    assert isinstance(bincal, str)
    assert isinstance(cosThetamin, (int, float))
    assert isinstance(cosThetamax, (int, float))
    assert isinstance(phi_dependence, bool)

    if not os.path.exists(ltcube):
        raise FileNotFoundError('The ltcube not found, please check it.')
    if not evtype in [0, 1, 2]:
        raise RuntimeError('The evtype must be one of [0, 1, 2], please check it.')
    if not ordering in ['RING', 'NESTED']:
        raise RuntimeError('The ordering must be one of RING or NESTED, please check it.')
    if not emax > emin:
        raise RuntimeError('The emax must be larger than emin, please check it.')
    if not ebins > 1:
        raise RuntimeError('The ebins must be larger than 1, please check it.')
    if not GorC in ['G', 'C']:
        raise RuntimeError('The GorC must be one of G or C, please check it.')
    if not bincal in ['CENTER', 'EDGE']:
        raise RuntimeError('The bincal must be one of CENTER or EDGE, please check it.')
    if not cosThetamax > cosThetamin:
        raise RuntimeError('The cosThetamax must be larger than cosThetamin, please check it.')

    exposure = Exposure(ltcube, irf, evtype, cosThetamin, cosThetamax)
    nside = hp.order2nside(order)
    npix = hp.nside2npix(nside)
    ipix = np.arange(npix)

    if GorC == 'G':
        ipix = maptools.ipixG2C(ordering, nside, ipix)
    
    if bincal == 'CENTER':
        logEl = np.linspace(np.log10(emin), np.log10(emax), ebins + 1)[:-1]
        logEh = np.linspace(np.log10(emin), np.log10(emax), ebins + 1)[1:]
        logE = (logEl + logEh) / 2
    if bincal == 'EDGE':
        logE = np.linspace(np.log10(emin), np.log10(emax), ebins + 1)

    skycrd_C = maptools.ipix2skycrd(ordering, nside, ipix)
    exphpx = exposure.pixexp(skycrd_C, logE, 'C', phi_dependence)

    hp.write_map(outfile, exphpx.T, nest=(ordering == 'NESTED'), coord=GorC, fits_IDL=False, overwrite=True)
    ebshdu = fits.BinTableHDU.from_columns(
        [fits.Column(name='ENERGY', format='1E', unit='MeV', array=10**logE)])
    fits.append(outfile, ebshdu.data, ebshdu.header, extname='ENERGY')

def exposure_weighted_psf(ltcube, outfile, irf, evtype, ra, dec,
                          emin=2e3, emax=2e6, nenergies=12,
                          thetamax=30., ntheta=300):
    """
    This function does the same thing as gtpsf.
    It calculate the exposure weighted psf and save to the fits file
    which has the same structure as fermi's.
    emin, emax in [MeV], thetamax in [deg], nenergies and ntheta should
    be integers.
    """
    # calculate the exposure&psf
    a = thetamax/ntheta**2
    thetas = a*np.linspace(0., ntheta, ntheta)**2
    lgE = np.linspace(np.log10(emin), np.log10(emax), nenergies)

    expobj = Exposure(ltcube=ltcube, irf=irf, evtype=evtype)
    wexp = expobj.pixexp([ra, dec], lgE)[0, :] # (lgE.size, )
    wpsf = expobj.wpsf([ra, dec], lgE, thetas) # (thetas.size, lgE.size)

    # construct fits file
    hdu_psf = fits.BinTableHDU.from_columns(
        [fits.Column(name='Energy', format='D', array=10**lgE),
         fits.Column(name='Exposure', format='D', array=wexp),
         fits.Column(name='Psf', format='%iD'%thetas.size, array=wpsf.T)],
        name='PSF')
    hdu_theta = fits.BinTableHDU.from_columns(
        [fits.Column(name='Theta', format='D', array=thetas)],
        name='THETA')

    hdu_psf.header['HISTORY'] = 'ltcube=%s, irf=%s, evtype=%s, ra=%s, dec=%s.' %\
                                (ltcube, irf, evtype, ra, dec)

    hdulist = fits.HDUList()
    hdulist.append(hdu_psf)
    hdulist.append(hdu_theta)
    hdulist.writeto(outfile, overwrite=True)
    hdulist.close()


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('ltcube', help='The name of ltcube file')
    #parser.add_argument('expcube', help='The name of expcube file', default='expcube.fits')
    #parser.add_argument('exphpx', help='The name od exphpx file', default='exphpx.fits')
    args = parser.parse_args()
    ltcube = args.ltcube
    #expcube = args.expcube
    #exphpx = args.exphpx

    irf = 'DmpIRFs'
    evtype = 1
    #expCube(ltcube, expcube, irf, evtype, 648, 324, 0.5, 1e3, 10e3, 1, 'G', 0, 0, 'AIT', 'EDGE')
    #expHpx(ltcube, exphpx, irf, evtype, 6, 'NESTED', 'G', 1e3, 10e3, 1, 'EDGE')
    exposure = Exposure(ltcube, irf, evtype)
    skycrd_C = (0, 0)
    exposure.PlotWpsf_dTheta(skycrd_C, 2, 10, 1000)
    #exposure.PlotWpsf_E(skycrd_C, 20)
    #exposure.PlotWedisp_dEratio(skycrd_C, 2, 10, 1000)
    #exposure.PlotWedisp_E(skycrd_C, 20)

    # exposure_weighted_psf(ltcube, 'wpsf.fits', irf, evtype, skycrd_C[0], skycrd_C[1])
