'''
File: LikelihoodAnalysis.py
Author: Duan Kai-Kai
Email: duankk@pmo.ac.cn
'''
import yaml
import pickle
import numpy as np
from copy import deepcopy
from pprint import pprint
from multiprocessing import Pool
from scipy.interpolate import interp1d

from . import SkyMap
from . import maptools
from .Events import Events
from .Exposure import Exposure
from .SpaceCraft import SpaceCraft

from .Model import Model
from .Source import Source
from .FitRegion import FitRegion
from .BinnedLikelihood import BinnedLikelihood
from .UnbinnedLikelihood import UnbinnedLikelihood


normParList = ['Prefactor', 'Normalization', 'norm', 'Integral', 'Value']
class LikelihoodAnalysis:
    '''
    This class do the likelihood Analysis to get the free parameters' fitted value.
    with the values, you can calculate the sources' flux, energyflux, TS value and/or upperlimit
    It will save the fitted model into yaml file, and also save the results of sources.
    '''

    def Fit(self, mcmc=False, hesse=False, minos=False, Npred=False, Flux=False, EnergyFlux=False, TsValue=False, upperLimit=False, lowerLimit=False, LikelihoodProfile=False, srcmaps=False, mdlcube=False):
        assert isinstance(mcmc, bool)
        assert isinstance(hesse, bool)
        assert isinstance(minos, bool)
        assert isinstance(Npred, bool)
        assert isinstance(Flux, bool)
        assert isinstance(EnergyFlux, bool)
        assert isinstance(TsValue, bool)
        assert isinstance(upperLimit, bool)
        assert isinstance(lowerLimit, bool)
        assert isinstance(LikelihoodProfile, bool)

        if mcmc:
            self.Likelihood.mcmc_pre()
            self.results, self.errors, self.covs, self.likelihoodValue = self.Likelihood.minuit()
            self.Likelihood.mcmc_post()
        else:
            self.results, self.errors, self.covs, self.likelihoodValue = self.Likelihood.minuit(hesse, minos)

        if srcmaps:
            self.Likelihood.SaveSrcMaps(self.srcName + '_srcmaps.fits')
        if mdlcube:
            self.Likelihood.SaveMdlCube(self.srcName + '_mdlcube.fits')

        self.Npreds = {}
        if Npred:
            for srcName in list(self.Likelihood.SrcDict.keys()):
                print('Calculate Npred of '+ srcName)
                self.SetFittedValue()
                Npred = self.Likelihood.mdlcubes[srcName].sum()
                self.Npreds[srcName] = Npred

        self.Fluxes = {}
        if Flux:
            for srcName in list(self.Likelihood.SrcDict.keys()):
                print('Calculate Flux of '+ srcName)
                self.SetFittedValue()
                Flux, Error = self.CalculateFlux(srcName, self.emin, self.emax)
                self.Fluxes[srcName] = (Flux, Error)

        self.EnergyFluxes = {}
        if EnergyFlux:
            for srcName in list(self.Likelihood.SrcDict.keys()):
                print('Calculate EnergyFlux of '+ srcName)
                self.SetFittedValue()
                EnergyFlux, Error = self.CalculateEnergyFlux(srcName, self.emin, self.emax)
                self.EnergyFluxes[srcName] = (EnergyFlux, Error)

        self.TsValues = {}
        if TsValue:
            for srcName in list(self.Likelihood.FreeSrcDict.keys()):
                print('Calculate TsValue of '+ srcName)
                self.SetFittedValue()
                TsValue = self.CalculateTsValue(srcName)
                self.TsValues[srcName] = TsValue

        self.parUpperErrors = {}
        self.parUpperLimits = {}
        self.FluxUpperErrors = {}
        self.FluxUpperLimits = {}
        if upperLimit:
            for freePar in list(self.Likelihood.FreeParsDict.keys()):
                self.SetFittedValue()
                srcName, parName = freePar.split('__')
                if parName in normParList:
                    print('Calculate UpperError of '+freePar)
                    parUpperError, FluxUpperError = self.CalculateUpperError(srcName, parName, 1)
                    self.parUpperErrors[freePar] = parUpperError
                    self.FluxUpperErrors[srcName] = FluxUpperError
                    self.SetFittedValue()
                    print('Calculate UpperLimit of '+srcName)
                    parUpperLimit, FluxUpperLimit = self.CalculateUpperLimit(srcName, parName, 2.71)
                    self.parUpperLimits[freePar] = parUpperLimit
                    self.FluxUpperLimits[srcName] = FluxUpperLimit

        self.parLowerErrors = {}
        self.parLowerLimits = {}
        self.FluxLowerErrors = {}
        self.FluxLowerLimits = {}
        if lowerLimit:
            for freePar in list(self.Likelihood.FreeParsDict.keys()):
                self.SetFittedValue()
                srcName, parName = freePar.split('__')
                if parName in normParList:
                    print('Calculate LowerError of '+freePar)
                    parLowerError, FluxLowerError = self.CalculateLowerError(srcName, parName, 1)
                    self.parLowerErrors[freePar] = parLowerError
                    self.FluxLowerErrors[srcName] = FluxLowerError
                    self.SetFittedValue()
                    print('Calculate LowerLimit of '+srcName)
                    parLowerLimit, FluxLowerLimit = self.CalculateLowerLimit(srcName, parName, 2.71)
                    self.parLowerLimits[freePar] = parLowerLimit
                    self.FluxLowerLimits[srcName] = FluxLowerLimit

        self.parValues = {}
        self.FluxValues = {}
        self.deltaLikelihoods = {}
        if LikelihoodProfile:
            for freePar in list(self.Likelihood.FreeParsDict.keys()):
                srcName, parName = freePar.split('__')
                if parName in normParList:
                    self.SetFittedValue()
                    print('Calculate LikelihoodProfile of '+srcName)
                    self.CalculateLikelihoodProfile(srcName, parName)


    def SetFittedValue(self):
        for freePar in list(self.Likelihood.FreeParsDict.keys()):
            srcName, parName = freePar.split('__')
            parValue = self.results[freePar]
            parError = self.errors[freePar]
            self.Likelihood.SetParValue(srcName, parName, parValue)
            self.Likelihood.SetParError(srcName, parName, parError)


    def CalculateFlux(self, srcName, emin, emax):
        '''
        Calculate the flux of srcName from emin to emax
        '''
        assert isinstance(srcName, str)
        assert isinstance(emin, (int, float))
        assert isinstance(emax, (int, float))

        if not srcName in self.Likelihood.SrcDict.keys():
            raise RuntimeError('The srcName must be in the SrcDict, please check it.')
        if not emax > emin:
            raise RuntimeError('The emax must be larger than emin, please check it.')

        Flux = self.Likelihood.SrcDict[srcName].CalculateFlux(emin, emax)
        Error = 0
        if self.covs == None:
            return Flux, np.sqrt(Error)
        for freepars, parsCov in list(self.covs.items()):
            freepar1, freepar2 = freepars
            srcName1, parName1 = freepar1.split('__')
            srcName2, parName2 = freepar2.split('__')
            if srcName1 == srcName and srcName2 == srcName:
                parValue1 = self.results[freepar1]
                parValue2 = self.results[freepar2]
                parError1 = self.errors[freepar1]
                parError2 = self.errors[freepar2]
        
                self.Likelihood.SetParValue(srcName1, parName1, parValue1 + parError1)
                Flux_up = self.Likelihood.SrcDict[srcName1].CalculateFlux(emin, emax)
                self.Likelihood.SetParValue(srcName1, parName1, parValue1 - parError1)
                Flux_down = self.Likelihood.SrcDict[srcName1].CalculateFlux(emin, emax)
                D1 = (Flux_up - Flux_down) / (2 * parError1)
            
                self.Likelihood.SetParValue(srcName2, parName2, parValue2 + parError2)
                Flux_up = self.Likelihood.SrcDict[srcName2].CalculateFlux(emin, emax)
                self.Likelihood.SetParValue(srcName2, parName2, parValue2 - parError2)
                Flux_down = self.Likelihood.SrcDict[srcName2].CalculateFlux(emin, emax)
                D2 = (Flux_up - Flux_down) / (2 * parError2)

                Error += D1 * D2 * parsCov

        return Flux, np.sqrt(Error)

    def CalculateEnergyFlux(self, srcName, emin, emax):
        '''
        Calculate the energy flux of srcName from emin to emax
        '''
        assert isinstance(srcName, str)
        assert isinstance(emin, (int, float))
        assert isinstance(emax, (int, float))

        if not srcName in self.likelihood.SrcDict.keys():
            raise RuntimeError('The srcName must be in the SrcDict, please check it.')
        if not emax > emin:
            raise RuntimeError('The emax must be larger than emin, please check it.')

        EnergyFlux = self.Likelihood.SrcDict[srcName].CalculateEnergyFlux(emin, emax)
        Error = 0
        if self.covs == None:
            return EnergyFlux, np.sqrt(Error)
        for freepars, parsCov in list(self.covs.items()):
            freepar1, freepar2 = freepars
            srcName1, parName1 = freepar1.split('__')
            srcName2, parName2 = freepar2.split('__')
            if srcName1 == srcName and srcName2 == srcName:
                parValue1 = self.results[freepar1]
                parValue2 = self.results[freepar2]
                parError1 = self.errors[freepar1]
                parError2 = self.errors[freepar2]
        
                self.Likelihood.SetParValue(srcName1, parName1, parValue1 + parError1)
                Flux_up = self.Likelihood.SrcDict[srcName1].CalculateEnergyFlux(emin, emax)
                self.Likelihood.SetParValue(srcName1, parName1, parValue1 - parError1)
                Flux_down = self.Likelihood.SrcDict[srcName1].CalculateEnergyFlux(emin, emax)
                D1 = (Flux_up - Flux_down) / (2 * parError1)
            
                self.Likelihood.SetParValue(srcName2, parName2, parValue2 + parError2)
                Flux_up = self.Likelihood.SrcDict[srcName2].CalculateEnergyFlux(emin, emax)
                self.Likelihood.SetParValue(srcName2, parName2, parValue2 - parError2)
                Flux_down = self.Likelihood.SrcDict[srcName2].CalculateEnergyFlux(emin, emax)
                D2 = (Flux_up - Flux_down) / (2 * parError2)

                Error += D1 * D2 * parsCov

        return EnergyFlux, np.sqrt(Error)

    def CalculateUpperError(self, srcName, parName, deltaLikelihoodlimit):
        '''
        Calculate the upper limit of parName of srcName with deltaLikelihoodlimit.
        '''
        assert isinstance(srcName, str)
        assert isinstance(parName, str)
        assert isinstance(deltaLikelihoodlimit, (int, float))

        if not srcName in self.Likelihood.SrcDict.keys():
            raise RuntimeError('The srcName must be in the SrcDict, please check it.')
        if not parName in self.Likelihood.SrcDict[srcName].ParsDict.keys():
            raise RuntimeError('The parName must be in the ParsDict, please check it.')
        if not (srcName+'__'+parName in self.results.keys()):
            raise RuntimeError('The srcName__parName is not in the results.')
        if not (srcName+'__'+parName in self.errors.keys()):
            raise RuntimeError('The srcName__parName is not in the errors.')

        self.Likelihood.SetFix(srcName, parName)
        parValue = self.results[srcName+'__'+parName]
        parError = self.errors[srcName+'__'+parName]

        likelihoodValue = self.Likelihood.minuit()[3]
        deltaLikelihood = likelihoodValue - self.likelihoodValue
        if deltaLikelihood > deltaLikelihoodlimit:
            raise RuntimeError(f'{srcName}_{parName}_{parValue} Best Fit Value is wrong.')

        parValue1 = parValue
        parValue2 = parValue
        loop = 0
        while deltaLikelihood < deltaLikelihoodlimit:
            loop += 1
            parValue2 += parError * 2**loop
            self.Likelihood.SetParValue(srcName, parName, parValue2)
            likelihoodValue = self.Likelihood.minuit()[3]
            deltaLikelihood = likelihoodValue - self.likelihoodValue
            #if loop > 10:
            #    raise RuntimeError(f'parValue_{parValue} and parError_{parError} are wrong!')
        

        parValue = parValue2
        distance = abs(deltaLikelihood - deltaLikelihoodlimit)
        while distance > 1e-2:
            parValue_upper = parValue
            deltaLikelihood_upper = deltaLikelihood

            parValue = (parValue1 + parValue2) / 2
            self.Likelihood.SetParValue(srcName, parName, parValue)
            likelihoodValue = self.Likelihood.minuit()[3]
            deltaLikelihood = likelihoodValue - self.likelihoodValue
            distance = abs(deltaLikelihood - deltaLikelihoodlimit)

            slope = (deltaLikelihood_upper - deltaLikelihood) / (parValue_upper - parValue)
            if slope <= 0:
                raise RuntimeError(f'The {srcName}_{parName} {parValue} slope {slope} of deltalikelihood {deltaLikelihood} is smaller than 0.')

            if deltaLikelihood > deltaLikelihoodlimit:
                parValue2 = parValue
            if deltaLikelihood < deltaLikelihoodlimit:
                parValue1 = parValue
        parUpperError = parValue
        self.Likelihood.SetParValue(srcName, parName, parUpperError)
        FluxUpperError = self.Likelihood.SrcDict[srcName].CalculateFlux(self.emin, self.emax)

        self.Likelihood.SetFree(srcName, parName)
        return parUpperError, FluxUpperError

    def CalculateLowerError(self, srcName, parName, deltaLikelihoodlimit):
        '''
        Calculate the lower limit of parName of srcName with deltaLikelihoodlimit.
        '''
        assert isinstance(srcName, str)
        assert isinstance(parName, str)
        assert isinstance(deltaLikelihoodlimit, (int, float))

        if not srcName in self.Likelihood.SrcDict.keys():
            raise RuntimeError('The srcName must be in the SrcDict, please check it.')
        if not parName in self.Likelihood.SrcDict[srcName].ParsDict.keys():
            raise RuntimeError('The parName must be in the ParsDict, please check it.')
        if not (srcName+'__'+parName in self.results.keys()):
            raise RuntimeError('The srcName__parName is not in the results.')
        if not (srcName+'__'+parName in self.errors.keys()):
            raise RuntimeError('The srcName__parName is not in the errors.')

        self.Likelihood.SetFix(srcName, parName)
        parValue = self.results[srcName+'__'+parName]
        parError = self.errors[srcName+'__'+parName]

        likelihoodValue = self.Likelihood.minuit()[3]
        deltaLikelihood = likelihoodValue - self.likelihoodValue
        if deltaLikelihood > deltaLikelihoodlimit:
            print('Best Fit Error:', parValue, deltaLikelihood)
            raise RuntimeError('Best Fit Value is wrong, you should refit it.')

        parValue1 = parValue
        parValue2 = parValue
        loop = 0
        while deltaLikelihood < deltaLikelihoodlimit:
            loop += 1
            parValue2 -= parError * 2**loop
            self.Likelihood.SetParValue(srcName, parName, parValue2)
            likelihoodValue = self.Likelihood.minuit()[3]
            deltaLikelihood = likelihoodValue - self.likelihoodValue
            if loop > 10:
                print(parValue, parError)
                break
        if deltaLikelihood < deltaLikelihoodlimit:
            self.Likelihood.SetFree(srcName, parName)
            return 0, 0

        parValue = parValue2
        distance = abs(deltaLikelihood - deltaLikelihoodlimit)
        while distance > 1e-3:
            parValue_lower = parValue
            deltaLikelihood_lower = deltaLikelihood

            parValue = (parValue1 + parValue2) / 2
            self.Likelihood.SetParValue(srcName, parName, parValue)
            likelihoodValue = self.Likelihood.minuit()[3]
            deltaLikelihood = likelihoodValue - self.likelihoodValue
            distance = abs(deltaLikelihood - deltaLikelihoodlimit)

            slope = (deltaLikelihood_lower - deltaLikelihood) / (parValue - parValue_lower)
            if slope <= 0:
                print(srcName, parName, parValue, deltaLikelihood, deltaLikelihoodlimit)
                raise RuntimeError('The slope of deltalikelihood is smaller than 0.')

            if deltaLikelihood > deltaLikelihoodlimit:
                parValue2 = parValue
            if deltaLikelihood < deltaLikelihoodlimit:
                parValue1 = parValue
        parLowerError = parValue
        self.Likelihood.SetParValue(srcName, parName, parLowerError)
        FluxLowerError = self.Likelihood.SrcDict[srcName].CalculateFlux(self.emin, self.emax)

        self.Likelihood.SetFree(srcName, parName)
        return parLowerError, FluxLowerError

    def CalculateUpperLimit(self, srcName, normPar, deltaLikelihoodlimit):
        '''
        Calculate the upper limit of normPar of srcName with deltaLikelihoodlimit.
        '''
        assert isinstance(srcName, str)
        assert isinstance(normPar, str)
        assert isinstance(deltaLikelihoodlimit, (int, float))

        return self.CalculateUpperError(srcName, normPar, deltaLikelihoodlimit)

    def CalculateLowerLimit(self, srcName, normPar, deltaLikelihoodlimit):
        '''
        Calculate the lower limit of normPar of srcName with deltaLikelihoodlimit.
        '''
        assert isinstance(srcName, str)
        assert isinstance(normPar, str)
        assert isinstance(deltaLikelihoodlimit, (int, float))

        return self.CalculateLowerError(srcName, normPar, deltaLikelihoodlimit)

    def CalculateLikelihoodProfile(self, srcName, normPar):
        '''
        Calculate the likelihood profile of parName of srcName.
        '''
        assert isinstance(srcName, str)
        assert isinstance(normPar, str)

        if not srcName in self.Likelihood.SrcDict.keys():
            raise RuntimeError('The srcName must be in the SrcDict, please check it.')
        if not normPar in self.Likelihood.SrcDict[srcName].ParsDict.keys():
            raise RuntimeError('The parName must be in the ParsDict, please check it.')
        if not (srcName+'__'+normPar in self.results.keys()):
            raise RuntimeError('The srcName__parName is not in the results.')
        if not (srcName+'__'+normPar in self.errors.keys()):
            raise RuntimeError('The srcName__parName is not in the errors.')

        try:
            parUpperLimit = self.parUpperLimits[srcName+'__'+normPar]
        except KeyError:
            parUpperLimit = self.CalculateUpperLimit(srcName, normPar, 2.71)
        try:
            parLowerLimit = self.parLowerLimits[srcName+'__'+normPar]
        except KeyError:
            parLowerLimit = self.CalculateLowerLimit(srcName, normPar, 2.71)
        
        N = 100
        deltaPar = (parUpperLimit - parLowerLimit) / N
        self.Likelihood.SetFix(srcName, normPar)
        parValues = []
        FluxValues = []
        deltaLikelihoods = []
        for i in range(N):
            parValue = parLowerLimit + i * deltaPar
            self.Likelihood.SetParValue(srcName, normPar, parValue)
            likelihoodValue = self.Likelihood.minuit()[3]
            deltaLikelihood = likelihoodValue - self.likelihoodValue
            Flux = self.Likelihood.SrcDict[srcName].CalculateFlux(self.emin, self.emax)
            FluxValues.append(Flux)
            parValues.append(parValue)
            deltaLikelihoods.append(deltaLikelihood)
        self.Likelihood.SetFree(srcName, normPar)

        self.parValues[srcName] = parValues
        self.FluxValues[srcName] = FluxValues
        self.deltaLikelihoods[srcName] = deltaLikelihoods

    def CalculateTsValue(self, srcName):
        '''
        Calculate the TS value of srcName.
        '''
        assert isinstance(srcName, str)

        if not srcName in self.Likelihood.FreeSrcDict.keys():
            raise RuntimeError('The srcName must be in the FreeSrcDict, please check it.')

        ChangedPars = []
        ChangedValues = []
        Source = self.Likelihood.FreeSrcDict[srcName]
        for parName in list(Source.FreeParsDict.keys()):
            ChangedPars.append(srcName+'__'+parName)
            ChangedValue = Source.GetParValue(parName)
            ChangedValues.append(ChangedValue)
            self.Likelihood.SetFix(srcName, parName)
            if parName in normParList:
                self.Likelihood.SetParValue(srcName, parName, 1e-5)

        likelihoodValue = self.Likelihood.minuit()[3]
        TsValue = likelihoodValue - self.likelihoodValue
        if TsValue < 0:
            tol = 0.01
            likelihoodValue = self.Likelihood.minuit(tol=tol)[3]
            TsValue = likelihoodValue - self.likelihoodValue

        for ChangedPar, ChangedValue in zip(ChangedPars, ChangedValues):
            srcName, parName = ChangedPar.split('__')
            self.Likelihood.SetFree(srcName, parName)
            self.Likelihood.SetParValue(srcName, parName, ChangedValue)

        return TsValue


    def resTSmap(self, srcName, outfile):
        '''
        Calculate the residual TS map as outfile.
        '''
        assert isinstance(outfile, str)

        testSourceDict = self.Likelihood.FreeSrcDict[srcName].srcDict
        testSourceDict['name'] = 'testSource'
        testSourceDict['spatialModel']['type'] = 'SkyDirFunction'
        testSourceDict['spectrum']['Index']['free'] = 0

        self.SetFittedValue()
        self.likelihoodValue0 = self.Likelihood.minuit()[3]
        argsList = [(self.Likelihood, skycrd_C, self.likelihoodValue0, testSourceDict) for skycrd_C in self.ccube.skycrd_C]
        pool = Pool(5)
        tsmap = pool.map(makeTSmap_mp, argsList)
        pool.close()
        pool.join()

        #skycrds_C= [skycrd_C for skycrd_C in self.ccube.skycrd_C]
        #tsmap = np.zeros(self.ccube.nxpix * self.ccube.nypix)
        #for i, skycrd in enumerate(skycrds_C):
        #    tsmap[i] = self.makeTSmap(skycrd)

        TSmap = np.array(tsmap).reshape(self.ccube.nxpix, self.ccube.nypix)
        SkyMap.NewCmap(outfile, TSmap, self.ccube.wcs).WriteFits()
        del tsmap, TSmap, argsList

    def srcTSmap(self, srcName, outfile):
        '''
        Calculate the source TS map as outfile.
        '''
        assert isinstance(outfile, str)

        testSourceDict = self.Likelihood.FreeSrcDict[srcName].srcDict
        testSourceDict['name'] = 'testSource'
        testSourceDict['spatialModel']['type'] = 'SkyDirFunction'
        testSourceDict['spectrum']['Index']['free'] = 0

        self.SetFittedValue()
        self.Likelihood.DelSource(srcName)
        self.likelihoodValue0 = self.Likelihood.minuit()[3]
        argsList = [(self.Likelihood, skycrd_C, self.likelihoodValue0, testSourceDict) for skycrd_C in self.ccube.skycrd_C]
        pool = Pool(5)
        tsmap = pool.map(makeTSmap_mp, argsList)
        pool.close()
        pool.join()

        #skycrds_C= [skycrd_C for skycrd_C in self.ccube.skycrd_C]
        #tsmap = np.zeros(self.ccube.nxpix * self.ccube.nypix)
        #for i, skycrd in enumerate(skycrds_C):
        #    Tsmap[i] = self.makeTSmap(skycrd)

        TSmap = np.array(tsmap).reshape(self.ccube.nxpix, self.ccube.nypix)
        SkyMap.NewCmap(outfile, TSmap, self.ccube.wcs).WriteFits()
        del tsmap, TSmap, argsList

    def makeTSmap(self, skycrd_C, testSourceDict):
        tsLikelihood = deepcopy(self.Likelihood)
        ra, dec = skycrd_C
        testSourceDict['spatialModel']['RA']['value'] = ra
        testSourceDict['spatialModel']['DEC']['value'] = dec
        testSource = Source(testSourceDict)
        self.SetFittedValue()
        tsLikelihood.AddSource('testSource', testSource)
        tslikelihoodValue = tsLikelihood.minuit()[3]
        TsValue = self.likelihoodValue0 - tslikelihoodValue
        del tsLikelihood, testSource, testSourceDict
        return TsValue

    def SaveSrcMaps(self, outfile):
        '''
        Save srcmaps as outfile.
        '''
        assert isinstance(outfile, str)

        self.Likelihood.SaveSrcMaps(outfile)

    def SaveMdlCube(self, outfile):
        '''
        Save mdlcube as outfile.
        '''
        assert isinstance(outfile, str)

        self.Likelihood.SaveMdlCube(outfile)

    def SaveFitModel(self, outfile):
        '''
        Save fitted model as outfile.
        '''
        assert isinstance(outfile, str)

        for freePar in list(self.Likelihood.FreeParsDict.keys()):
            srcName, parName = freePar.split('__')
            parValue = self.results[freePar]
            parError = self.errors[freePar]
            self.model.ChangeParValue(srcName, parName, parValue, parError)
        self.model.SaveModel(outfile)

    def SaveFitResults(self, outfile):
        '''
        Save fitted results as outfile.
        '''
        assert isinstance(outfile, str)

        results = {}
        results['LikelihoodValue'] = float(self.likelihoodValue)
        for srcName in list(self.Likelihood.SrcDict.keys()):
            result = {}
            for parName, parameter in list(self.Likelihood.SrcDict[srcName].FixParsDict.items()):
                if parName in normParList:
                    result[parName+'_value'] = '{:.2f}'.format(parameter.value)
                    result[parName+'_scale'] = '{:.2e}'.format(parameter.scale)
                else:
                    result[parName+'_value'] = '{:.2f}'.format(parameter.value * parameter.scale)
                
            for parName, parameter in list(self.Likelihood.SrcDict[srcName].FreeParsDict.items()):
                result[parName+'_value'] = '{:.2f}'.format(self.results[srcName + '__' + parName])
                result[parName+'_error'] = '{:.2f}'.format(self.errors[srcName + '__' + parName])
                result[parName+'_scale'] = '{:.2e}'.format(parameter.scale)
                try:
                    result[parName+'_UpperError'] = '{:.2f}'.format(self.parUpperErrors[srcName + '__' + parName])
                except KeyError:
                    pass
                try:
                    result[parName+'_LowerError'] = '{:.2f}'.format(self.parLowerErrors[srcName + '__' + parName])
                except KeyError:
                    pass
                try:
                    result[parName+'_UpperLimit'] = '{:.2f}'.format(self.parUpperLimits[srcName + '__' + parName])
                except KeyError:
                    pass
                try:
                    result[parName+'_LowerLimit'] = '{:.2f}'.format(self.parLowerLimits[srcName + '__' + parName])
                except KeyError:
                    pass

            try:
                result['Npred'] = '{:.2f}'.format(self.Npreds[srcName])
            except KeyError:
                pass
            try:
                result['ROI_Center_Distance'] = '{}'.format(self.Likelihood.SrcDict[srcName].srcDict['ROI_Center_Distance'])
            except KeyError:
                pass
            try:
                result['Flux_value'] = '{:.2e}'.format(self.Fluxes[srcName][0])
                result['Flux_error'] = '{:.2e}'.format(self.Fluxes[srcName][1])
            except KeyError:
                pass
            try:
                result['EnergyFlux_value'] = '{:.2e}'.format(self.EnergyFluxes[srcName][0])
                result['EnergyFlux_error'] = '{:.2e}'.format(self.EnergyFluxes[srcName][1])
            except KeyError:
                pass
            try:
                result['TsValue'] = float(self.TsValues[srcName])
            except KeyError:
                pass

            try:
                result['Flux_UpperError'] = '{:.2e}'.format(self.FluxUpperErrors[srcName])
            except KeyError:
                pass
            try:
                result['Flux_LowerError'] = '{:.2e}'.format(self.FluxLowerErrors[srcName])
            except KeyError:
                pass
            try:
                result['Flux_UpperLimit'] = '{:.2e}'.format(self.FluxUpperLimits[srcName])
            except KeyError:
                pass
            try:
                result['Flux_LowerLimit'] = '{:.2e}'.format(self.FluxLowerLimits[srcName])
            except KeyError:
                pass

            results[srcName] = result
        with open(outfile, 'w') as f:
            yaml.dump(results, f)

    def SaveFitCovariance(self, outfile):
        '''
        Save fitted covariance as outfile.
        '''
        assert isinstance(outfile, str)

        Covariance = {}
        for freepars, parsCov in list(self.covs.items()):
            freepar1, freepar2 = freepars
            Covariance[freepar1 + '___' + freepar2] = parsCov
        with open(outfile, 'w') as f:
            yaml.dump(Covariance, f)

    def SaveLikelihoodProfile(self, outfile):
        '''
        Save likelihood profile as outfile.
        '''
        assert isinstance(outfile, str)

        for srcName in list(self.Likelihood.FreeSrcDict.keys()):
            parValues = self.parValues[srcName]
            FluxValues = self.FluxValues[srcName]
            deltaLikelihoods = self.deltaLikelihoods[srcName]
            np.savetxt(srcName+'_'+outfile+'.dat', np.array([parValues, FluxValues, deltaLikelihoods]).T, header='parValue, FlusValues, deltaLikelihood')


class BinnedAnalysis(LikelihoodAnalysis):
    def __init__(self, **kwd):
        pprint('Binned Analysis parameters:')
        pprint(kwd)
        self.srcName = kwd['srcName']
        CubeType = kwd['CubeType']
        assert isinstance(CubeType, str)
        if not CubeType in ['Map', 'Hpx', 'MaskedHpx']:
            raise RuntimeError('The CubeType must be one of Map or Hpx, please check it.')

        ccube = kwd.get('ccube', None)
        self.mask = None
        if ccube:
            if CubeType == 'Map':
                self.ccube = SkyMap.Ccube(ccube)
            if CubeType == 'Hpx':
                self.ccube = SkyMap.HpxCcube(ccube)
            if CubeType == 'MaskedHpx':
                self.mask = kwd.get('mask', None)
                self.ccube = SkyMap.MaskedHpxCcube(ccube, self.mask)
        else:
            evfile = kwd['evfile']
            GorC = kwd['GorC']
            xref = kwd['xref']
            yref = kwd['yref']
            proj = kwd['proj']
            emin = kwd['emin']
            emax = kwd['emax']
            ebins = kwd['ebins']

            events = Events(evfile)
            if CubeType == 'Map':
                nxpix = kwd['nxpix']
                nypix = kwd['nypix']
                binsz = kwd['binsz']
                ccube = self.srcName + '_ccube.fits'
                self.ccube = events.Ccube(ccube, nxpix, nypix, binsz,
                                          GorC, xref, yref, proj, emin, emax, ebins)

            if CubeType == 'Hpx':
                order = kwd['order']
                ordering = kwd['ordering']
                hpxcube = self.srcName + 'hpxcube.fits'
                self.ccube = events.HpxCube(hpxcube, order, ordering,
                                            GorC, xref, yref, emin, emax, ebins)

        self.phi_dependence = kwd.get('phi_dependence', False)
        srcmaps = kwd.get('srcmaps', None)
        print(srcmaps)
        try:
            if self.mask:
                self.srcmaps = SkyMap.ReadMaskedHpxSrcMaps(srcmaps, self.mask)
            else:
                self.srcmaps = SkyMap.ReadSrcMaps(srcmaps)
        except:
            self.srcmaps = {}

        fitRegion = kwd.get('fitRegion', None)
        if fitRegion:
            self.fitRegion = FitRegion(fitRegion)
        else:
            self.fitRegion = None

        irf = kwd['irf']
        evtype = kwd['evtype']
        ltcube = kwd['ltcube']
        cosThetamin = kwd['cosThetamin']
        cosThetamax = kwd['cosThetamax']
        self.exposure = Exposure(ltcube, irf, evtype, cosThetamin, cosThetamax)

        self.emin = kwd['emin']
        self.emax = kwd['emax']
        srcmdl = kwd['srcmdl']
        self.model = Model(srcmdl)
        self.Likelihood = BinnedLikelihood(CubeType, self.ccube, self.exposure, self.emin, self.emax, self.srcmaps, self.fitRegion, self.phi_dependence)

        for srcName, SourceObj in list(self.model.SrcDict.items()):
            print('Add ', srcName)
            self.Likelihood.AddSource(srcName, SourceObj)


class UnbinnedAnalysis(LikelihoodAnalysis):
    def __init__(self, **kwd):
        print('Unbinned Analysis parameters:')
        pprint(kwd)
        self.srcName = kwd['srcName']
        CubeType = kwd['CubeType']
        assert isinstance(CubeType, str)
        if not CubeType in ['Map', 'Hpx']:
            raise RuntimeError('The CubeType must be one of Map or Hpx, please check it.')

        events = kwd['events']
        if 'pkl' in events:
            self.events = pickle.load(open(events, 'rb'))
        else:
            self.events = Events(events)
            scfile = kwd['scfile']
            self.spacecraft = SpaceCraft(scfile)

        ccube = kwd.get('ccube', None)
        if ccube:
            self.ccube = SkyMap.Ccube(ccube)
        else:
            GorC = kwd['GorC']
            xref = kwd['xref']
            yref = kwd['yref']
            proj = kwd['proj']
            emin = kwd['emin']
            emax = kwd['emax']
            ebins = kwd['ebins']

            if CubeType == 'Map':
                skycrd = kwd['skycrd']
                nxpix = kwd['nxpix']
                nypix = kwd['nypix']
                binsz = kwd['binsz']
                ccube = self.srcName + '_ccube.fits'
                self.ccube = self.events.Ccube(ccube, nxpix, nypix, binsz,
                                          GorC, xref, yref, proj, emin, emax, ebins)

            if CubeType == 'Hpx':
                order = kwd['order']
                ordering = kwd['ordering']
                hpxcube = self.srcName + 'hpxcube.fits'
                self.ccube = self.events.HpxCube(hpxcube, order, ordering,
                                            GorC, skycrd, emin, emax, ebins)

        phi_dependence = kwd.get('phi_dependence', False)
        srcmaps = kwd.get('srcmaps', None)
        try:
            if srcmaps:
                self.srcmaps = SkyMap.ReadSrcMaps(srcmaps)
            else:
                self.srcmaps = {}
        except:
            self.srcmaps = {}

        fitRegion = kwd.get('fitRegion', None)
        if fitRegion:
            self.fitRegion = FitRegion(fitRegion)
        else:
            self.fitRegion = None

        irf = kwd['irf']
        evtype = kwd['evtype']
        ltcube = kwd['ltcube']
        cosThetamin = kwd['cosThetamin']
        cosThetamax = kwd['cosThetamax']
        self.exposure = Exposure(ltcube, irf, evtype, cosThetamin, cosThetamax)

        self.emin = kwd['emin']
        self.emax = kwd['emax']
        srcmdl = kwd['srcmdl']
        self.model = Model(srcmdl)
        self.events = events_in_ccube(self.events, self.ccube)
        self.Likelihood = UnbinnedLikelihood(CubeType, self.events, self.ccube, self.exposure, self.emin, self.emax, self.srcmaps, self.fitRegion, phi_dependence)

        for srcName, SourceObj in list(self.model.SrcDict.items()):
            print('Add ', srcName)
            self.Likelihood.AddSource(srcName, SourceObj)

        if not hasattr(self.events, 'PsfProbs'):
            print('Calculate PsfProbs for events......')
            self.CalPsfProbs(self.events.evfile.replace('.fits', '_CalPsfProbs.pkl'))

    def CalPsfProbs(self, outfile):
        self.events.PsfProbs = []
        for srcName, Source in list(self.Likelihood.SrcDict.items()):
            CalPsfProbs(Source, self.events, self.spacecraft, self.exposure, self.ccube)

        for event in self.events.events:
            self.events.PsfProbs.append(event.psfProb)

        self.events.PsfProbs = np.array(self.events.PsfProbs)
        pickle.dump(self.events, open(outfile, 'wb'))


def events_in_ccube(events, ccube):
    wcs = ccube.wcs
    events_in_ccube = []
    Energy = []
    Skycrd_C = []
    Skycrd_G = []
    for event in events.events:
        skycrd_C = event.skycrd_C
        skycrd_G = event.skycrd_G
        if wcs.axis_type_names == ['RA', 'DEC']:
            pixcrd = wcs.wcs_world2pix([skycrd_C], 0)[0]
        if wcs.axis_type_names == ['GLON', 'GLAT']:
            pixcrd = wcs.wcs_world2pix([skycrd_G], 0)[0]
        if -0.5 <= pixcrd[0] <= ccube.nxpix-0.5 and -0.5 <= pixcrd[1] <= ccube.nypix-0.5:
            events_in_ccube.append(event)
            Energy.append(event.energy)
            Skycrd_C.append(event.skycrd_C)
            Skycrd_G.append(event.skycrd_G)

    assert len(events_in_ccube) == ccube.data.sum()
    events.events = events_in_ccube
    events.Energy = np.array(Energy)
    events.Skycrd_C = np.array(Skycrd_C)
    events.Skycrd_G = np.array(Skycrd_G)
    return events

def CalPsfProbs(Source, events, spacecraft, exposure, ccube):
    for i, event in enumerate(events.events):
        E = event.energy
        logE = float(np.log10(E))
        time = event.time
        evtype = event.evtype

        if not hasattr(event, 'psfProb'):
            event.psfProb = {}

        if Source.type == 'PointSource':
            skycrd_C = Source.spatialModel.GetSkycrd()
            Theta = spacecraft.ObservedAngle('C', *skycrd_C, time)
            cosTheta = float(np.cos(np.deg2rad(Theta)))
            dTheta = maptools.Sep(event.skycrd_C, skycrd_C)
            psfProb = 0
            if evtype in [0, 2]:
                srcExp = exposure.pixexp_LE_unbinned(skycrd_C, logE, cosTheta)
                Sp, x, P = exposure.psf.GetPsf_dTheta(0, logE, cosTheta, dTheta)
                psfProb += sum(srcExp * P / np.deg2rad(Sp)**2 * ccube.dOmega0)
            if evtype in [1, 2]:
                srcExp = exposure.pixexp_HE_unbinned(skycrd_C, logE, cosTheta)
                Sp, x, P = exposure.psf.GetPsf_dTheta(1, logE, cosTheta, dTheta)
                psfProb += sum(srcExp * P / np.deg2rad(Sp)**2 * ccube.dOmega0)

        else:
            # nlarger = 1
            # LargerMap = ccube.get_larger_map(nlarger, nlarger)
            if Source.spatialModel.type == 'SpatialMap':
                SpatialMap = Source.spatialModel.GetSpatialMap()
                skycrd_C = SpatialMap.skycrd_C
                dOmega = SpatialMap.dOmega
                larger_spatialmap = maptools.from_Map_to_Map(SpatialMap, ccube)

            if Source.spatialModel.type == 'MapCubeFunction':
                skycrd_C = ccube.skycrd_C
                dOmega = ccube.dOmega
                SpatialCube = Source.spatialModel.GetSpatialCube()
                larger_spatialCube = maptools.from_Map_to_Map(SpatialCube, ccube)
                larger_spatialmap = 10**interp1d(np.log10(SpatialCube.ebounds), np.log10(larger_spatialCube.data), fill_value='extrapolate')(logE)

            if Source.spatialModel.type == 'ConstantValue':
                skycrd_C = ccube.skycrd_C
                dOmega = ccube.dOmega
                larger_spatialmap = np.ones((ccube.nxpix, ccube.nypix)) / (4*np.pi)

            psfProb = 0
            for skycrd0_C, spatialmap, dOmega0 in zip(skycrd_C, larger_spatialmap.ravel(), dOmega.ravel()):
                print(skycrd0_C, spatialmap, dOmega0)
                # LargerSep = maptools.Sep(LargerMap.skycrd0_C, LargerMap.skycrd_C).reshape(LargerMap.nxpix, LargerMap.nypix)
                Theta = spacecraft.ObservedAngle('C', *skycrd0_C, time)
                cosTheta = float(np.cos(np.deg2rad(Theta)))
                dTheta = maptools.Sep(event.skycrd_C, skycrd0_C)
                if exposure.evtype in [0, 2]:
                    srcExp = exposure.pixexp_LE_unbinned(skycrd0_C, logE, cosTheta)
                    Sp, x, P = exposure.psf.GetPsf_dTheta(0, logE, cosTheta, dTheta)
                    psfProb += sum(sum(spatialmap * srcExp * P / np.deg2rad(Sp)**2 * dOmega0))

                if exposure.evtype in [1, 2]:
                    srcExp = exposure.pixexp_HE_unbinned(skycrd0_C, logE, cosTheta)
                    Sp, x, P = exposure.psf.GetPsf_dTheta(1, logE, cosTheta, dTheta)
                    psfProb += sum(sum(spatialmap * srcExp * P / np.deg2rad(Sp)**2 * dOmega0))

        event.psfProb[Source.name] = psfProb
        print(i, event.psfProb)


def makeTSmap_mp(args):
    Likelihood = args[0]
    skycrd_C = args[1]
    likelihoodValue = args[2]
    testSourceDict = args[3]
    tsLikelihood = deepcopy(Likelihood)
    testSourceDict = deepcopy(testSourceDict)
    ra, dec = skycrd_C
    testSourceDict['spatialModel']['RA']['value'] = ra
    testSourceDict['spatialModel']['DEC']['value'] = dec
    testSource = Source(testSourceDict)
    tsLikelihood.AddSource('testSource', testSource)
    tslikelihoodValue = tsLikelihood.minuit()[3]
    TsValue = likelihoodValue - tslikelihoodValue
    tsLikelihood.DelSource('testSource')
    del tsLikelihood, testSource, testSourceDict, Likelihood
    return TsValue
