'''
File: LikelihoodBase.py
Author: Duan Kaikai
Email: duankk@pmo.ac.cn
'''

from pprint import pprint
import numpy as np
import numpy.ma as ma
import healpy as hp
from astropy.io import fits
from . import maptools
from . import SkyMap
import yaml
import math

# import emcee
# import corner
from iminuit import Minuit
from iminuit.util import make_func_code
from multiprocessing import Pool
from scipy.interpolate import interp1d


def MapConvolvePSF(SpatialModel, E, ccube, exposure, phi_dependence):
    import numpy as np
    logE = np.log10(E)

    if SpatialModel.type == 'SkyDirFunction':
        skycrd0_C = SpatialModel.GetSkycrd()

        dTheta = maptools.Sep(skycrd0_C, ccube.skycrd_C).reshape(ccube.nxpix, ccube.nypix)

        #cosTheta Binned Method
        srcmap = np.zeros((ccube.nxpix, ccube.nypix, exposure.cosTheta.size, E.size))
        if exposure.evtype in [0, 2]:
            srcExp = exposure.pixexp_LE(skycrd0_C, logE, phi_dependence=phi_dependence)
            psfmap = exposure.psf_LE(logE, dTheta) * ccube.dOmega[:, :, np.newaxis, np.newaxis]
            srcmap += srcExp * psfmap
        if exposure.evtype in [1, 2]:
            srcExp = exposure.pixexp_HE(skycrd0_C, logE, phi_dependence=phi_dependence)
            psfmap = exposure.psf_HE(logE, dTheta) * ccube.dOmega[:, :, np.newaxis, np.newaxis]
            srcmap += srcExp * psfmap
        srcmap = srcmap.sum(axis=-2)


    else:
        if SpatialModel.type == 'MapCubeFunction': #galactic model
            SpatialCube = SpatialModel.GetSpatialCube()
            Spatialdata = maptools.from_Map_to_Map(SpatialCube, ccube)

            spatialmap = 10**(interp1d(np.log10(SpatialCube.ebounds), np.log10(Spatialdata), fill_value='extrapolate')(logE))[:, :, np.newaxis, :] * ccube.dOmega[:, :, np.newaxis, np.newaxis]

        elif SpatialModel.type == 'HpxSpatialCube':
            SpatialCube = SpatialModel.GetSpatialCube()
            Spatialdata = maptools.from_Hpx_to_Map(SpatialCube, ccube)

            spatialmap = 10**(interp1d(np.log10(SpatialCube.ebounds), np.log10(Spatialdata), fill_value='extrapolate')(logE))[:, :, np.newaxis, :] * ccube.dOmega[:, :, np.newaxis, np.newaxis]

        elif SpatialModel.type == 'ConstantValue': #isotropic mode
            spatialmap = np.ones((ccube.nxpix, ccube.nypix))[:, :, np.newaxis, np.newaxis] * ccube.dOmega[:, :, np.newaxis, np.newaxis]

        else: #extend source
            SpatialMap = SpatialModel.GetSpatialMap()
            spatialmap = maptools.from_Map_to_Map(SpatialMap, ccube)[:, :, np.newaxis, np.newaxis] * ccube.dOmega[:, :, np.newaxis, np.newaxis]

        dTheta = maptools.Sep(ccube.skycrd0_C, ccube.skycrd_C).reshape(ccube.nxpix, ccube.nypix)
        srcmap = np.zeros((ccube.nxpix, ccube.nypix, exposure.cosTheta.size, E.size))
        from scipy.signal import fftconvolve
        if exposure.evtype in [0, 2]:
            srcExp = exposure.pixexp_LE(ccube.skycrd_C, logE, phi_dependence=phi_dependence).reshape(srcmap.shape)
            psfmap = exposure.psf_LE(logE, dTheta) * ccube.dOmega[:, :, np.newaxis, np.newaxis]
            psfmap[round(ccube.nxpix/2 - 0.5), round(ccube.nypix/2 - 0.5), :, :] += 1 - psfmap.sum(axis=(0, 1))
            # print(psfmap.sum(axis=(0, 1)))
            srcmap += fftconvolve(srcExp * spatialmap, psfmap, mode='same', axes=(0, 1))

        if exposure.evtype in [1, 2]:
            srcExp = exposure.pixexp_HE(ccube.skycrd_C, logE, phi_dependence=phi_dependence).reshape(srcmap.shape)
            psfmap = exposure.psf_HE(logE, dTheta) * ccube.dOmega[:, :, np.newaxis, np.newaxis]
            psfmap[round(ccube.nxpix/2 - 0.5), round(ccube.nypix/2 - 0.5), :, :] += 1 - psfmap.sum(axis=(0, 1))
            # print(psfmap.sum(axis=(0, 1)))
            srcmap += fftconvolve(srcExp * spatialmap, psfmap, mode='same', axes=(0, 1))

        srcmap = srcmap.sum(axis=-2)
    return srcmap

def HpxConvolvePSF(SpatialModel, E, HpxCube, exposure, phi_dependence):
    logE = np.log10(E)

    if SpatialModel.type == 'SkyDirFunction':
        skycrd0_C = SpatialModel.GetSkycrd
        ipix0 = hp.ang2pix(HpxCube.nside, *skycrd0_C, nest=True, lonlat=True)
        if HpxCube.type == 'MaskedHpx':
            ipix0 = sum(HpxCube.mask[:ipix0]) - 1
        dTheta = maptools.Sep(skycrd0_C, HpxCube.skycrd_C)

        #cosTheta Binned Method
        srcmap = np.zeros((HpxCube.npix, exposure.cosTheta.size, E.size))
        if exposure.evtype in [0, 2]:
            srcExp = exposure.pixexp_LE(skycrd0_C, logE, phi_dependence=phi_dependence)
            psfmap = exposure.psf_LE(logE, dTheta) * HpxCube.dOmega
            srcmap += srcExp * psfmap
        if exposure.evtype in [1, 2]:
            srcExp = exposure.pixexp_HE(skycrd0_C, logE, phi_dependence=phi_dependence)
            psfmap = exposure.psf_HE(logE, dTheta) * HpxCube.dOmega
            srcmap += srcExp * psfmap
        srcmap = srcmap.sum(axis=-2)

    else:
        if SpatialModel.type == 'HpxSpatialMap':
            SpatialMap = SpatialModel.GetSpatialMap()
            spatialmap = maptools.from_Hpx_to_Hpx(SpatialMap, HpxCube)
            spatialHpx = spatialmap.data * spatialmap.dOmega

        elif SpatialModel.type == 'HpxSpatialCube':
            SpatialCube = SpatialModel.GetSpatialCube()
            Spatialdata = maptools.from_Hpx_to_Hpx(SpatialCube, HpxCube)

            spatialHpx = 10**(interp1d(np.log10(SpatialCube.ebounds), np.log10(Spatialdata), fill_value='extrapolate')(logE)) * HpxCube.dOmega

        elif SpatialModel.type == 'MapCubeFunction':
            SpatialCube = SpatialModel.GetSpatialCube()
            Spatialdata = maptools.from_Map_to_Hpx(SpatialCube, HpxCube)

            spatialHpx = 10**(interp1d(np.log10(SpatialCube.ebounds), np.log10(Spatialdata), fill_value='extrapolate')(logE)) * HpxCube.dOmega

        elif SpatialModel.type == 'ConstantValue':
            spatialHpx = np.ones(HpxCube.npix)[:, np.newaxis] * HpxCube.dOmega

        else:
            SpatialMap = SpatialModel.GetSpatialMap()
            spatialHpx = maptools.from_Map_to_Hpx(SpatialMap, HpxCube)[:, np.newaxis, np.newaxis] * HpxCube.dOmega

        #ignore the PSF distribution
        srcExp = exposure.pixexp(HpxCube.skycrd_C, logE, phi_dependence=phi_dependence)
        srcmap = srcExp * spatialHpx

        '''
        if HpxCube.coord == 'G':
            dTheta = maptools.Sep(HpxCube.skycrd_G, np.array([0, 0]))
        if HpxCube.coord == 'C':
            dTheta = maptools.Sep(HpxCube.skycrd_C, np.array([0, 0]))

        #cosTheta Binned Method
        lmax = 3 * HpxCube.nside - 1
        srcmap = np.zeros((HpxCube.npix, exposure.cosTheta.size, E.size))
        if exposure.evtype in [0, 2]:
            srcExp = exposure.pixexp_LE(HpxCube.skycrd_C, logE, phi_dependence=phi_dependence)
            psfmap = exposure.psf_LE(logE, dTheta) * HpxCube.dOmega
            for i in range(E.size):
                for j in range(exposure.cosTheta.size):
                    psfAl = hp.map2alm(psfmap[:, j, i], mmax=0)\
                          * np.sqrt(4*np.pi / (2 * np.arange(0, lmax+1) + 1))
                    mapAlm = hp.map2alm(srcExp[:, j, i] * spatialHpx[:, 0, i])
                    srcmapAlm = hp.almxfl(mapAlm, psfAl)
                    srcmap[:, j, i] += hp.alm2map(srcmapAlm, HpxCube.nside)
        if exposure.evtype in [1, 2]:
            srcExp = exposure.pixexp_HE(HpxCube.skycrd_C, logE, phi_dependence=phi_dependence)
            psfmap = exposure.psf_HE(logE, dTheta) * HpxCube.dOmega
            for i in range(E.size):
                for j in range(exposure.cosTheta.size):
                    psfAl = hp.map2alm(psfmap[:, j, i], mmax=0)\
                          * np.sqrt(4*np.pi / (2 * np.arange(0, lmax+1) + 1))
                    mapAlm = hp.map2alm(srcExp[:, j, i] * spatialHpx[:, 0, i])
                    srcmapAlm = hp.almxfl(mapAlm, psfAl)
                    srcmap[:, j, i] += hp.alm2map(srcmapAlm, HpxCube.nside)
        srcmap = srcmap.sum(axis=-2)
        '''

    return srcmap

class FuncCode(object):
    __slots__ = ('co_varnames', 'co_argcount')

class LikelihoodBase:
    ''' 
    This class store and manage the Source Dict and ParDict

    def __init__(self, ccube, exposure, emin, emax, srcmaps=None):
        self.modelDict = {}
        self.SrcDict = {}
        self.ParsDict = {}
        self.FixSrcDict = {}
        self.FreeSrcDict = {}
        self.FixParsDict = {}
        self.FreeParsDict = {}
    '''

    def __str__(self):
        return str(self.modelDict)

    def __repr(self):
        return str(self.modelDict)


    def AddSource(self, srcName, SourceObj):
        '''
        This method will add source object into the Likelihood class
        the source Dict will add the object and it will add into
        Fix or Free source and parameter depending on the parameter fix or free.
        '''
        assert isinstance(srcName, str)

        self.modelDict[srcName] = SourceObj.srcDict
        self.SrcDict[srcName] = SourceObj
        if SourceObj.FreeParsDict:
            self.FreeSrcDict[srcName] = SourceObj
        else:
            self.FixSrcDict[srcName] = SourceObj

        for parName, parameter in list(SourceObj.ParsDict.items()):
            self.ParsDict[srcName + '__' + parName] = parameter
            if SourceObj.GetParFree(parName):
                self.FreeParsDict[srcName + '__' + parName] = parameter
            else:
                self.FixParsDict[srcName + '__' + parName] = parameter

    def DelSource(self, srcName):
        '''
        This method will delete source from the srcDict and modelDict, and delete all the parameters about this source from the parDict.
        '''
        assert isinstance(srcName, str)

        if not srcName in self.SrcDict.keys():
            raise RuntimeError('The srcName not in the srcDict, please check it.')

        SourceObj = self.SrcDict[srcName]
        del self.modelDict[srcName]
        del self.SrcDict[srcName]
        if SourceObj.FreeParsDict == {}:
            del self.FixSrcDict[srcName]
        else:
            del self.FreeSrcDict[srcName]

        for parName, parameter in list(SourceObj.ParsDict.items()):
            del self.ParsDict[srcName + '__' + parName]
            if SourceObj.GetParFree(parName):
                del self.FreeParsDict[srcName + '__' + parName]
            else:
                del self.FixParsDict[srcName + '__' + parName]


    def SetFree(self, srcName, parName):
        '''
        This method will set one parameter of one source free
        change the free parameter dict and fix parameter dict
        '''
        assert isinstance(srcName, str)
        assert isinstance(parName, str)

        if not srcName in self.SrcDict.keys():
            raise RuntimeError('The srcName not in the srcDict, please check it.')
        if not parName in self.SrcDict[srcName].ParsDict.keys():
            raise RuntimeError('The parName not in the parDict, please check it.')
        if not (srcName + '__' + parName) in self.ParsDict.keys():
            raise RuntimeError('The srcName__parName not in the ParsDict, please check it.')

        if not self.SrcDict[srcName].FreeParsDict:
            self.FreeSrcDict[srcName] = self.FixSrcDict.pop(srcName)
        self.FreeParsDict[srcName + '__' + parName] = self.FixParsDict.pop(srcName + '__' + parName)
        self.SetParFree(srcName, parName, True)

    def SetFix(self, srcName, parName):
        '''
        This method can set one parameter of one source fix
        change the free parameter dict and fix parameter dict
        '''
        assert isinstance(srcName, str)
        assert isinstance(parName, str)

        if not srcName in self.SrcDict.keys():
            raise RuntimeError('The srcName not in the srcDict, please check it.')
        if not parName in self.SrcDict[srcName].ParsDict.keys():
            raise RuntimeError('The parName not in the parDict, please check it.')
        if not (srcName + '__' + parName) in self.ParsDict.keys():
            raise RuntimeError('The srcName__parName not in the ParsDict, please check it.')

        self.SetParFree(srcName, parName, False)
        self.FixParsDict[srcName + '__' + parName] = self.FreeParsDict.pop(srcName + '__' + parName)
        if not self.SrcDict[srcName].FreeParsDict:
            self.FixSrcDict[srcName] = self.FreeSrcDict.pop(srcName)


    def SetParFree(self, srcName, parName, new_free):
        try:
            self.modelDict[srcName]['spectrum'][parName]['free'] = new_free
        except:
            self.modelDict[srcName]['spatialModel'][parName]['free'] = new_free
        self.SrcDict[srcName].SetParFree(parName, new_free)
        self.ParsDict[srcName + '__' + parName].SetFree(new_free)

    def SetParValue(self, srcName, parName, new_value):
        try:
            self.modelDict[srcName]['spectrum'][parName]['value'] = new_value
        except:
            self.modelDict[srcName]['spatialModel'][parName]['value'] = new_value
        self.SrcDict[srcName].SetParValue(parName, new_value)
        self.ParsDict[srcName + '__' + parName].SetValue(new_value)

    def SetParError(self, srcName, parName, new_error):
        try:
            self.modelDict[srcName]['spectrum'][parName]['error'] = new_error
        except:
            self.modelDict[srcName]['spatialModel'][parName]['error'] = new_error
        self.SrcDict[srcName].SetParError(parName, new_error)
        self.ParsDict[srcName + '__' + parName].SetError(new_error)

    def SetParScale(self, srcName, parName, new_scale):
        try:
            self.modelDict[srcName]['spectrum'][parName]['scale'] = new_scale
        except:
            self.modelDict[srcName]['spatialModel'][parName]['scale'] = new_scale
        self.SrcDict[srcName].SetParScale(parName, new_scale)
        self.ParsDict[srcName + '__' + parName].SetScale(new_scale)


    def FixModelCube(self):
        '''
        if self.ebins < 10 * (math.log10(self.emax) - math.log10(self.emin)):
            bins = round(10 * (math.log10(self.emax) - math.log10(self.emin)) / self.ebins)
        else:
            bins = 1
        '''
        bins = 1
        Fix_mdlcubes = np.zeros_like(self.N, dtype='float')
        ebounds = np.logspace(np.log10(self.emin), np.log10(self.emax), self.ebins*bins + 1)

        for srcName, FixSource in list(self.FixSrcDict.items()):
            try: 
                if self.ccube.type == 'Map':
                    srcmap = self.srcmaps[srcName][:, :, self.imin*bins:(self.imax+1)*bins+1]
                if self.ccube.type in ['Hpx', 'MaskedHpx']:
                    srcmap = self.srcmaps[srcName][:, self.imin*bins:(self.imax+1)*bins+1]
            except KeyError:
                # print('Calculate the srcmap for', srcName)
                SpatialModel = FixSource.spatialModel
                if self.ccube.type == 'Map':
                    srcmap = MapConvolvePSF(SpatialModel, ebounds, self.ccube, self.exposure, self.phi_dependence)
                if self.ccube.type in ['Hpx', 'MaskedHpx']:
                    srcmap = HpxConvolvePSF(SpatialModel, ebounds, self.ccube, self.exposure, self.phi_dependence)
                self.srcmaps[srcName] = srcmap
            mdlcube = FixSource.MdlCube(self.CubeType, srcmap, ebounds)

            Fix_mdlcube = np.zeros_like(self.N, dtype='float')
            for i in range(self.imax-self.imin+1):
                if self.ccube.type == 'Map':
                    Fix_mdlcube[:, :, i] = mdlcube[:, :, i*bins:(i+1)*bins].sum(axis=2)
                if self.ccube.type in ['Hpx', 'MaskedHpx']:
                    Fix_mdlcube[:, i] = mdlcube[:, i*bins:(i+1)*bins].sum(axis=1)
            self.mdlcubes[srcName] = Fix_mdlcube
            #if self.ccube.type == 'Map':
                #print(srcName, Fix_mdlcube.sum(axis=(0, 1)), Fix_mdlcube.sum())
            #if self.ccube.type == 'Hpx':
                #print(srcName, Fix_mdlcube.sum(axis=0), Fix_mdlcube.sum())
            Fix_mdlcubes += Fix_mdlcube
        return Fix_mdlcubes

    def FreeModelCube(self, **kwd):
        '''
        This method calculate the model counts cube for free source
        depend on the energy range you have setted,
        the free parameter will be setted as the argument for fitting
        convolve the Spatial Map and the PSF, then Integrate on energy
        '''
        for key, value in list(kwd.items()):
            srcName, parName = key.split('__')
            self.SetParValue(srcName, parName, value)

        '''
        if self.ebins < 10 * (math.log10(self.emax) - math.log10(self.emin)):
            bins = round(10 * (math.log10(self.emax) - math.log10(self.emin)) / self.ebins)
        else:
            bins = 1
        '''
        bins = 1
        Free_mdlcubes = np.zeros_like(self.N, dtype='float')
        ebounds = np.logspace(np.log10(self.emin), np.log10(self.emax), self.ebins*bins + 1)
        for srcName, FreeSource in list(self.FreeSrcDict.items()):
            try:
                if self.ccube.type == 'Map':
                    srcmap = self.srcmaps[srcName][:, :, self.imin*bins:(self.imax+1)*bins+1]
                if self.ccube.type in ['Hpx', 'MaskedHpx']:
                    srcmap = self.srcmaps[srcName][:, self.imin*bins:(self.imax+1)*bins+1]
            except KeyError:
                print('Calculate the srcmap for', srcName)
                SpatialModel = FreeSource.spatialModel
                if self.ccube.type == 'Map':
                    srcmap = MapConvolvePSF(SpatialModel, ebounds, self.ccube, self.exposure, self.phi_dependence)
                if self.ccube.type in ['Hpx', 'MaskedHpx']:
                    srcmap = HpxConvolvePSF(SpatialModel, ebounds, self.ccube, self.exposure, self.phi_dependence)
                self.srcmaps[srcName] = srcmap
            mdlcube = FreeSource.MdlCube(self.CubeType, srcmap, ebounds)

            Free_mdlcube = np.zeros_like(self.N, dtype='float')
            for i in range(self.imax-self.imin+1):
                if self.ccube.type == 'Map':
                    Free_mdlcube[:, :, i] = mdlcube[:, :, i*bins:(i+1)*bins].sum(axis=2)
                if self.ccube.type in ['Hpx', 'MaskedHpx']:
                    Free_mdlcube[:, i] = mdlcube[:, i*bins:(i+1)*bins].sum(axis=1)
            self.mdlcubes[srcName] = Free_mdlcube
            # if self.ccube.type == 'Map':
            #     print(srcName, Free_mdlcube.sum(axis=(0, 1)), Free_mdlcube.sum())
            # if self.ccube.type in ['Hpx', 'MaskedHpx']:
            #     print(srcName, Free_mdlcube.sum(axis=0), Free_mdlcube.sum())
            Free_mdlcubes += Free_mdlcube

        return Free_mdlcubes


    def SaveSrcMaps(self, outfile):
        '''
        As a middle production, you can save the srcmaps file
        and for the same model and exposure, it can be reused.
        '''
        assert isinstance(outfile, str)

        self.ccube.header['EXTNAME'] = 'SKYMAP'
        hdu0 = fits.PrimaryHDU(self.ccube.data.T, self.ccube.header)
        hdu1 = fits.BinTableHDU.from_columns([
                fits.Column(name='E_MIN', format='D', unit='MeV', array=self.emins),
                fits.Column(name='E_MAX', format='D', unit='MeV', array=self.emaxs)
                ], name='EBOUNDS')
        hdulist = fits.HDUList([hdu0, hdu1])
        for srcName in list(self.SrcDict.keys()):
            header = self.ccube.header.copy()
            header['EXTNAME'] = srcName
            hdu = fits.ImageHDU(data=self.srcmaps[srcName].T, header=header)
            hdulist.append(hdu)
        hdulist.writeto(outfile, overwrite=True)

    def SaveMdlCube(self, outfile):
        '''
        Save the mdlcube as outfile.
        '''
        assert isinstance(outfile, str)

        self.ccube.header['EXTNAME'] = 'SKYMAP'
        hdu0 = fits.PrimaryHDU(self.ccube.data.T, self.ccube.header)
        hdu1 = fits.BinTableHDU.from_columns([
                fits.Column(name='E_MIN', format='D', unit='MeV', array=self.emins),
                fits.Column(name='E_MAX', format='D', unit='MeV', array=self.emaxs)
                ], name='EBOUNDS')
        hdulist = fits.HDUList([hdu0, hdu1])
        mdlcube = np.zeros_like(self.N, dtype='float')
        for srcName in list(self.SrcDict.keys()):
            header = self.ccube.header
            header['EXTNAME'] = srcName
            hdu = fits.ImageHDU(data=self.mdlcubes[srcName].T, header=header)
            hdulist.append(hdu)

            mdlcube += self.mdlcubes[srcName]
        header['EXTNAME'] = 'AllSource'
        hdu = fits.ImageHDU(data=mdlcube.T, header=header)
        hdulist.append(hdu)
        hdulist.writeto(outfile, overwrite=True)

    def makeFuncCode(self):
        self.__code__ = FuncCode()
        self.__code__.co_varnames=sorted(self.FreeParsDict.keys())
        self.__code__.co_argcount=len(self.FreeParsDict)

        self.kwd = {}
        for freePar in self.__code__.co_varnames:
            parameter = self.FreeParsDict[freePar]
            self.kwd[freePar] = parameter.value
            self.kwd['limit_' + freePar] = (parameter.min, parameter.max)
            try:
                self.kwd['error_' + freePar] = parameter.error
            except:
                self.kwd['error_' + freePar] = 0.1 * parameter.value

    def lnprior(self, args):
        for i, freePar in enumerate(self.__code__.co_varnames):
            if args[i] < self.kwd['limit_' + freePar][0] or args[i] > self.kwd['limit_' + freePar][1]:
                return -np.inf
        return 0.0

    def lnprob(self, args):
        lnprior = self.lnprior(args)
        if not np.isfinite(lnprior):
            return -np.inf
        lnlike = -1 * self.__call__(*args)
        return lnprior + lnlike

    def mcmc_pre(self):
        self.Fixmdlcube = self.FixModelCube()
        self.makeFuncCode()
        ndim, nwalkers, nstep = self.__code__.co_argcount, 10*self.__code__.co_argcount, 1000
        pos = np.array([[self.kwd['limit_'+freePar][0] + (self.kwd['limit_'+freePar][1] - self.kwd['limit_'+freePar][0]) * np.random.rand() for freePar in self.__code__.co_varnames] for i in range(nwalkers)])
        pool = Pool(10)
        sampler = emcee.EnsembleSampler(nwalkers, ndim, self.lnprob, pool=pool)
        sampler.run_mcmc(pos, nstep, progress=True)
        pool.close()
        burnin = round(nstep / 10)
        samples = sampler.chain[:, burnin:, :].reshape((-1, ndim))
        values = np.percentile(samples, 50, axis=0)
        for freePar, value in zip(self.__code__.co_varnames, values):
            self.kwd[freePar] = value

        fig = corner.corner(samples)
        fig.savefig('corner_pre.png')

    def mcmc_post(self):
        self.Fixmdlcube = self.FixModelCube()
        self.makeFuncCode()
        ndim, nwalkers, nstep = self.__code__.co_argcount, 2*self.__code__.co_argcount, 1000
        pos = [np.array([self.kwd[freePar] for freePar in self.__code__.co_varnames]) + 1e-4 * np.random.randn(ndim) for i in range(nwalkers)]
        pool = Pool(nwalkers)
        sampler = emcee.EnsembleSampler(nwalkers, ndim, self.lnprob, pool=pool)
        sampler.run_mcmc(pos, nstep, progress=True)
        pool.close()
        burnin = round(nstep / 10)
        samples = sampler.chain[:, burnin:, :].reshape((-1, ndim))

        fig = corner.corner(samples)
        fig.savefig('corner_post.png')

    def minuit(self, hesse=False, minos=False, tol=None):
        '''
        This method execute the minuit fit
        with the counts cube and model counts cube
        call the minus likelihood function and minimize it
        '''
        self.Fixmdlcube = self.FixModelCube()
        self.makeFuncCode()

        if self.kwd == {}:
            likelihood = self.__call__()
            return {}, {}, {}, likelihood

        print('Minuit Starting......')
        if tol:
            m = Minuit(self, print_level=1, **self.kwd)
            m.tol = tol
        else:
            m = Minuit(self, print_level=1, **self.kwd)
        print('Migrad Start......')
        m.migrad()
        # if not m.valid:
        #     print('Migrad is not valid!')
        results = m.values
        errors = m.errors
        covs = m.covariance
        likelihood = m.fval

        if hesse:
            try:
                print('Hesse Start......')
                m.hesse()
                # if not m.accurate:
                #     print('Hesse is not accurate!')
                results = m.values
                errors = m.errors
                covs = m.covariance
                likelihood = m.fval
            except:
                pass

        if minos:
            try:
                print('Minos Start......')
                for freePar in self.func_code.co_varnames:
                    m.minos(freePar)
                    if not m.merrors[freePar].is_valid:
                        print(f'Minos of {freePar} is not valid!')
                results = m.values
                errors = m.errors
                covs = m.covariance
                likelihood = m.fval
            except:
                pass

        mdlcube = 0
        if self.ccube.type == 'Map':
            for srcName in self.SrcDict.keys():
                srcmdlcube = self.mdlcubes[srcName].sum(axis=(0, 1))
                #print(srcName, srcmdlcube, srcmdlcube.sum())
                mdlcube += srcmdlcube
            # print('mdlcube:', mdlcube, mdlcube.sum())
            # print('ccube:', self.N.sum(axis=(0, 1)), self.N.sum())
        if self.ccube.type == 'Hpx':
            for srcName in self.SrcDict.keys():
                srcmdlcube = self.mdlcubes[srcName].sum(axis=0)
                #print(srcName, srcmdlcube, srcmdlcube.sum())
                mdlcube += srcmdlcube
            # print('mdlcube:', mdlcube, mdlcube.sum())
            # print('ccube:', self.N.sum(axis=0), self.N.sum())
        # print('likelihood = ', likelihood)
        return results, errors, covs, likelihood
