'''
File: Model.py
Author: Duan Kai-Kai
Email: duankk@pmo.ac.cn
'''

import os
import yaml
import xml.etree.ElementTree as ET
from .Source import Source


'''example
    >>> from Model import Model

    >>> srcmdl = 'GC_model.yaml'
    >>> model = Model(srcmdl)
    >>> print model.modelDict, model.SrcDict, model.SrcList, model.FixSrcList, model.FreeSrcList

    >>> model.GetSpectralType(srcname)
    >>> model.GetSpecFunc(srcname)
    >>> model.GetSpatialType(srcname)
    >>> model.GetSpatMap(srcname)

    >>> model.ChangeParValue(srcName, parName, parValue, parError)

    >>> model.DelSource(srcname)
    >>> model.AddSource(srcname, SrcDict)
    >>> model.ChangeSrcSpec(srcName, SpecDict)
    >>> model.ChangeSrcSpat(srcName, SpatDict)
    >>> model.SaveYaml(outfile)
'''

class Model:
    def __init__(self, filename):
        if not os.path.exists(filename):
            raise FileNotFoundError(f'The model file "{filename}" not be found, please check it.')

        self.filename = filename.split('.')[0]
        if '.xml' in filename:
            tree = ET.parse(filename)
            root = tree.getroot()

            self.modelDict = {}
            for SrcTree in root:
                srcinfo = self.__xml2dict(SrcTree)
                self.modelDict[SrcTree.get('name')] = srcinfo

            with open(self.filename + '.yaml', 'w') as outfile:
                yaml.dump(self.modelDict, outfile)

        if '.yaml' in filename:
            self.modelDict = yaml.load(open(filename, 'r'), Loader=yaml.FullLoader)

        if isinstance(filename, dict):
            self.modelDict = filename

        self.__update__()
        
    def __update__(self):
        self.SrcDict = {}
        self.SrcList = []
        self.FixSrcDict = {}
        self.FixSrcList = []
        self.FreeSrcDict = {}
        self.FreeSrcList = []
        for srcName, SrcDict in list(self.modelDict.items()):
            self.SrcList.append(srcName)
            SourceObj = Source(SrcDict)
            self.SrcDict[srcName] = SourceObj
            if SourceObj.FreeParsDict:
                self.FreeSrcList.append(srcName)
                self.FreeSrcDict[srcName] = SourceObj
            else:
                self.FixSrcList.append(srcName)
                self.FixSrcDict[srcName] = SourceObj

    def __str__(self):
        return str(self.modelDict)

    def __repr(self):
        return str(self.modelDict)

    def __xml2dict(self, source):
        srcinfo = dict(source.attrib)
        for Spectrum in source.findall('spectrum'):
            spectrum = dict(Spectrum.attrib)

            for parameter in Spectrum:
                parainfo = dict(parameter.attrib)
                spectrum[parameter.get('name')] = parainfo
        srcinfo['spectrum'] = spectrum

        for SpatialModel in source.findall('spatialModel'):
            spatialModel = dict(SpatialModel.attrib)

            for parameter in SpatialModel:
                parainfo = dict(parameter.attrib)
                spatialModel[parameter.get('name')] = parainfo
        srcinfo['spatialModel'] = spatialModel

        return srcinfo


    def AddSource(self, srcName, srcDict):
        self.modelDict[srcName] = srcDict
        self.__update__()

    def DelSource(self, srcName):
        del self.modelDict[srcName]
        self.__update__()

    def ChangeSrcSpec(self, srcName, SpecDict):
        self.modelDict[srcName]['spectrum'] = SpecDict
        self.__update__()

    def ChangeSrcSpat(self, srcName, SpatDict):
        self.modelDict[srcName]['spatialModel'] = SpatDist
        self.__update__()

    def ChangeParValue(self, srcName, parName, parValue, parError=None):
        try:
            self.modelDict[srcName]['spectrum'][parName]['value'] = parValue
        except KeyError:
            self.modelDict[srcName]['spatialModel'][parName]['value'] = parValue

        if parError:
            try:
                self.modelDict[srcName]['spectrum'][parName]['error'] = parError
            except KeyError:
               self.modelDict[srcName]['spatialModel'][parName]['error'] = parError
        self.__update__()

    def SaveSrcModel(self, srcName, outfile):
        srcModelDict = self.modelDict[srcName]
        with open(outfile, 'w') as f:
            yaml.dump(srcModelDict, f)

    def SaveSpecModel(self, srcName, outfile):
        SpecModelDict = self.modelDict[srcName]['spectrum']
        with open(outfile, 'w') as f:
            yaml.dump(SpecModelDict, f)

    def SaveSpatModel(self, srcName, outfile):
        SpatModelDict = self.modelDict[srcName]['spatialModel']
        with open(outfile, 'w') as f:
            yaml.dump(SpatModelDict, f)

    def SaveModel(self, outfile):
        with open(outfile, 'w') as f:
            yaml.dump(self.modelDict, f)
