'''
File: SkyMap.py
Author: Duan Kai-Kai
Email: duankk@pmo.ac.cn
'''

import os
import numpy as np
import healpy as hp
import astropy.units as u
from astropy.io import fits
from astropy.wcs import WCS
from astropy.wcs import utils
from astropy.coordinates import SkyCoord
from . import maptools


class Cmap(object):
    '''
    This class is the counts map class,
    stores infomation about the map include wcs, pixcrd, skycrd, proj, coord,
    and allow user get larger, smaller, higher, lowwer map based on the origin map.
    '''
    def __init__(self, filename, hdu=0):
        assert isinstance(filename, str)
        assert isinstance(hdu, int)
        if not os.path.exists(filename):
            raise FileNotFoundError(f'The infile {filename} not be found, please check it.')

        self.dirname = os.path.dirname(filename)
        self.basename = os.path.basename(filename)
        self.filename = self.basename.split('.')[0]

        hdulist = fits.open(filename)
        self.data = hdulist[hdu].data.T
        self.header = hdulist[hdu].header
        hdulist.close()

        self.GetInfo()
        self.type = 'Map'

    def GetInfo(self):
        self.naxis = self.header['NAXIS']
        self.nxpix = self.header['NAXIS1']
        self.nypix = self.header['NAXIS2']
        self.npix  = self.nxpix * self.nypix
        self.deltax = self.header['CDELT1']
        self.deltay = self.header['CDELT2']

        self.wcs = self.Getwcs(self.header)
        self.proj = self.Getproj
        self.coord = self.Getcoord

        self.pixcrd = self.Getpixcrd
        self.pixcrd0 = self.Getpixcrd0

        self.skycrd, self.skycrd_C, self.skycrd_G = self.Getskycrd
        self.skycrd0, self.skycrd0_C, self.skycrd0_G = self.Getskycrd0

        self.dOmega = self.GetdOmega
        self.dOmega0 = self.GetdOmega0

    def Getwcs(self, header):
        new_header = header.copy()
        new_header['NAXIS'] = 2
        new_header['WCSAXES'] = 2
        for key in list(new_header.keys()):
            if '3' in key or '4' in key:
                del new_header[key]
        wcs = WCS(new_header)
        return wcs

    @property
    def Getpixcrd(self):
        pixcrd = np.mgrid[0:self.nxpix, 0:self.nypix].reshape(2, self.npix).T
        return pixcrd

    @property
    def Getpixcrd0(self):
        self.xrpix = self.header['CRPIX1']
        self.yrpix = self.header['CRPIX2']
        pixcrd0 = np.array([self.xrpix, self.yrpix])
        return pixcrd0

    @property
    def Getproj(self):
        return self.wcs.wcs.ctype[0].split('-')[-1]

    @property
    def Getcoord(self):
        if self.wcs.wcs.ctype[0].split('-')[0] == 'GLON'\
        or self.wcs.wcs.ctype[1].split('-')[0] == 'GLAT':
            coord = 'G'
        if self.wcs.wcs.ctype[0].split('-')[0] == 'RA'\
        or self.wcs.wcs.ctype[1].split('-')[0] == 'DEC':
            coord = 'C'
        return coord

    @property
    def Getskycrd(self):
        skycrd = self.wcs.wcs_pix2world(self.pixcrd, 0)
        if self.coord == 'G':
            skycrd_G = skycrd
            skycrd_C = maptools.skycrdG2C(skycrd_G)
        if self.coord == 'C':
            skycrd_C = skycrd
            skycrd_G = maptools.skycrdC2G(skycrd_C)
        skycrd = SkyCoord(ra=skycrd_C[:, 0]*u.degree, dec=skycrd_C[:, 1]*u.degree, frame='fk5')
        return skycrd, skycrd_C, skycrd_G

    @property
    def Getskycrd0(self):
        xrval = self.header['CRVAL1']
        yrval = self.header['CRVAL2']
        skycrd0 = np.array([xrval, yrval])
        if self.coord == 'G':
            skycrd0_G = skycrd0
            skycrd0_C = maptools.skycrdG2C(skycrd0_G)
            skycrd0 = SkyCoord(l=xrval*u.degree, b=yrval*u.degree, frame='galactic')
        if self.coord == 'C':
            skycrd0_C = skycrd0
            skycrd0_G = maptools.skycrdC2G(skycrd0_C)
            skycrd0 = SkyCoord(ra=xrval*u.degree, dec=yrval*u.degree, frame='fk5')
        return skycrd0, skycrd0_C, skycrd0_G

    @property
    def GetdOmega(self):
        yrval = self.header['CRVAL2']
        if self.coord == 'G':
            cosb = np.cos(np.deg2rad(self.skycrd_G[:, 1] - yrval))
        if self.coord == 'C':
            cosb = np.cos(np.deg2rad(self.skycrd_C[:, 1] - yrval))
        dOmega = abs(np.deg2rad(self.deltax) * np.deg2rad(self.deltay) * cosb)
        return dOmega.reshape(self.nxpix, self.nypix)

    @property
    def GetdOmega0(self):
        return utils.proj_plane_pixel_area(self.wcs) * (np.pi / 180.)**2

    def ChangeXY(self, x1=0, x2=0, y1=0, y2=0):
        new_header = self.header.copy()
        new_header['NAXIS'] = 2
        new_header['WCSAXES'] = 2
        new_header['NAXIS1'] = self.nxpix + x1 + x2
        new_header['NAXIS2'] = self.nypix + y1 + y2
        new_header['CRPIX1'] = new_header['NAXIS1'] / 2. + 0.5
        new_header['CRPIX2'] = new_header['NAXIS2'] / 2. + 0.5
        new_header['CRVAL1'] = self.header['CRVAL1'] + self.deltax * x1
        new_header['CRVAL2'] = self.header['CRVAL2'] + self.deltay * y1
        new_header['HISTORY'] = 'Change ' + self.filename + ' {},{},{},{}'.format(x1, x2, y1, y2)
        new_wcs = self.Getwcs(new_header)

        if self.naxis == 2:
            new_data = np.zeros((self.nxpix + x1 + x2, self.nypix + y1 + y2))
        if self.naxis == 3:
            new_data = np.zeros((self.nxpix + x1 + x2, self.nypix + y1 + y2, self.nzpix))

        if x1 >= 0 and x2 >= 0 and y1 >= 0 and y2 >= 0:
            new_data[x1 : self.nxpix + x1, y1 : self.nypix + y1] = self.data

        if x1 < 0 and x2 >= 0 and y1 >= 0 and y2 >= 0:
            new_data[0 : self.nxpix + x1, y1 : self.nypix + y1] = self.data[abs(x1):, :]
        if x1 >= 0 and x2 >= 0 and y1 < 0 and y2 >= 0:
            new_data[x1 : self.nxpix + x1, 0 : self.nypix + y1] = self.data[:, abs(y1):]
        if x1 >= 0 and x2 < 0 and y1 >= 0 and y2 >= 0:
            new_data[x1:, y1 : self.nypix + y1] = self.data[:self.nxpix-abs(x2), :]
        if x1 >= 0 and x2 >= 0 and y1 >= 0 and y2 < 0:
            new_data[x1 : self.nxpix + x1, y1:] = self.data[:, :self.nypix-abs(y2)]

        #new_filename = self.filename + '_changeXY.fits'
        new_filename = os.path.join(self.dirname, self.filename) + '_changeXY.fits'
        if self.naxis == 2:
            return NewCmap(new_filename, new_data, new_wcs)
        if self.naxis == 3:
            try:
                return NewCcube(new_filename, new_data, new_wcs, self.emins, self.emaxs)
            except:
                return NewSpatialCube(new_filename, new_data, new_wcs, self.ebounds)

    def get_larger_map(self, nx=3, ny=3):
        assert isinstance(nx, int)
        assert isinstance(ny, int)

        new_header = self.header.copy()
        new_header['NAXIS'] = 2
        new_header['WCSAXES'] = 2
        new_header['NAXIS1'] = self.nxpix * nx
        new_header['NAXIS2'] = self.nypix * ny
        new_header['CRPIX1'] = new_header['NAXIS1'] / 2. + 0.5
        new_header['CRPIX2'] = new_header['NAXIS2'] / 2. + 0.5
        new_header['HISTORY'] = 'larger ' + self.filename + '{0}*{1}-times'.format(nx, ny)
        new_wcs = self.Getwcs(new_header)

        if self.naxis == 2:
            new_data = np.zeros((self.nxpix * nx, self.nypix * ny))
        if self.naxis == 3:
            new_data = np.zeros((self.nxpix * nx, self.nypix * ny, self.nzpix))

        new_data[round(nx/2. - 0.5) * self.nxpix : round(nx/2. + 0.5) * self.nxpix,\
                 round(ny/2. - 0.5) * self.nypix : round(ny/2. + 0.5) * self.nypix] = self.data

        #new_filename = self.filename + '_larger.fits'
        new_filename = os.path.join(self.dirname, self.filename) + '_larger.fits'
        if self.naxis == 2:
            return NewCmap(new_filename, new_data, new_wcs)
        if self.naxis == 3:
            if hasattr(self, 'emins') and hasattr(self, 'emaxs'):
                return NewCcube(new_filename, new_data, new_wcs, self.emins, self.emaxs)
            elif hasattr(self, 'ebounds'):
                return NewSpatialCube(new_filename, new_data, new_wcs, self.ebounds)
            else:
                return NewCmap(new_filename, new_data, new_wcs)

    def get_smaller_map(self, nx=3, ny=3):
        assert isinstance(nx, int)
        assert isinstance(ny, int)

        if self.nxpix % nx != 0 or self.nypix % ny != 0:
            raise Exception('{0} cannot divided by {2} or {1} cannot divided by {3},\
                             You Cannot get the {2} or {3} times lowwer map.'\
                            .format(self.nxpix, self.nypix, nx, ny))

        new_header = self.header.copy()
        new_header['NAXIS'] = 2
        new_header['NAXIS1'] = self.nxpix / nx
        new_header['NAXIS2'] = self.nypix / ny
        new_header['CRPIX1'] = new_header['NAXIS1'] / 2. + 0.5
        new_header['CRPIX2'] = new_header['NAXIS2'] / 2. + 0.5
        new_header['HISTORY'] = 'smaller ' + self.filename + '{0}*{1}-times'.format(nx, ny)
        new_wcs = self.Getwcs(new_header)

        if self.naxis == 2:
            new_data = np.zeros((int(self.nxpix/nx), int(self.nypix/ny)))
        if self.naxis == 3:
            new_data = np.zeros((int(self.nxpix/nx), int(self.nypix/ny), self.nzpix))

        new_data = self.data[round(nx/2.-0.5) * int(self.nxpix/nx) : round(nx/2. + 0.5) * int(self.nxpix / nx),\
                             round(ny/2.-0.5) * int(self.nypix/ny) : round(ny/2. + 0.5) * int(self.nypix / ny)]

        #new_filename = self.filename + '_smaller.fits'
        new_filename = os.path.join(self.dirname, self.filename) + '_smaller.fits'
        if self.naxis == 2:
            return NewCmap(new_filename, new_data, new_wcs)
        if self.naxis == 3:
            if hasattr(self, 'emins') and hasattr(self, 'emaxs'):
                return NewCcube(new_filename, new_data, new_wcs, self.emins, self.emaxs)
            elif hasattr(self, 'ebounds'):
                return NewSpatialCube(new_filename, new_data, new_wcs, self.ebounds)
            else:
                return NewCmap(new_filename, new_data, new_wcs)

    def get_higher_map(self, nx=2, ny=2, area_depended=False):
        assert isinstance(nx, int)
        assert isinstance(ny, int)
        assert isinstance(area_depended, bool)

        new_header = self.header.copy()
        new_header['NAXIS'] = 2
        new_header['WCSAXES'] = 2
        new_header['NAXIS1'] = self.nxpix * nx
        new_header['NAXIS2'] = self.nypix * ny
        new_header['CDELT1'] = self.deltax / nx
        new_header['CDELT2'] = self.deltay / ny
        new_header['CRPIX1'] = new_header['NAXIS1'] / 2. + 0.5
        new_header['CRPIX2'] = new_header['NAXIS2'] / 2. + 0.5
        new_header['HISTORY'] = 'higher ' + self.filename + '{0}*{1}-times'.format(nx, ny)
        new_wcs = self.Getwcs(new_header)

        if self.naxis == 2:
            new_data = np.zeros((self.nxpix * nx, self.nypix * ny))
        if self.naxis == 3:
            new_data = np.zeros((self.nxpix * nx, self.nypix * ny, self.nzpix))

        for a in range(nx):
            for b in range(ny):
                new_data[a::nx, b::ny] = self.data

        if area_depended:
            new_data = new_data / (nx * ny)

        #new_filename = self.filename + '_higher.fits'
        new_filename = os.path.join(self.dirname, self.filename) + '_higher.fits'
        if self.naxis == 2:
            return NewCmap(new_filename, new_data, new_wcs)
        if self.naxis == 3:
            if hasattr(self, 'emins') and hasattr(self, 'emaxs'):
                return NewCcube(new_filename, new_data, new_wcs, self.emins, self.emaxs)
            elif hasattr(self, 'ebounds'):
                return NewSpatialCube(new_filename, new_data, new_wcs, self.ebounds)
            else:
                return NewCmap(new_filename, new_data, new_wcs)

    def get_lowwer_map(self, nx=2, ny=2, area_depended=False):
        assert isinstance(nx, int)
        assert isinstance(ny, int)
        assert isinstance(area_depended, bool)

        if self.nxpix % nx != 0 or self.nypix % ny != 0:
            raise Exception('{0} cannot divided by {2} or {1} cannot divided by {3},\
                             You Cannot get the {2} or {3} times lowwer map.'\
                            .format(self.nxpix, self.nypix, nx, ny))

        new_header = self.header.copy()
        new_header['NAXIS'] = 2
        new_header['WCSAXES'] = 2
        new_header['NAXIS1'] = self.nxpix / nx
        new_header['NAXIS2'] = self.nypix / ny
        new_header['CDELT1'] = self.deltax * nx
        new_header['CDELT2'] = self.deltay * ny
        new_header['CRPIX1'] = new_header['NAXIS1'] / 2. + 0.5
        new_header['CRPIX2'] = new_header['NAXIS2'] / 2. + 0.5
        new_header['HISTORY'] = 'lowwer ' + self.filename + '{0}*{1}-times'.format(nx, ny)
        new_wcs = self.Getwcs(new_header)

        if self.naxis == 2:
            new_data = np.zeros((self.nxpix / nx, self.nypix / ny))
        if self.naxis == 3:
            new_data = np.zeros((self.nxpix / nx, self.nypix / ny, self.nzpix))

        for a in range(nx):
            for b in range(ny):
                new_data += self.data[a::nx, b::ny] / (nx * ny)

        if area_depended:
            new_data = new_data * (nx * ny)

        #new_filename = self.filename + '_lowwer.fits'
        new_filename = os.path.join(self.dirname, self.filename) + '_lowwer.fits'
        if self.naxis == 2:
            return NewCmap(new_filename, new_data, new_wcs)
        if self.naxis == 3:
            if hasattr(self, 'emins') and hasattr(self, 'emaxs'):
                return NewCcube(new_filename, new_data, new_wcs, self.emins, self.emaxs)
            elif hasattr(self, 'ebounds'):
                return NewSpatialCube(new_filename, new_data, new_wcs, self.ebounds)
            else:
                return NewCmap(new_filename, new_data, new_wcs)


class SpatialMap(Cmap):
    '''
    This class is the spatial map class,
    stores some useful infomation like cmap class.
    '''
    def __init__(self, filename):
        super(SpatialMap, self).__init__(filename)


class Ccube(Cmap):
    '''
    This class is the counts cube class,
    stores the useful information like cmap class,
    and the more information about the energy bounds
    '''
    def __init__(self, filename, hdu=0):
        assert isinstance(filename, str)
        assert isinstance(hdu, int)
        if not os.path.exists(filename):
            raise FileNotFoundError(f'The infile {filename} not be found, please check it.')

        self.dirname = os.path.dirname(filename)
        self.basename = os.path.basename(filename)
        self.filename = self.basename.split('.')[0]

        hdulist = fits.open(filename)
        self.data = hdulist[hdu].data.T
        self.header = hdulist[hdu].header

        try:
            self.emins = hdulist['EBOUNDS'].data.field('E_MIN')
            self.emaxs = hdulist['EBOUNDS'].data.field('E_MAX')
        except:
            self.emins = None
            self.emaxs = None
        hdulist.close()

        self.nzpix = self.header['NAXIS3']
        self.GetInfo()
        self.type = 'Map'
        self.dOmega = self.dOmega.reshape(self.nxpix, self.nypix)


class SpatialCube(Ccube):
    '''
    This class is the spatial cube class,
    stores the useful information like cmap class,
    and the more information about the energy bounds
    '''
    def __init__(self, filename, hdu=0):
        assert isinstance(filename, str)
        assert isinstance(hdu, int)
        if not os.path.exists(filename):
            raise FileNotFoundError(f'The infile {filename} not be found, please check it.')

        self.dirname = os.path.dirname(filename)
        self.basename = os.path.basename(filename)
        self.filename = self.basename.split('.')[0]

        hdulist = fits.open(filename)
        self.data = hdulist[hdu].data.T
        self.header = hdulist[hdu].header

        try:
            self.ebounds = hdulist[hdu+1].data.field(0)
        except:
            self.ebounds = None
        hdulist.close()

        self.nzpix = self.header['NAXIS3']
        self.GetInfo()
        self.type = 'Map'


class HpxCmap:
    def __init__(self, filename, hdu=1):
        assert isinstance(filename, str)
        assert isinstance(hdu, int)
        if not os.path.exists(filename):
            raise FileNotFoundError(f'The infile {filename} not be found, please check it.')

        self.dirname = os.path.dirname(filename)
        self.basename = os.path.basename(filename)
        self.filename = self.basename.split('.')[0]

        hdulist = fits.open(filename)
        self.fdata = hdulist[hdu].data.T
        self.header = hdulist[hdu].header
        hdulist.close()

        self.nside = self.header['NSIDE']
        self.npix, self.order, self.resol, self.dOmega = self.GetInfo()

        self.ordering = self.header['ORDERING']
        self.coord = self.Getcoord
        self.skycrd, self.skycrd_C, self.skycrd_G = self.Getskycrd
        self.type = 'Hpx'

        self.data = self.fdata.field(0)
        #self.alm = self.GetAlm
        #self.cl = self.GetCl


    def GetInfo(self):
        npix = hp.nside2npix(self.nside)
        order = hp.nside2order(self.nside)
        resol = hp.nside2resol(self.nside)
        pixarea = hp.nside2pixarea(self.nside)
        return npix, order, resol, pixarea

    @property
    def Getcoord(self):
        try:
            coord = self.header['COORDSYS']
            if coord == 'GAL': coord = 'G' 
            if coord == 'CEL': coord = 'C'
        except:
            coord = ''
            print('Please set coord')
        return coord

    @property
    def Getskycrd(self):
        if self.coord == 'G':
            skycrd_G = maptools.ipix2skycrd(self.ordering, self.nside, list(range(self.npix)))
            skycrd_C = maptools.skycrdG2C(skycrd_G)
        if self.coord == 'C':
            skycrd_C = maptools.ipix2skycrd(self.ordering, self.nside, list(range(self.npix)))
            skycrd_G = maptools.skycrdC2G(skycrd_C)
        skycrd = SkyCoord(ra=skycrd_C[:, 0]*u.degree, dec=skycrd_C[:, 1]*u.degree, frame='fk5')
        return skycrd, skycrd_C, skycrd_G

    @property
    def GetAlm(self):
        if self.ordering == 'RING':
            alm = hp.map2alm(self.data)
        if self.ordering == 'NESTED':
            data = hp.nest2ring(self.data)
            alm = hp.map2alm(data)
        return alm

    @property
    def GetCl(self):
        if self.ordering == 'RING':
            cl = hp.anafast(self.data)
        if self.ordering == 'NESTED':
            data = hp.nest2ring(self.data)
            cl = hp.anafast(data)
        return cl

class HpxSpatialMap(HpxCmap):
    def __init__(self, filename):
        super(HpxSpatialMap, self).__init__(filename)


class HpxCcube(HpxCmap):
    def __init__(self, filename, hdu=1):
        assert isinstance(filename, str)
        assert isinstance(hdu, int)
        if not os.path.exists(filename):
            raise FileNotFoundError(f'The infile {filename} not be found, please check it.')

        self.dirname = os.path.dirname(filename)
        self.basename = os.path.basename(filename)
        self.filename = self.basename.split('.')[0]

        hdulist = fits.open(filename)
        self.fdata = hdulist[hdu].data
        self.header = hdulist[hdu].header
        try:
            self.emins = hdulist[hdu+1].data.field('E_MIN')
            self.emaxs = hdulist[hdu+1].data.field('E_MAX')
        except:
            self.emins = None
            self.emaxs = None
        hdulist.close()

        self.nside = self.header['NSIDE']
        self.npix, self.order, self.resol, self.dOmega = self.GetInfo()

        self.ordering = self.header['ORDERING']
        self.coord = self.Getcoord
        self.skycrd, self.skycrd_C, self.skycrd_G = self.Getskycrd
        self.type = 'Hpx'

        try:
            self.fields = self.header['TFIELDS']
            self.data = np.zeros((self.npix, self.fields))
            for i in range(self.fields):
                self.data[:, i] = self.fdata.field(i)
        except:
            self.fields = self.header['NAXIS2']
            self.data = self.fdata


class HpxSpatialCube(HpxCcube):
    def __init__(self, filename, hdu=1):
        assert isinstance(filename, str)
        assert isinstance(hdu, int)
        if not os.path.exists(filename):
            raise FileNotFoundError(f'The infile {filename} not be found, please check it.')

        self.dirname = os.path.dirname(filename)
        self.basename = os.path.basename(filename)
        self.filename = self.basename.split('.')[0]

        hdulist = fits.open(filename)
        self.fdata = hdulist[hdu].data
        self.header = hdulist[hdu].header
        try:
            self.ebounds = hdulist[hdu+1].data.field(0)
        except:
            self.ebounds = None
        hdulist.close()

        self.nside = self.header['NSIDE']
        self.npix, self.order, self.resol, self.dOmega = self.GetInfo()

        self.ordering = self.header['ORDERING']
        self.coord = self.Getcoord
        self.skycrd, self.skycrd_C, self.skycrd_G = self.Getskycrd

        self.fields = self.header['TFIELDS']
        if self.fields > 1:
            self.data = np.zeros((self.npix, self.fields))
            for i in range(self.fields):
                self.data[:, i] = self.fdata.field(i)
        else:
            self.data = self.fdata.field(0)
        

class MaskedHpxCMap(HpxCmap):
    def __init__(self, hpxmapFile, maskFile, hdu=1):
        assert isinstance(hpxmapFile, str)
        assert isinstance(maskFile, str)
        assert isinstance(hdu, int)

        if not os.path.exists(hpxmapFile):
            raise FileNotFoundError(f'The infile {hpxmapFile} not be found, please check it.')

        if not os.path.exists(maskFile):
            raise FileNotFoundError(f'The infile {maskFile} not be found, please check it.')

        self.dirname = os.path.dirname(hpxmapFile)
        self.basename = os.path.basename(hpxmapFile)
        self.filename = self.basename.split('.')[0]

        hdulist = fits.open(hpxmapFile)
        self.fdata = hdulist[hdu].data
        self.header = hdulist[hdu].header
        hdulist.close()

        self.nside = self.header['NSIDE']
        self.npix, self.order, self.resol, self.dOmega = self.GetInfo()

        self.ordering = self.header['ORDERING']
        self.coord = self.Getcoord
        self.skycrd, self.skycrd_C, self.skycrd_G = self.Getskycrd
        self.type = 'MaskedHpx'

        self.data = self.fdata.field(0)

        hdulist = fits.open(maskFile)
        self.mask = hdulist[hdu].data.field(0).astype('bool')

        self.skycrd_C = self.skycrd_C[self.mask]
        self.skycrd_G = self.skycrd_G[self.mask]
        self.data = self.data[self.mask]
        self.npix = sum(self.mask)


class MaskedHpxSpatialMap(MaskedHpxCMap):
    def __init__(self, hpxSpatialmapFile, maskFile):
        super(HpxSpatialMap, self).__init__(hpxSpatialmapFile, maskFile)


class MaskedHpxCcube(MaskedHpxCMap):
    def __init__(self, hpxCcubeFile, maskFile, hdu=1):
        assert isinstance(hpxCcubeFile, str)
        assert isinstance(maskFile, str)
        assert isinstance(hdu, int)

        if not os.path.exists(hpxCcubeFile):
            raise FileNotFoundError(f'The infile {hpxCcubeFile} not be found, please check it.')

        if not os.path.exists(maskFile):
            raise FileNotFoundError(f'The infile {maskFile} not be found, please check it.')

        self.dirname = os.path.dirname(hpxCcubeFile)
        self.basename = os.path.basename(hpxCcubeFile)
        self.filename = self.basename.split('.')[0]

        hdulist = fits.open(hpxCcubeFile)
        self.fdata = hdulist[hdu].data
        self.header = hdulist[hdu].header
        try:
            self.emins = hdulist[hdu+1].data.field('E_MIN')
            self.emaxs = hdulist[hdu+1].data.field('E_MAX')
        except:
            self.emins = None
            self.emaxs = None
        hdulist.close()

        self.nside = self.header['NSIDE']
        self.npix, self.order, self.resol, self.dOmega = self.GetInfo()

        self.ordering = self.header['ORDERING']
        self.coord = self.Getcoord
        self.skycrd, self.skycrd_C, self.skycrd_G = self.Getskycrd
        self.type = 'MaskedHpx'

        self.fields = self.header['TFIELDS']
        if self.fields > 1:
            self.data = np.zeros((self.npix, self.fields))
            for i in range(self.fields):
                self.data[:, i] = self.fdata.field(i)
        else:
            self.data = self.fdata.field(0)

        hdulist = fits.open(maskFile)
        self.mask = hdulist[hdu].data.field(0).astype('bool')

        self.skycrd_C = self.skycrd_C[self.mask]
        self.skycrd_G = self.skycrd_G[self.mask]
        self.data = self.data[self.mask]
        self.npix = sum(self.mask)


class MaskedHpxSpatialCube(MaskedHpxCcube):
    def __init__(self, hpxSpatialcubeFile, maskFile, hdu=1):
        assert isinstance(hpxSpatialcubeFile, str)
        assert isinstance(maskFile, str)
        assert isinstance(hdu, int)

        if not os.path.exists(hpxSpatialcubeFile):
            raise FileNotFoundError(f'The infile {hpxSpatialcubeFile} not be found, please check it.')

        if not os.path.exists(maskFile):
            raise FileNotFoundError(f'The infile {maskFile} not be found, please check it.')

        self.dirname = os.path.dirname(hpxSpatialcubeFile)
        self.basename = os.path.basename(hpxSpatialcubeFile)
        self.filename = self.basename.split('.')[0]

        hdulist = fits.open(hpxSpatialcubeFile)
        self.fdata = hdulist[hdu].data
        self.header = hdulist[hdu].header
        try:
            self.ebounds = hdulist[hdu+1].data.field('Energy')
        except:
            self.ebounds = None
        hdulist.close()

        self.nside = self.header['NSIDE']
        self.npix, self.order, self.resol, self.dOmega = self.GetInfo()

        self.ordering = self.header['ORDERING']
        self.coord = self.Getcoord
        self.skycrd, self.skycrd_C, self.skycrd_G = self.Getskycrd
        self.type = 'MaskedHpx'

        self.fields = self.header['TFIELDS']
        self.data = np.zeros((self.npix, self.fields))
        for i in range(self.fields):
            self.data[:, i] = self.fdata.field(i)
        
        hdulist = fits.open(maskFile)
        self.mask = hdulist[hdu].data.field(0).astype('bool')

        self.skycrd_C = self.skycrd_C[self.mask]
        self.skycrd_G = self.skycrd_G[self.mask]
        self.data = self.data[self.mask]
        self.npix = sum(self.mask)


class NewCmap(Cmap):
    def __init__(self, filename, data, wcs, gtidata=None, gtiheader=None):
        self.filename = filename
        self.dirname = os.path.dirname(filename)
        self.basename = os.path.basename(filename)
        self.data = data
        self.wcs = wcs
        self.gtidata = gtidata
        self.gtiheader = gtiheader

        self.GetHeader()
        self.GetInfo()
        self.type = 'Map'

    def WriteFits(self):
        fits.writeto(self.filename, self.data.T, self.header, overwrite=True)
        if self.gtidata is not None:
            fits.append(self.filename, self.gtidata, self.gtiheader, extname='GTI')

    def GetHeader(self):
        self.header = self.wcs.to_header()
        self.header['NAXIS'] = self.data.ndim
        self.header['NAXIS1'] = self.data.shape[0]
        self.header['NAXIS2'] = self.data.shape[1]


class NewSpatialMap(NewCmap):
    def __init__(self, filename, data, wcs):
        super(NewSpatialMap, self).__init__(filename, data, wcs)


class NewCcube(NewCmap):
    def __init__(self, filename, data, wcs, emins, emaxs, gtidata=None, gtiheader=None):
        self.filename = filename
        self.dirname = os.path.dirname(filename)
        self.basename = os.path.basename(filename)
        self.data = data
        self.wcs = wcs
        self.emins = emins
        self.emaxs = emaxs
        self.gtidata = gtidata
        self.gtiheader = gtiheader

        self.GetHeader()
        self.header['NAXIS3'] = data.shape[2]
        self.nzpix = self.header['NAXIS3']
        self.GetInfo()
        self.type = 'Map'
        self.dOmega = self.dOmega.reshape(self.nxpix, self.nypix)


    def WriteFits(self):
        fits.writeto(self.filename, self.data.T, self.header, overwrite=True)
        ebshdu = fits.BinTableHDU.from_columns([\
            fits.Column(name='E_MIN', format='E', unit='MeV', array=self.emins),
            fits.Column(name='E_MAX', format='E', unit='MeV', array=self.emaxs)],
            name='EBOUNDS')
        fits.append(self.filename, ebshdu.data, ebshdu.header, extname='EBOUNDS')
        if self.gtidata is not None:
            fits.append(self.filename, self.gtidata, self.gtiheader, extname='GTI')


class NewSpatialCube(NewCcube):
    def __init__(self, filename, data, wcs, ebounds):
        self.filename = filename
        self.dirname = os.path.dirname(filename)
        self.basename = os.path.basename(filename)
        self.data = data
        self.wcs = wcs
        self.ebounds = ebounds

        self.GetHeader()
        self.header['NAXIS3'] = data.shape[2]
        self.nzpix = self.header['NAXIS3']
        self.GetInfo()
        self.type = 'Map'

    def WriteFits(self):
        fits.writeto(self.filename, self.data.T, self.header, overwrite=True)
        ebshdu = fits.BinTableHDU.from_columns([\
            fits.Column(name='Energy', format='E', unit='MeV', array=self.ebounds)],
            name='EBOUNDS')
        fits.append(self.filename, ebshdu.data, ebshdu.header, extname='EBOUNDS')


class NewHpxmap(HpxCmap):
    def __init__(self, filename, data, nside, ordering, coord, gtidata=None, gtiheader=None):
        self.filename = filename
        self.dirname = os.path.dirname(filename)
        self.basename = os.path.basename(filename)
        self.data = data
        self.gtidata = gtidata
        self.gtiheader = gtiheader

        self.nside = nside
        self.npix, self.order, self.resol, self.dOmega = self.GetInfo()

        self.ordering = ordering
        self.coord = coord
        self.skycrd, self.skycrd_C, self.skycrd_G = self.Getskycrd
        self.type = 'Hpx'

    def WriteFits(self):
        hp.write_map(self.filename, self.data, nest=(self.ordering=='NESTED'), coord=self.coord, fits_IDL=False, overwrite=True)
        if self.gtidata is not None:
            fits.append(self.filename, self.gtidata, self.gtiheader)

class NewHpxcube(HpxCcube):
    def __init__(self, filename, data, nside, ordering, coord, emins, emaxs, gtidata=None, gtiheader=None):
        self.filename = filename
        self.dirname = os.path.dirname(filename)
        self.basename = os.path.basename(filename)
        self.data = data
        self.emins = emins
        self.emaxs = emaxs
        self.gtidata = gtidata
        self.gtiheader = gtiheader

        self.nside = nside
        self.npix, self.order, self.resol, self.dOmega = self.GetInfo()

        self.ordering = ordering
        self.coord = coord
        self.skycrd, self.skycrd_C, self.skycrd_G = self.Getskycrd
        self.type = 'Hpx'

    def WriteFits(self):
        hp.write_map(self.filename, self.data.T, nest=(self.ordering=='NESTED'), coord=self.coord, fits_IDL=False, overwrite=True)
        ebshdu = fits.BinTableHDU.from_columns([\
            fits.Column(name='E_MIN', format='E', unit='MeV', array=self.emins),\
            fits.Column(name='E_MAX', format='E', unit='MeV', array=self.emaxs)],\
            name='EBOUNDS')
        fits.append(self.filename, ebshdu.data, ebshdu.header, extname='EBOUNDS')
        if self.gtidata is not None:
            fits.append(self.filename, self.gtidata, self.gtiheader)


def ReadSrcMaps(srcmaps):
    srcmapsDict = {}
    hdulist = fits.open(srcmaps)
    for hdu in hdulist:
        if hdu.name in ['SKYMAP', 'EBOUNDS']:
            continue
        srcmapsDict[hdu.name] = hdu.data.T
    return srcmapsDict

def SaveSrcMaps(ccube, srcmapsDict, outfile):
    ccube.header['EXTNAME'] = 'SKYMAP'
    hdu0 = fits.PrimaryHDU(ccube.data.T, ccube.header)
    hdu1 = fits.BinTableHDU.from_columns([
            fits.Column(name='E_MIN', format='D', unit='MeV', array=ccube.emins),
            fits.Column(name='E_MAX', format='D', unit='MeV', array=ccube.emaxs)
            ], name='EBOUNDS')
    hdulist = fits.HDUList([hdu0, hdu1])

    for srcName, srcmap in list(srcmapsDict.items()):
        header = ccube.header.copy()
        header['EXTNAME'] = srcName
        hdu = fits.ImageHDU(data=srcmap.T, header=header)
        hdulist.append(hdu)
    hdulist.writeto(outfile, overwrite=True)

def ReadMaskedHpxSrcMaps(srcmaps, maskFile):
    srcmapsDict = {}

    hdulist = fits.open(maskFile)
    mask = hdulist[1].data.field(0).astype('bool')

    hdulist = fits.open(srcmaps)
    for hdu in hdulist:
        if hdu.name in ['SKYMAP', 'EBOUNDS']:
            continue
        srcmapsDict[hdu.name] = hdu.data.T[mask]

    return srcmapsDict


if __name__ == '__main__':
    filename = 'cmap.fits'
    cmap = Cmap(filename)
    NewCmap('NewCmap.fits', cmap.data, cmap.wcs)
    larger_cmap = cmap.get_larger_map()
    larger_cmap.get_smaller_map()
    higher_cmap = cmap.get_higher_map(area_depended=True)
    higher_cmap.get_lowwer_map(area_depended=True)

    filename = 'ccube.fits'
    ccube = Ccube(filename)
    NewCcube('NewCcube.fits', ccube.data, ccube.wcs)
