'''
File: Source.py
Author: Duan Kai-Kai
Email: duankk@pmo.ac.cn
'''

import numpy as np
from . import Spectrum
from . import SpatialModel
from scipy.integrate import simps
from scipy.interpolate import interp1d


'''example
    >>> from Source import Source

    >>> srcDict = {'name': '3FGL J0000.1+6545',
                   'type': 'PointSource',

           'spatialModel': {'type': 'SkyDirFunction',
                             'DEC': {'free': '0',
                                      'max': '90',
                                      'min': '-90',
                                     'name': 'DEC',
                                    'scale': '1',
                                    'value': '65.7517'
                                    },
                              'RA': {'free': '0',
                                      'max': '360',
                                      'min': '-360',
                                     'name': 'RA',
                                    'scale': '1',
                                    'value': '0.0377'
                                    }
                           },

         'spectrum': {'type': 'PowerLaw',
                   'normPar': 'Prefactor',
                     'Index': {'error': '0.0823208136',
                                'free': '1',
                                 'max': '-0',
                                 'min': '-5',
                                'name': 'Index',
                               'scale': '1',
                               'value': '-2.410955835'
                              },
                 'Prefactor': {'error': '0.07322381815',
                                'free': '1',
                                 'max': '100000',
                                 'min': '1e-05',
                                'name': 'Prefactor',
                               'scale': '1e-12',
                               'value': '0.4869737351'
                              },
                     'Scale': {'free': '0',
                                'max': '1000',
                                'min': '0.01',
                                'max': '1000',
                               'name': 'Scale',
                              'scale': '1000',
                              'value': '1.565626445'
                              }
                     }
            }

    >>> source = Source(srcDict)
    >>> print source.name, source.type, source.SpectralType, source.Spectrum, source.SpatialType, source.SpatialModel, source.ParsDict, source.FixParsDict, source.FreeParsDict, source.SpecFunc, self.Skycrd

    >>> parName = 'Index'
    >>> source.GetParFree(parName)
    >>> source.GetParScale(parName)
    >>> source.GetParValue(parName)
    >>> source.GetParError(parName)
    >>> source.GetParRange(parName)
    >>> source.GetScaledValue(parName)
    >>> source.GetScaledError(parName)
    >>> source.GetScaledRange(parName)

    >>> source.SetParFree(parName, new_free)
    >>> source.SetParScale(parName, new_scale)
    >>> source.SetParValue(parName, new_value)
    >>> source.SetParError(parName, new_error)
    >>> source.SetParRange(parName, new_min, new_max)
    >>> source.SetScaledValue(parName, new_value)
    >>> source.SetScaledError(parName, new_error)
    >>> source.SetScaledRange(parName, new_min, new_max)
'''


class Source:
    def __init__(self, srcDict):
        self.srcDict = srcDict
        self.name = self.srcDict['name']
        self.type = self.srcDict['type']
        ParsList = list(self.srcDict.keys())

        if 'TS_Value' in ParsList:
            self.TS_Value = self.srcDict['TS_Value']

        if 'Variability_Index' in ParsList:
            self.Variability_Index = self.srcDict['Variability_Index']

        self.SpectralType = self.srcDict['spectrum']['type']
        self.spectrum = eval('Spectrum.' + self.SpectralType)\
                            (self.srcDict['spectrum'])

        self.SpatialType = self.srcDict['spatialModel']['type']
        self.spatialModel = eval('SpatialModel.' + self.SpatialType)\
                                (self.srcDict['spatialModel'])

        self.ParsDict = dict(list(self.spectrum.ParsDict.items())
                           + list(self.spatialModel.ParsDict.items()))

        self.FixParsDict = self.GetFixPars()
        self.FreeParsDict = self.GetFreePars()
        self.ParsDict = dict(list(self.FixParsDict.items())
                           + list(self.FreeParsDict.items()))

        self.SpecFunc = self.spectrum.GetSpecFunc
        #self.Skycrd = self.spatialModel.GetSkycrd

    def __updateSpectrumPars__(self):
        self.SpectralType = self.srcDict['spectrum']['type']
        self.spectrum = eval('Spectrum.' + self.SpectralType)\
                            (self.srcDict['spectrum'])
        self.ParsDict = dict(list(self.spectrum.ParsDict.items())
                           + list(self.spatialModel.ParsDict.items()))

        self.FixParsDict = self.GetFixPars()
        self.FreeParsDict = self.GetFreePars()
        self.ParsDict = dict(list(self.FixParsDict.items())
                           + list(self.FreeParsDict.items()))

        self.SpecFunc = self.spectrum.GetSpecFunc
        self.Skycrd = self.spatialModel.GetSkycrd

    def __str__(self):
        return str(self.srcDict)

    def __repr__(self):
        return str(self.srcDict)

    def MdlCube(self, CubeType, srcmap, ebounds):
        '''
        Calculate the Mdlcube based on srcmap.
        '''
        assert isinstance(CubeType, str)
        assert isinstance(srcmap, np.ndarray)
        assert isinstance(ebounds, np.ndarray)

        if not CubeType in ['Map', 'Hpx', 'MaskedHpx']:
            raise RuntimeError('The CubeType must be one of Map or Hpx, please check it.')

        emins = ebounds[:-1]
        emaxs = ebounds[1:]
        dlnE = np.log(emaxs) - np.log(emins)

        if CubeType == 'Map':
            mdlcube = (srcmap[:, :, :-1] * self.SpecFunc(emins) * emins\
                      + srcmap[:, :, 1:] * self.SpecFunc(emaxs) * emaxs) * dlnE / 2
        if CubeType in ['Hpx', 'MaskedHpx']:
            mdlcube = (srcmap[:, :-1] * self.SpecFunc(emins) * emins\
                      + srcmap[:, 1:] * self.SpecFunc(emaxs) * emaxs) * dlnE / 2
        return mdlcube

    def CalculateFlux(self, emin, emax, ebins=100):
        '''
        Calculate flux from emin to emax.
        emin: MeV
        emax: MeV
        '''
        assert isinstance(emin, (int, float))
        assert isinstance(emax, (int, float))
        assert isinstance(ebins, int)

        if not emax > emin:
            raise RuntimeError('The emax must be larger than emin, please check it.')

        E = np.logspace(np.log10(emin), np.log10(emax), ebins)
        dnde = self.SpecFunc(E)
        lnE = np.log(E)
        if self.SpatialType == 'SkyDirFunction':
            Flux = simps(E * dnde, lnE)

        if self.SpatialType == 'ConstantValue': #isotropic model
            Flux = 4*np.pi * simps(E * dnde, lnE)

        if self.SpatialType in ['SpatialMap', 'HpxSpatialMap']: #extend source
            SpatialMap = self.spatialModel.GetSpatialMap()
            Flux = np.sum(SpatialMap.data * SpatialMap.dOmega) * simps(E * dnde, lnE)
        if self.SpatialType == 'MapCubeFunction':#galactic model
            SpatialCube = self.spatialModel.GetSpatialCube()
            Spatialmin = 10**(interp1d(np.log10(SpatialCube.ebounds), np.log10(SpatialCube.data), fill_value='extrapolate')(np.log10(emin)))
            Spatialmax = 10**(interp1d(np.log10(SpatialCube.ebounds), np.log10(SpatialCube.data), fill_value='extrapolate')(np.log10(emax)))

            E = SpatialCube.ebounds
            SpatialData = np.concatenate([Spatialmin[:, :, np.newaxis], SpatialCube.data[:, :, (E>emin) * (E<emax)], Spatialmax[:, :, np.newaxis]], axis=-1)
            E = np.hstack([emin, E[(E>emin) * (E<emax)], emax])
            dnde = self.SpecFunc(E)
            lnE = np.log(E)
            Flux = simps(np.sum(SpatialData.T * SpatialCube.dOmega.T, axis=(1, 2)) * E * dnde, lnE)

        if self.SpatialType == 'HpxSpatialCube':
            SpatialCube = self.spatialModel.GetSpatialCube()
            Spatialmin = 10**(interp1d(np.log10(SpatialCube.ebounds), np.log10(SpatialCube.data), fill_value='extrapolate')(np.log10(emin)))
            Spatialmax = 10**(interp1d(np.log10(SpatialCube.ebounds), np.log10(SpatialCube.data), fill_value='extrapolate')(np.log10(emax)))

            E = SpatialCube.ebounds
            SpatialData = np.concatenate([Spatialmin[:, np.newaxis], SpatialCube.data[:, (E>emin) * (E<emax)], Spatialmax[:, np.newaxis]], axis=-1)
            E = np.hstack([emin, E[(E>emin) * (E<emax)], emax])
            dnde = self.SpecFunc(E)
            lnE = np.log(E)
            Flux = simps(np.sum(SpatialData * SpatialCube.dOmega) * E * dnde, lnE)
        return Flux

    def CalculateEnergyFlux(self, emin, emax, ebins=100):
        '''
        Calculate energy flux from emin to emax.
        emin: MeV
        emax: MeV
        '''
        assert isinstance(emin, (int, float))
        assert isinstance(emax, (int, float))
        assert isinstance(ebins, int)

        if not emax > emin:
            raise RuntimeError('The emax must be larger than emin, please check it.')

        E = np.logspace(np.log10(emin), np.log10(emax), ebins)
        dnde = self.SpecFunc(E)
        lnE = np.log(E)
        if self.SpatialType == 'SkyDirFunction':
            EnergyFlux = simps(E**2 * dnde, lnE)

        if self.SpatialType == 'ConstantValue': #isotropic model
            EnergyFlux = 4*np.pi * simps(E**2 * dnde, lnE)

        if self.SpatialType in ['SpatialMap', 'HpxSpatialMap']: #extend source
            SpatialMap = self.spatialModel.GetSpatialMap
            EnergyFlux = np.sum(SpatialMap.data * SpatialMap.dOmega, axis=1) * simps(E**2 * dnde, lnE)

        if self.SpatialType == 'MapCubeFunction':#galactic model
            SpatialCube = self.spatialModel.GetSpatialCube
            from scipy.interpolate import interp1d
            Spatialmin = 10**(interp1d(np.log10(SpatialCube.ebounds), np.log10(SpatialCube.data), fill_value='extrapolate')(np.log10(emin)))
            Spatialmax = 10**(interp1d(np.log10(SpatialCube.ebounds), np.log10(SpatialCube.data), fill_value='extrapolate')(np.log10(emax)))

            E = SpatialCube.ebounds
            SpatialData = np.concatenate([Spatialmin[:, :, np.newaxis], SpatialCube.data[:, :, (E>emin) * (E<emax)], Spatialmax[:, :, np.newaxis]], axis=-1)
            E = np.hstack([emin, E[(E>emin) * (E<emax)], emax])
            dnde = self.SpecFunc(E)
            EnergyFlux = simps(np.sum(SpatialData.T * SpatialCube.dOmega.T, axis=(1, 2)) * E**2 * dnde, lnE)

        if self.SpatialType == 'HpxSpatialCube':
            SpatialCube = self.spatialModel.GetSpatialCube
            from scipy.interpolate import interp1d
            Spatialmin = 10**(interp1d(np.log10(SpatialCube.ebounds), np.log10(SpatialCube.data), fill_value='extrapolate')(np.log10(emin)))
            Spatialmax = 10**(interp1d(np.log10(SpatialCube.ebounds), np.log10(SpatialCube.data), fill_value='extrapolate')(np.log10(emax)))

            E = SpatialCube.ebounds
            SpatialData = np.concatenate([Spatialmin[:, np.newaxis], SpatialCube.data[:, (E>emin) * (E<emax)], Spatialmax[:, np.newaxis]], axis=-1)
            E = np.hstack([emin, E[(E>emin) * (E<emax)], emax])
            dnde = self.SpecFunc(E)
            EnergyFlux = simps(np.sum(SpatialData.T * SpatialCube.dOmega.T, axis=1) * E**2 * dnde, lnE)

        return EnergyFlux

    def GetFixPars(self):
        FixParsDict = {}
        for parName in list(self.ParsDict.keys()):
            if self.GetParFree(parName) == False:
                FixParsDict[parName] = self.ParsDict[parName]
        return FixParsDict

    def GetFreePars(self):
        FreeParsDict = {}
        for parName in list(self.ParsDict.keys()):
            if self.GetParFree(parName) == True:
                FreeParsDict[parName] = self.ParsDict[parName]
        return FreeParsDict

    def GetParFree(self, parName):
        return self.ParsDict[parName].free

    def GetParScale(self, parName):
        return self.ParsDict[parName].scale

    def GetParValue(self, parName):
        return self.ParsDict[parName].GetValue

    def GetParError(self, parName):
        return self.ParsDict[parName].GetError

    def GetParRange(self, parName):
        return self.ParsDict[parName].GetRange

    def GetScaledValue(self, parName):
        return self.ParsDict[parName].value

    def GetScaledError(self, parName):
        return self.ParsDict[parName].error

    def GetScaledRange(self, parName):
        scaledmin = self.ParsDict[parName].min
        scaledmax = self.ParsDict[parName].max
        return (scaledmin, scaledmax)

    def SetParFree(self, parName, new_free):
        self.ParsDict[parName].SetFree(new_free)
        try:
            self.srcDict['spectrum'][parName]['free'] = new_free
        except:
            self.srcDict['spatialModel'][parName]['free'] = new_free
        if new_free:
            self.FreeParsDict[parName] = self.FixParsDict.pop(parName)
        else:
            self.FixParsDict[parName] = self.FreeParsDict.pop(parName)

    def SetParScale(self, parName, new_scale):
        self.ParsDict[parName].SetScale(new_scale)
        try:
            self.srcDict['spectrum'][parName]['scale'] = new_scale
        except:
            self.srcDict['spatialModel'][parName]['scale'] = new_scale

    def SetParValue(self, parName, new_value):
        self.ParsDict[parName].SetValue(new_value)
        try:
            self.srcDict['spectrum'][parName]['value'] = new_value
        except:
            self.srcDict['spatialModel'][parName]['value'] = new_value
        self.SpecFunc = self.spectrum.GetSpecFunc

    def SetParError(self, parName, new_error):
        self.ParsDict[parName].SetError(new_error)
        try:
            self.srcDict['spectrum'][parName]['error'] = new_error
        except:
            self.srcDict['spatialModel'][parName]['error'] = new_error

    def SetParRange(self, parName, new_min, new_max):
        self.ParsDict[parName].SetRange(new_min, new_max)
        try:
            self.srcDict['spectrum'][parName]['min'] = new_min
            self.srcDict['spectrum'][parName]['max'] = new_max
        except:
            self.srcDict['spatialModel'][parName]['min'] = new_min
            self.srcDict['spatialModel'][parName]['max'] = new_max
