'''
File: SpaceCraft.py
Author: Duan Kai-Kai
Email: duankk@pmo.ac.cn
'''

import os
import numpy as np
from . import maptools
from . import DmpTime
from astropy.io import fits


class SpaceCraft:
    '''
    This class stores the spacecraft information including the time, position of satellite, coordinates of subsatellite point, zenith, scx, scy, scz and livetime for observation.
    scfile: str
    '''
    def __init__(self, scfile):
        if not os.path.exists(scfile):
            raise FileNotFoundError(f'The scfile "{scfile}" not be found, please check it.')

        hdulist = fits.open(scfile)
        hdu = hdulist['SC_DATA']
        self.header = hdu.header
        self.scdata = hdu.data

        self.SCmin = self.scdata.field('START')
        self.SCmax = self.scdata.field('STOP')
        self.position = self.scdata.field('SC_POSITION')
        self.velocity = self.scdata.field('SC_VELOCITY')

        lon_geo = self.scdata.field('LON_GEO')
        lat_geo = self.scdata.field('LAT_GEO')
        self.geocrd = np.vstack((lon_geo, lat_geo)).T

        ra_zenith = self.scdata.field('RA_ZENITH')
        dec_zenith = self.scdata.field('DEC_ZENITH')
        self.skycrd_zenith = np.vstack((ra_zenith, dec_zenith)).T

        ra_scx = self.scdata.field('RA_SCX')
        dec_scx = self.scdata.field('DEC_SCX')
        self.skycrd_scx = np.vstack((ra_scx, dec_scx)).T

        ra_scy = self.scdata.field('RA_SCY')
        dec_scy = self.scdata.field('DEC_SCY')
        self.skycrd_scy = np.vstack((ra_scy, dec_scy)).T

        ra_scz = self.scdata.field('RA_SCZ')
        dec_scz = self.scdata.field('DEC_SCZ')
        self.skycrd_scz = np.vstack((ra_scz, dec_scz)).T

        self.in_SAA = self.scdata.field('IN_SAA')
        self.livetime = self.scdata.field('LIVETIME') * (1 - self.in_SAA)

    def __repr__(self):
        num = len(self.scdata)
        METmin = min(self.SCmin)
        METmax = max(self.SCmax)
        UTCmin = DmpTime.MET2UTC(METmin)
        UTCmax = DmpTime.MET2UTC(METmax)
        livetime = sum(self.livetime)
        summary = f"""
        {num} records:
        MET: from {METmin} to {METmax}
        UTC: from {UTCmin} to {UTCmax}
        livetime = {livetime} s
        """
        return summary

    def ObservedTime(self, tmin, tmax):
        '''
        Calculate the livetime from tmin to tmax
        tmin: MET
        tmax: MET
        '''
        assert isinstance(tmin, (int, float))
        assert isinstance(tmax, (int, float))

        if not tmax > tmin:
            raise RuntimeError('The tmax must be larger than tmin, please check it.')

        tfilter = (self.SCmin > tmin) * (self.SCmax < tmax)
        livetime = self.livetime[tfilter].sum()
        return livetime

    def ObservedAngle(self, GorC, xref, yref, time):
        '''
        Calculate the observed angle for (GorC, xref, yref) at time
        GorC: ['G', 'C']
        xref: deg
        yref: deg
        time: MET
        '''
        assert isinstance(GorC, str)
        assert isinstance(xref, (int, float))
        assert isinstance(yref, (int, float))
        assert isinstance(time, (int, float))

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')
        if not 0 <= xref <= 360:
            raise RuntimeError('The xref must be in the range of from 0 to 360, please check it.')
        if not -90 <= yref <= 90:
            raise RuntimeError('The yref must be in the range of from -90 to 90, please check it.')

        if GorC == 'G':
            skycrd_G = (xref, yref)
            skycrd_C = maptools.skycrdG2C(skycrd_G)
        if GorC == 'C':
            skycrd_C = (xref, yref)
            skycrd_G = maptools.skycrdC2G(skycrd_C)

        skycrd_scz = self.skycrd_scz[(self.SCmin <= time) * (self.SCmax > time)]
        ObservedAngle = 180 - maptools.Sep(skycrd_C, skycrd_scz)
        return ObservedAngle

    def ObservedAngles(self, GorC, xref, yref, tmin, tmax):
        '''
        Calculate the observed angle for (GorC, xref, yref) from tmin to tmax.
        GorC: ['G', 'C']
        xref: deg
        yref: deg
        time: MET
        '''
        assert isinstance(GorC, str)
        assert isinstance(xref, (int, float))
        assert isinstance(yref, (int, float))
        assert isinstance(tmin, (int, float))
        assert isinstance(tmax, (int, float))

        if not GorC in ['G', 'C']:
            raise RuntimeError('The GorC must be one of G or C, please check it.')
        if not 0 <= xref <= 360:
            raise RuntimeError('The xref must be in the range of from 0 to 360, please check it.')
        if not -90 <= yref <= 90:
            raise RuntimeError('The yref must be in the range of from -90 to 90, please check it.')
        if not tmax > tmin:
            raise RuntimeError('The tmax must be larger than tmin, please check it.')
        
        if GorC == 'G':
            skycrd_G = (xref, yref)
            skycrd_C = maptools.skycrdG2C(skycrd_G)
        if GorC == 'C':
            skycrd_C = (xref, yref)
            skycrd_G = maptools.skycrdC2G(skycrd_C)

        tfilter = (self.SCmin > tmin) * (self.SCmax < tmax)
        ObservedAngles = 180 - maptools.Sep(skycrd_C, self.skycrd_scz[tfilter])
        return ObservedAngles


if __name__ == '__main__':
    scfile = 'sc.fits'
    sc = SpaceCraft(scfile)
    print(sc)