'''
File: SpatialModel.py
Author: Duan Kai-Kai
Email: duankk@pmo.ac.cn
'''

import numpy as np
from . import maptools
from . import SkyMap
import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.modeling import models
from .Parameter import Parameter

'''example
    >>> from SpatialModel import *
    >>> pars =  {'type'  : 'SkyDirFunction',
               'RA'  : {'free': '0', 'max': '360', 'min': '-360', 'name': 'RA', 'scale': '1', 'value': '0.0377'},
              'DEC'  : {'free': '0', 'max': '90', 'min': '-90', 'name': 'DEC', 'scale': '1', 'value': '65.7517'}
            }

    >>> SpatialModel = SkyDirFunction(pars)
    >>> print SpatialModel.type, SpatialModel.ParsDict
    >>> SpatialModel.GetSkycrd


    >>> pars =  {'file'  : '$(FERMI_EXTENDED)/RXJ1713.fits',
     'map_based_integral': 'true',
             'type'  : 'SpatialMap'
            }

    >>> SpatialModel = SpatialMap(par)
    >>> print SpatialModel.type, SpatialModel.ParsDict
    >>> SpatialModel.GetSpatialMap
'''


class SpatialModel:
    def __init__(self, pars):
        self.pars = pars
        self.type = self.pars['type']

        if 'file' in list(self.pars.keys()):
            self.file = self.pars['file']

        if 'map_based_integral' in list(self.pars.keys()):
            self.map_based_integral = self.pars['map_based_integral']

        self.ParsDict = {}
        for parName, parDict in list(self.pars.items()):
            try:
                self.ParsDict[parName] = Parameter(parDict)
            except TypeError:
                pass

        if 'Prefactor' in list(self.ParsDict.keys()):
            self.ParsDict.pop('Prefactor')

        if self.type == 'SkyDirFunction':
            self.skycrd = self.GetSkycrd()
        elif self.type in ['MapCubeFunction', 'HpxSpatialCube']:
            self.SpatialCube = self.GetSpatialCube()
        elif self.type in ['ConstantValue', 'SpatialMap', 'HpxSpatialMap', 'RadialDisk', 'RadialGaussian', 'Ring2D', 'Ellipse2D', 'Gaussian2D']:
            self.SpatialMap = self.GetSpatialMap()
        else:
            assert RunTimeError('The type of spatial model is not supported!')


    def __str__(self):
        return str(self.pars)

    def __repr__(self):
        return str(self.pars)

    def GetParValue(self, par):
        return self.ParsDict[par].GetValue


class SkyDirFunction(SpatialModel):
    def GetSkycrd(self):
        self.RA = self.GetParValue('RA')
        self.DEC = self.GetParValue('DEC')
        return np.array([self.RA, self.DEC])

    def SetSkycrd(self, new_skycrd):
        new_ra = new_skycrd[0]
        new_dec = new_skycrd[1]
        self.RA.SetParValue('RA', new_ra)
        self.DEC.SetParValue('DEC', new_dec)


class SpatialMap(SpatialModel):
    def GetSpatialMap(self):
        return SkyMap.SpatialMap(self.file)

    '''
    def GetSkycrd(self):
        SpatialMap = self.GetSpatialMap
        return SpatialMap.skycrd0_C
    '''

class HpxSpatialMap(SpatialModel):
    def GetSpatialMap(self):
        return SkyMap.HpxSpatialMap(self.file)


class MapCubeFunction(SpatialModel):
    def GetSpatialCube(self):
        return SkyMap.SpatialCube(self.file)

    def GetSkycrd(self):
        SpatialCube = self.GetSpatialCube
        return SpatialCube.skycrd0_C

class HpxSpatialCube(SpatialModel):
    def GetSpatialCube(self):
        return SkyMap.HpxSpatialCube(self.file)


class ConstantValue(SpatialModel):
    def GetSpatialMap(self):
        return 1


class RadialDisk(SpatialModel):
    def GetSpatialMap(self):
        self.RA = self.GetParValue('RA')
        self.DEC = self.GetParValue('DEC')
        self.Radius = self.GetParValue('Radius')

        deltax = deltay = 0.01
        nxpix = nypix = 2 * self.Radius / deltax
        radius = self.Radius / deltax
        mod = models.Disk2D(1., nxpix/2, nypix/2, radius)

        x, y = np.mgrid[0:nxpix, 0:nypix]
        data = mod(x, y)
        wcs = maptools.make_wcs(nxpix, nypix, deltax, deltay, 'C', self.RA, self.DEC, 'CAR')
        return SkyMap.NewSpatialMap('RadialDisk.fits', data, wcs)

    def GetSkycrd(self):
        return SkyCoord(ra=self.RA*u.degree, dec=self.DEC*u.degree, frame='fk5')


class RadialGaussian(SpatialModel):
    def GetSpatialMap(self):
        self.RA = self.GetParValue('RA')
        self.DEC = self.GetParValue('DEC')
        self.Sigma = self.GetParValue('Sigma')

        deltax = deltay = 0.01
        nxpix = nypix = 5 * self.Sigma / deltax
        sigma = self.Sigma / deltax
        mod = models.Gaussian2D(1., nxpix/2, nypix/2, sigma, sigma, 0)

        x, y = np.mgrid[0:nxpix, 0:nypix]
        data = mod(x, y)
        wcs = maptools.make_wcs(nxpix, nypix, deltax, deltay, 'C', self.RA, self.DEC, 'CAR')
        return SkyMap.NewSpatialMap('RadialGaussian.fits', data, wcs)

    def GetSkycrd(self):
        return SkyCoord(ra=self.RA*u.degree, dec=self.DEC*u.degree, frame='fk5')


class Ring2D(SpatialModel):
    def GetSpatialMap(self):
        self.RA = self.GetParValue('RA')
        self.DEC = self.GetParValue('DEC')
        self.R_in = self.GetParValue('R_in')
        self.R_out = self.GetParValue('R_out')

        deltax = deltay = 0.01
        nxpix = nypix = 2 * self.R_out / deltax
        r_in = self.R_in / deltax
        r_out = self.R_out / deltax
        mod = models.Disk2D(1., nxpix/2, nypix/2, r_out) - models.Disk2D(1., nxpix/2, nypix/2, r_in)

        x, y = np.mgrid[0:nxpix, 0:nypix]
        data = mod(x, y)
        wcs = maptools.make_wcs(nxpix, nypix, deltax, deltay, 'C', self.RA, self.DEC, 'CAR')
        return SkyMap.NewSpatialMap('Ring2D.fits', data, wcs)

    def GetSkycrd(self):
        return SkyCoord(ra=self.RA*u.degree, dec=self.DEC*u.degree, frame='fk5')


class Ellipse2D(SpatialModel):
    def GetSpatialMap(self):
        self.RA = self.GetParValue('RA')
        self.DEC = self.GetParValue('DEC')
        self.A = self.GetParValue('A')
        self.B = self.GetParValue('B')
        self.Theta = self.GetParValue('Theta')

        deltax = deltay = 0.01
        nxpix = nypix = 2 * max(self.A, self.B) / deltax
        a = self.A / deltax
        b = self.B / deltax
        theta = np.deg2rad(self.Theta)
        mod = models.Ellipse2D(1., nxpix/2, nypix/2, a, b, theta)

        x, y = np.mgrid[0:nxpix, 0:nypix]
        data = mod(x, y)
        wcs = maptools.make_wcs(nxpix, nypix, deltax, deltay, 'C', self.RA, self.DEC, 'CAR')
        return SkyMap.NewSpatialMap('Ellipse2D.fits', data, wcs)
 
    def GetSkycrd(self):
        return SkyCoord(ra=self.RA*u.degree, dec=self.DEC*u.degree, frame='fk5')


class Gaussian2D(SpatialModel):
    def GetSpatialMap(self):
        self.RA = self.GetParValue('RA')
        self.DEC = self.GetParValue('DEC')
        self.X_stddev = self.GetParValue('X_stddev')
        self.Y_stddev = self.GetParValue('Y_stddev')
        self.Theta = self.GetParValue('Theta')

        deltax = deltay = 0.01
        nxpix = nypix = 5 * max(self.X_stddev, self.Y_stddev) / deltax
        x_stddev = self.X_stddev / deltax
        y_stddev = self.Y_stddev / deltay
        theta = np.deg2rad(self.Theta)
        mod = models.Gaussian2D(1., nxpix/2, nypix/2, x_stddev, y_stddev, theta)

        x, y = np.mgrid[0:nxpix, 0:nypix]
        data = mod(x, y)
        wcs = maptools.make_wcs(nxpix, nypix, deltax, deltay, 'C', self.RA, self.DEC, 'CAR')
        return SkyMap.NewSpatialMap('Gaussian2D.fits', data, wcs)

    def GetSkycrd(self):
        return SkyCoord(ra=self.RA*u.degree, dec=self.DEC*u.degree, frame='fk5')