'''
File: Spectrum.py
Author: Duan Kai-Kai
Email: duankk@pmo.ac.cn
'''

import numpy as np
import matplotlib.pyplot as plt
from scipy.integrate import quad
from .Parameter import Parameter

'''example
    >>> import Spectrum

    >>> pars = {'type': 'PowerLaw',
             'normPar': 'Prefactor',
               'Index': {'error': '0.0823208136',
                          'free': '1',
                           'max': '-0',
                           'min': '-5',
                          'name': 'Index',
                         'scale': '1',
                         'value': '-2.410955835'},
           'Prefactor': {'error': '0.07322381815',
                          'free': '1',
                           'max': '100000',
                           'min': 1e-05,
                          'name': 'Prefactor',
                         'scale': 1e-12,
                         'value': '0.4869737351'},
               'Scale': { 'free': '0',
                           'max': '1000',
                           'min': '0.01',
                          'name': 'Scale',
                         'scale': '1000',
                         'value': '1.565626445'}
               }
    >>> spectrum = Spectrum.PowerLaw(pars)

    >>> print spectrum.type, spectrum.normPar, spectrum.ParsDict

    >>> spectrum.GetSpecFunc
'''


class Spectrum:
    def __init__(self, pars):
        self.type = pars['type']

        if 'normPar' in list(pars.keys()):
            self.normPar = pars['normPar']

        if 'apply_edisp' in list(pars.keys()):
            self.apply_edisp = pars['apply_edisp']

        if 'file' in list(pars.keys()):
            self.file = pars['file']

        if 'Branch' in list(pars.keys()):
            self.Branch = pars['Branch']

        self.ParsDict = {}
        for parName, parDict in list(pars.items()):
            try:
                self.ParsDict[parName] = Parameter(parDict)
            except TypeError:
                pass

    def __str__(self):
        return str(self.ParsDict)

    def __repr__(self):
        return str(self.ParsDict)

    def GetSpec(self, e):
        return self.GetSpecFunc(e)

    def GetParValue(self, par):
        return self.ParsDict[par].GetValue

    def plot(self):
        E = np.logspace(3, 6, 100)
        Spec = self.GetSpec(E)
        plt.figure()
        plt.loglog(E, E**2*Spec)
        plt.xlim(1e3, 1e6)
        plt.xlabel('Energy [MeV]')
        plt.ylabel('E$^2$ dN/dE [MeV / cm$^2$ / s]')
        plt.savefig(f'{self.type}.png')

class ConstantValue(Spectrum):
    def GetSpecFunc(self, e):
        value = self.GetParValue('Value')

        if type(e) in (type(0), type(0.0)):
            dnde = value
        else:
            dnde = np.array([value] * len(e))
        return dnde


class PowerLaw(Spectrum):
    def GetSpecFunc(self, e):
        n = self.GetParValue('Prefactor')
        gamma = self.GetParValue('Index')
        e0 = self.GetParValue('Scale')

        return n * (e / e0)**gamma


class BrokenPowerLaw(Spectrum):
    def GetSpecFunc(self, e):
        n = self.GetParValue('Prefactor')
        gamma1 = self.GetParValue('Index1')
        gamma2 = self.GetParValue('Index2')
        eb = self.GetParValue('BreakValue')

        if type(e) in (int, float):
            if e < eb:
                dnde = n * (e / eb)**gamma1
            else:
                dnde = n * (e / eb)**gamma2
        else:
            dnde = np.zeros_like(e)
            dnde[e <= eb] = n * (e[e <= eb] / eb)**gamma1
            dnde[e >= eb] = n * (e[e >= eb] / eb)**gamma2
        return dnde


class PowerLaw2(Spectrum):
    def GetSpecFunc(self, e):
        n = self.GetParValue('Integral')
        gamma = self.GetParValue('Index')
        emin = self.GetParValue('LowerLimit')
        emax = self.GetParValue('UpperLimit')

        return (n * (gamma + 1) * e**gamma)\
             / (emax**(gamma + 1) - emin**(gamma + 1))


class BrokenPowerLaw2(Spectrum):
    def GetSpecFunc(self, e):
        n = self.GetParValue('Integral')
        gamma1 = self.GetParValue('Index1')
        gamma2 = self.GetParValue('Index2')
        eb = self.GetParValue('BreakValue')
        emin = self.GetParValue('LowerLimit')
        emax = self.GetParValue('UpperLimit')

        if emax < eb:
            n0 = n / quad(lambda ex: (ex / eb)**gamma1, emin, emax)[0]
        elif emin > eb:
            n0 = n / quad(lambda ex: (ex / eb)**gamma2, emin, emax)[0]
        else:
            n0 = n / (quad(lambda ex: (ex / eb)**gamma1, emin, eb)[0] +
                      quad(lambda ex: (ex / eb)**gamma2, eb, emax)[0])

        if type(e) in (int, float):
            if e < eb:
                dnde = n0 * (e / eb)**gamma1
            else:
                dnde = n0 * (e / eb)**gamma2
        else:
            dnde = np.zeros_like(e)
            dnde[e <= eb] = n0 * (e[e <= eb] / eb)**gamma1
            dnde[e >= eb] = n0 * (e[e >= eb] / eb)**gamma2
        return dnde


class SmoothBrokenPowerLaw(Spectrum):
    def GetSpecFunc(self, e):
        n = self.GetParValue('Prefactor')
        gamma1 = self.GetParValue('Index1')
        e0 = self.GetParValue('Scale')
        gamma2 = self.GetParValue('Index2')
        eb = self.GetParValue('BreakValue')
        beta = self.GetParValue('Beta')

        return n * (e / e0)**gamma1\
                 * (1 + (e / eb)**((gamma1 - gamma2) / beta))**(-1 * beta)


class LogParabola(Spectrum):
    def GetSpecFunc(self, e):
        n = self.GetParValue('norm')
        alpha = self.GetParValue('alpha')
        beta = self.GetParValue('beta')
        eb = self.GetParValue('Eb')

        return n * (e / eb)**(-1 * (alpha + beta * np.log(e / eb)))


class ExpCutoff(Spectrum):
    def GetSpecFunc(self, e):
        n = self.GetParValue('Prefactor')
        gamma = self.GetParValue('Index')
        e0 = self.GetParValue('Scale')
        eb = self.GetParValue('Ebreak')
        p1 = self.GetParValue('P1')
        p2 = self.GetParValue('P2')
        p3 = self.GetParValue('P3')

        if type(e) in (int, float):
            if e < eb:
                dnde = n * (e / e0)**gamma
            else:
                dnde = n * (e / e0)**gamma\
                         * np.exp(-1 * 
                           ((e - eb) / p1
                            + p2 * np.log(e / eb)
                            + 2. * p3 * np.log(e / eb)))
        else:
            dnde = np.zeros_like(e)
            dnde[e <= eb] = n * (e[e <= eb] / e0)**gamma
            dnde[e >= eb] = n * (e[e >= eb] / e0)**gamma\
                         * np.exp(-1 * 
                           ((e[e >= eb] - eb) / p1
                            + p2 * np.log(e[e >= eb] / eb)
                            + 2. * p3 * np.log(e[e >= eb] / eb)))
        return dnde


class BPLExpCutoff(Spectrum):
    def GetSpecFunc(self, e):
        n = self.GetParValue('Prefactor')
        gamma1 = self.GetParValue('Index1')
        gamma2 = self.GetParValue('Index2')
        eb = self.GetParValue('BreakValue')
        eabs = self.GetParValue('Eabs')
        p1 = self.GetParValue('P1')

        if type(e) in (int, float):
            if e <= eb and e <= eabs:
                dnde = n * (e / eb)**gamma1
            if e > eb and e < eabs:
                dnde = n * (e / eb)**gamma2
            if e < eb and e > eabs:
                dnde = n * (e / eb)**gamma1 * np.exp(-1 * (e - eabs) / p1)
            if e >= eb and e >= eabs:
                dnde = n * (e / eb)**gamma2 * np.exp(-1 * (e - eabs) / p1)
        else:
            dnde = np.zeros_like(e)
            dnde[(e <= eb)*(e <= eabs)] = n * (e[(e <= eb)*(e <= eabs)] / eb)**gamma1
            dnde[(e > eb) * (e < eabs)] = n * (e[(e > eb) * (e < eabs)] / eb)**gamma2
            dnde[(e < eb) * (e > eabs)] = n * (e[(e < eb) * (e > eabs)] / eb)**gamma1 \
                                        * np.exp(-1 * (e[(e < eb)*(e > eabs)] - eabs) / p1)
            dnde[(e >= eb)*(e >= eabs)] = n * (e[(e >= eb)*(e >= eabs)] / eb)**gamma2 \
                                        * np.exp(-1 * (e[(e >= eb)*(e >= eabs)] - eabs) / p1)
        return dnde


class PLSuperExpCutoff(Spectrum):
    def GetSpecFunc(self, e):
        n = self.GetParValue('Prefactor')
        gamma1 = self.GetParValue('Index1')
        e0 = self.GetParValue('Scale')
        ec = self.GetParValue('Cutoff')
        gamma2 = self.GetParValue('Index2')

        return n * (e / e0)**gamma1 * np.exp(-1 * (e / ec)**gamma2)


class PLSuperExpCutoff2(Spectrum):
    def GetSpecFunc(self, e):
        n = self.GetParValue('Prefactor')
        gamma1 = self.GetParValue('Index1')
        e0 = self.GetParValue('Scale')
        a = self.GetParValue('Expfactor')
        gamma2 = self.GetParValue('Index2')

        return n * (e / e0)**gamma1 * np.exp(-a * e**(gamma2))


class Gaussian(Spectrum):
    def GetSpecFunc(self, e):
        n = self.GetParValue('Prefactor')
        emean = self.GetParValue('Mean')
        sigma = self.GetParValue('Sigma')

        return n / (sigma * np.sqrt(2 * np.pi))\
             * np.exp(-1 * (e - emean)**2. / (2. * sigma**2.))


class BandFunction(Spectrum):
    def GetSpecFunc(self, e):
        n = self.GetParValue('norm')
        alpha = self.GetParValue('alpha')
        beta = self.GetParValue('beta')
        ep = self.GetParValue('Ep')

        if type(e) in (int, float):
            if e < ep * (alpha - beta) / (alpha - 2.):
                dnde = n * (e / 0.1)**alpha\
                     * np.exp(-1 * (e / ep)**(alpha + 2.))
            else:
                dnde = n * (e / 0.1)**beta\
                     * ((ep / 0.1) * (alpha - beta) / (alpha + 2.))**(alpha - beta)\
                     * np.exp(beta - alpha)
        else:
            dnde = np.zeros_like(e)
            dnde[e < ep * (alpha - beta) / (alpha - 2.)] = n * (e[e < ep * (alpha - beta) / (alpha - 2.)] / 0.1)**alpha\
                * np.exp(-1 * (e[e < ep * (alpha - beta) / (alpha - 2.)] / ep)**(alpha + 2.))
            dnde[e >= ep * (alpha - beta) / (alpha - 2.)] = n * (e[e >= ep * (alpha - beta) / (alpha - 2.)] / 0.1)**beta\
                * ((ep / 0.1) * (alpha - beta) / (alpha + 2.))**(alpha - beta)\
                    * np.exp(beta - alpha)
        return dnde


class FileFunction(Spectrum):
    def GetSpecFunc(self, e):
        n = self.GetParValue('Normalization')
        f = self.file
        if '$' in f:
            import os
            f1 = f.split(')')[0].split('(')[1]
            f2 = f.split(')')[1]
            f = os.environ.get(f1) + f2
        data = np.loadtxt(f)

        from scipy.interpolate import interp1d

        return n * 10**(interp1d(np.log10(data[:, 0]), np.log10(data[:, 1]), fill_value='extrapolate')(np.log10(e)))
