'''
File: UnbinnedLikelihood.py
Author: Duan Kai-Kai
Email: duankk@pmo.ac.cn
'''

from pprint import pprint
import numpy as np
from .maptools import Sep
from .LikelihoodBase import LikelihoodBase


class UnbinnedLikelihood(LikelihoodBase):
    '''
    This class is Unbinned Likelihood class.        
    It stores the events, exposure class and/or the srcmaps.
    for calculate the events' probability for each source and the likelihood
    '''
    def __init__(self, CubeType, events, ccube, exposure, emin, emax, srcmaps=None, fitRegion=None, phi_dependence=False):
        assert isinstance(CubeType, str)
        assert isinstance(emin, (int, float))
        assert isinstance(emax, (int, float))
        assert isinstance(phi_dependence, bool)

        if not CubeType in ['Map', 'Hpx']:
            raise RuntimeError('The CubeType must be one of Map or Hpx, please check it.')
        
        if not emax > emin:
            raise RuntimeError('The emax must be larger than emin, please check it.')
        
        self.CubeType = CubeType
        self.events = events
        self.exposure = exposure
        self.ccube = ccube
        emins = self.ccube.emins
        emaxs = self.ccube.emaxs
        self.emin = emins[abs(emins - emin) == min(abs(emins - emin))][0]
        self.emax = emaxs[abs(emaxs - emax) == min(abs(emaxs - emax))][0]
        self.imin = list(emins).index(self.emin)
        self.imax = list(emaxs).index(self.emax)

        if CubeType == 'Map':
            self.N = self.ccube.data[:, :, self.imin:self.imax+1]
        if CubeType == 'Hpx':
            self.N = self.ccube.data[:, self.imin:self.imax+1]
        self.emins = emins[self.imin:self.imax+1]
        self.emaxs = emaxs[self.imin:self.imax+1]
        self.ebins = self.N.shape[-1]
        print('the energy range is {}:{}MeV to {}:{}MeV'.format(self.imin, self.emin, self.imax, self.emax))

        if srcmaps:
            self.srcmaps = srcmaps
        else:
            self.srcmaps = {}

        self.fitRegion = fitRegion
        self.phi_dependence = phi_dependence

        self.mdlcubes = {}

        self.modelDict = {}
        self.SrcDict = {}
        self.ParsDict = {}
        self.FixSrcDict = {}
        self.FreeSrcDict = {}
        self.FixParsDict = {}
        self.FreeParsDict = {}

    def __str__(self):
        print('Fix Source:', self.FixSrcDict.keys())
        print('Free Source:', self.FreeSrcDict.keys())
        print('Fix Parameter:', self.FixParsDict.keys())
        print('Free Parameter:', self.FreeParsDict.keys())

    def EventProb(self, **kwargs):
        for key, value in list(kwargs.items()):
            srcName, parName = key.split('__')
            self.SetParValue(srcName, parName, value)

        Efilter = (self.events.Energy > self.emin) * (self.events.Energy <= self.emax)
        Energy = self.events.Energy[Efilter]
        PsfProbs = self.events.PsfProbs[Efilter]
        # print(self.events.skycrd_C)
        skycrds_C = self.events.skycrd_C[Efilter]

        lnprob = 0
        for(energy, psfProbs, skycrd_C) in zip(Energy, PsfProbs, skycrds_C):
            # print(energy, psfProbs, skycrd_C)
            if self.fitRegion:
                for region in self.fitRegion.inRegionList:
                    if Sep(skycrd_C, [region.skycrd.fk5.ra.value, region.skycrd.fk5.dec.value]) > region.rad.value:
                        continue
                for region in self.fitRegion.outRegionList:
                    if Sep(skycrd_C, [region.skycrd.fk5.ra.value, region.skycrd.fk5.dec.value]) < region.rad.value:
                        continue

            probability = 0
            for srcName, Source in list(self.SrcDict.items()):
                psfProb = psfProbs[srcName]
                dnde = Source.spectrum.GetSpecFunc(energy)
                probability += psfProb * dnde
            lnprob += np.log(probability)
        return lnprob

    def __call__(self, *arg):
        kwargs = dict(list(zip(sorted(self.FreeParsDict.keys()), arg)))
        # pprint(kwargs)
        ROIint = self.FreeModelCube(**kwargs) + self.Fixmdlcube
        EventProb = self.EventProb(**kwargs)
        if self.fitRegion:
            if self.CubeType == 'Map':
                fitRegion = np.zeros(self.ccube.nxpix * self.ccube.nypix, dtype=bool)
            if self.CubeType in ['Hpx', 'MaskedHpx']:
                fitRegion = np.zeros(self.ccube.npix, dtype=bool)

            for region in self.fitRegion.inRegionList:
                # print(region)
                fitRegion += region.skycrd.separation(self.ccube.skycrd) < region.rad
            for region in self.fitRegion.outRegionList:
                # print(region)
                fitRegion *= region.skycrd.separation(self.ccube.skycrd) > region.rad

            if self.CubeType == 'Map':
                fitRegion = fitRegion.reshape(self.ccube.nxpix, self.ccube.nypix)
            likelihood = -2 * (EventProb - sum(ROIint[fitRegion]))
        else:
            likelihood = -2 * (EventProb - np.sum(ROIint))

        # print('Observed Number = ', np.sum(self.N))
        # print('Model Number = ', np.sum(ROIint))
        # print('likelihood = ', likelihood)
        return likelihood
