'''
File: maptools.py
Author: Duan Kai-Kai
Email: duankk@pmo.ac.cn
'''

from time import time
import numpy as np
import healpy as hp
from astropy.io import fits
from astropy.wcs import WCS
from astropy import units as u
from astropy.coordinates import SkyCoord

def myprint(func):
    def wrapper(*arg, **kwd):
        startTime = time()
        func(*arg, **kwd)
        endTime = time()
        print('-------> elapsed time: %f s'%(endTime - startTime))
    return wrapper


def make_wcs(nxpix, nypix, deltax, deltay, coord, xref, yref, proj):
    wcs = WCS(naxis=2)
    wcs.wcs.crpix = [nxpix / 2. + 0.5, nypix / 2. + 0.5]
    wcs.wcs.cdelt = [-deltax, deltay]
    wcs.wcs.crval = [xref, yref]
    wcs.wcs.cunit = ['deg', 'deg']

    if coord == 'G':
        wcs.wcs.ctype = ['GLON-' + proj, 'GLAT-' + proj]
    if coord == 'C':
        wcs.wcs.ctype = ['RA---' + proj, 'DEC--' + proj]
    return wcs


def skycrdG2C(skycrd_G):
    if type(skycrd_G) is not np.ndarray:
        skycrd_G = np.array(skycrd_G)

    if skycrd_G.ndim == 1:
        skycrd_G = np.array([skycrd_G], dtype='float64')
    c = SkyCoord(skycrd_G[:, 0] * u.degree,
                 skycrd_G[:, 1] * u.degree,
                 frame='galactic')
    skycrd_C = np.array((c.fk5.ra.value,
                         c.fk5.dec.value)).T
    return skycrd_C


def skycrdC2G(skycrd_C):
    if type(skycrd_C) is not np.ndarray:
        skycrd_C = np.array(skycrd_C)

    if skycrd_C.ndim == 1:
        skycrd_C = np.array([skycrd_C], dtype='float64')
    c = SkyCoord(skycrd_C[:, 0] * u.degree,
                 skycrd_C[:, 1] * u.degree,
                 frame='fk5')
    skycrd_G = np.array((c.galactic.l.value,
                         c.galactic.b.value)).T
    return skycrd_G


def skycrd2ipix(ordering, nside, skycrd):
    if type(skycrd) is not np.ndarray:
        skycrd = np.array(skycrd)

    if skycrd.ndim == 1:
        skycrd = np.array([skycrd], dtype='float64')
    l = skycrd[:, 0]
    b = skycrd[:, 1]
    theta = np.deg2rad(90. - b)
    phi = np.deg2rad(l)
    if ordering == 'NESTED':
        nest = True
    if ordering == 'RING':
        nest = False
    ipix = hp.ang2pix(nside, theta, phi, nest)
    return ipix


def ipix2skycrd(ordering, nside, ipix):
    if ordering == 'NESTED':
        nest = True
    if ordering == 'RING':
        nest = False
    theta, phi = hp.pix2ang(nside, ipix, nest)
    l = np.rad2deg(phi)
    b = 90 - np.rad2deg(theta)
    skycrd = np.array([l, b]).T
    return skycrd


def ipixG2C(ordering, nside, ipix_G):
    skycrd_G = ipix2skycrd(ordering, nside, ipix_G)
    skycrd_C = skycrdG2C(skycrd_G)
    ipix_C = skycrd2ipix(ordering, nside, skycrd_C)
    return ipix_C


def ipixC2G(ordering, nside, ipix_C):
    skycrd_C = ipix2skycrd(ordering, nside, ipix_C)
    skycrd_G = skycrdC2G(skycrd_C)
    ipix_G = skycrd2ipix(ordering, nside, skycrd_G)
    return ipix_G

def cosSep(skycrd1, skycrd2):
    if type(skycrd1) is not np.ndarray:
        skycrd1 = np.array(skycrd1)
    if type(skycrd2) is not np.ndarray:
        skycrd2 = np.array(skycrd2)

    if skycrd1.ndim == 1:
        skycrd1 = np.array([skycrd1], dtype='float64')
    if skycrd2.ndim == 1:
        skycrd2 = np.array([skycrd2], dtype='float64')

    l1 = np.deg2rad(skycrd1[:, 0])
    b1 = np.deg2rad(skycrd1[:, 1])
    l2 = np.deg2rad(skycrd2[:, 0])
    b2 = np.deg2rad(skycrd2[:, 1])
    cos_sep = np.cos(abs(l1 - l2)) * np.cos(b1) * np.cos(b2) + np.sin(b1) * np.sin(b2)
    cos_sep[cos_sep > 1] = 1.0
    cos_sep[cos_sep < -1] = -1.0
    return cos_sep

def sinSep(skycrd1, skycrd2):
    cos_sep = cosSep(skycrd1, skycrd2)
    sin_sep = np.sqrt(1 - cos_sep**2)
    return sin_sep

def Sep(skycrd1, skycrd2):
    cos_sep = cosSep(skycrd1, skycrd2)
    sep = np.arccos(cos_sep)
    return np.rad2deg(sep)


def from_Map_to_Map(Map1, Map2):
    if Map1.naxis == 2:
        NewMap = np.zeros((Map2.nxpix, Map2.nypix))
    if Map1.naxis == 3:
        NewMap = np.zeros((Map2.nxpix, Map2.nypix, Map1.nzpix))

    if abs(Map1.deltax) >= abs(Map2.deltax):
        if Map1.coord == 'G':
            pixcrd = Map1.wcs.wcs_world2pix(Map2.skycrd_G, 0).astype('int')
        if Map1.coord == 'C':
            pixcrd = Map1.wcs.wcs_world2pix(Map2.skycrd_C, 0).astype('int')

        for i, pixcrd0 in enumerate(Map2.pixcrd):
            if pixcrd[i, 0] < 0 or pixcrd[i, 1] < 0 or \
               pixcrd[i, 0] >= Map1.nxpix or pixcrd[i, 1] >= Map1.nypix:
                continue
            NewMap[tuple(pixcrd0)] = Map1.data[tuple(pixcrd[i])]
    
    if abs(Map1.deltax) < abs(Map2.deltax):
        if Map2.coord == 'G':
            pixcrd = Map2.wcs.wcs_world2pix(Map1.skycrd_G, 0).astype('int')
        if Map2.coord == 'C':
            pixcrd = Map2.wcs.wcs_world2pix(Map1.skycrd_C, 0).astype('int')

        Omega = np.zeros_like(NewMap)
        for i, pixcrd0 in enumerate(Map1.pixcrd):
            if pixcrd[i, 0] < 0 or pixcrd[i, 1] < 0 or \
               pixcrd[i, 0] >= Map2.nxpix or pixcrd[i, 1] >= Map2.nypix:
                continue

            NewMap[tuple(pixcrd[i])] += Map1.data[tuple(pixcrd0)] * Map1.dOmega[tuple(pixcrd0)]
            Omega[tuple(pixcrd[i])] += Map1.dOmega[tuple(pixcrd0)]

        NewMap[Omega != 0] = NewMap[Omega != 0] / Omega[Omega != 0]
    return NewMap

def from_Map_to_Hpx(Map, Hpx):
    if Map.naxis == 2:
        NewHpx = np.zeros(Hpx.npix)
    if Map.naxis == 3:
        NewHpx = np.zeros((Hpx.npix, Map.nzpix))

    if abs(Map.deltax) >= abs(Hpx.resol):
        if Map.coord == 'G':
            pixcrd = Map.wcs.wcs_world2pix(Hpx.skycrd_G, 0).astype('int')
        if Map.coord == 'C':
            pixcrd = Map.wcs.wcs_world2pix(Hpx.skycrd_C, 0).astype('int')

        for i in range(Hpx.npix):
            if pixcrd[i, 0] < 0 or pixcrd[i, 1] < 0 or \
               pixcrd[i, 0] >= Map.nxpix or pixcrd[i, 1] >= Map.nypix:
                   continue
            NewHpx[i] = Map.data[tuple(pixcrd[i])]

    if abs(Map.deltax) < abs(Hpx.resol):
        if Hpx.coord == 'G':
            ipix = skycrd2ipix(Hpx.ordering, Hpx.nside, Map.skycrd_G)
        if Hpx.coord == 'C':
            ipix = skycrd2ipix(Hpx.ordering, Hpx.nside, Map.skycrd_C)

        Omega = np.zeros_like(NewHpx)
        for i, pixcrd0 in enumerate(Map.pixcrd):
            if ipix[i] < 0 or ipix[i] > Hpx.npix:
                continue

            NewHpx[ipix[i]] += Map.data[tuple(pixcrd0)] * Map.dOmega[tuple(pixcrd0)]
            Omega[ipix[i]] += Map.dOmega[tuple(pixcrd0)]

        NewHpx[Omega != 0] = NewHpx[Omega != 0] / Omega[Omega != 0]
    return NewHpx

def from_Hpx_to_Map(Hpx, Map):
    if Hpx.data.ndim == 1:
        NewMap = np.zeros((Map.nxpix, Map.nypix))
    if Hpx.data.ndim == 2:
        NewMap = np.zeros((Map.nxpix, Map.nypix, Hpx.fields))

    # if abs(Hpx.resol) >= abs(Map.deltax):
    if Hpx.coord == 'G':
        ipix = skycrd2ipix(Hpx.ordering, Hpx.nside, Map.skycrd_G)
    if Hpx.coord == 'C':
        ipix = skycrd2ipix(Hpx.ordering, Hpx.nside, Map.skycrd_C)

    if abs(Hpx.resol) >= abs(Map.deltax):
        for i, pixcrd in enumerate(Map.pixcrd):
            if ipix[i] < 0 or ipix[i] >= Hpx.npix:
                continue
            NewMap[pixcrd] = Hpx.data[ipix[i]]

    if abs(Hpx.resol) < abs(Map.deltax):
        if Hpx.coord == 'G':
            pixcrd = Map.wcs.wcs_world2pix(Hpx.skycrd_G, 0).astype('int')
        if Hpx.coord == 'C':
            pixcrd = Map.wcs.wcs_world2pix(Hpx.skycrd_C, 0).astype('int')

        Omega = np.zeros_like(NewMap)
        for i in range(Hpx.npix):
            if pixcrd[i, 0] < 0 or pixcrd[i, 1] < 0 or \
               pixcrd[i, 0] >= Map.nxpix or pixcrd[i, 1] >= Map.nypix:
                   continue
            NewMap[tuple(pixcrd[i])] += Hpx.data[i] * Hpx.dOmega
            Omega[tuple(pixcrd[i])] += Hpx.dOmega

        NewMap[Omega != 0] = NewMap[Omega != 0] / Omega[Omega != 0]

        for pixcrd in Map.pixcrd:
            # print(pixcrd, NewMap[tuple(pixcrd)])
            if sum(Omega[tuple(pixcrd)]) == 0:
                if pixcrd[0] == 0 and pixcrd[1] == 0:
                    NewMap[tuple(pixcrd)] = (NewMap[0, 1]
                                            +NewMap[1, 0]
                                            +NewMap[1, 1]) / 3
                elif pixcrd[0] == Map.nxpix-1 and pixcrd[1] == 0:
                    NewMap[tuple(pixcrd)] = (NewMap[-2, 0]
                                            +NewMap[-1, 1]
                                            +NewMap[-2, 1]) / 3
                elif pixcrd[0] == 0 and pixcrd[1] == Map.nypix-1:
                    NewMap[tuple(pixcrd)] = (NewMap[0, -2]
                                            +NewMap[1, -1]
                                            +NewMap[1, -2]) / 3
                elif pixcrd[0] == Map.nxpix-1 and pixcrd[1] == Map.nypix-1:
                    NewMap[tuple(pixcrd)] = (NewMap[-1,-2]
                                            +NewMap[-2,-1]
                                            +NewMap[-2,-2]) / 3
                elif pixcrd[0] == 0:
                    NewMap[tuple(pixcrd)] = (NewMap[0, pixcrd[1]-1]
                                           + NewMap[0, pixcrd[1]+1]
                                           + NewMap[1, pixcrd[1]-1]
                                           + NewMap[1, pixcrd[1]]
                                           + NewMap[1, pixcrd[1]+1]) / 5
                elif pixcrd[1] == 0:
                    NewMap[tuple(pixcrd)] = (NewMap[pixcrd[0]-1, 0]
                                           + NewMap[pixcrd[0]+1, 0]
                                           + NewMap[pixcrd[0], 1]
                                           + NewMap[pixcrd[0]-1, 1]
                                           + NewMap[pixcrd[0]+1, 1]) / 5
                elif pixcrd[0] == Map.nxpix-1:
                    NewMap[tuple(pixcrd)] = (NewMap[-1, pixcrd[1]-1]
                                           + NewMap[-1, pixcrd[1]+1]
                                           + NewMap[-2, pixcrd[1]-1]
                                           + NewMap[-2, pixcrd[1]]
                                           + NewMap[-2, pixcrd[1]+1]) / 5
                elif pixcrd[1] == Map.nypix-1:
                    NewMap[tuple(pixcrd)] = (NewMap[pixcrd[0]-1, -1]
                                           + NewMap[pixcrd[0]+1, -1]
                                           + NewMap[pixcrd[0], -2]
                                           + NewMap[pixcrd[0]-1, -2]
                                           + NewMap[pixcrd[0]+1, -2]) / 5
                else:
                    NewMap[tuple(pixcrd)] = (NewMap[pixcrd[0], pixcrd[1]-1]
                                           + NewMap[pixcrd[0], pixcrd[1]+1]
                                           + NewMap[pixcrd[0]-1, pixcrd[1]]
                                           + NewMap[pixcrd[0]+1, pixcrd[1]]
                                           + NewMap[pixcrd[0]-1, pixcrd[1]-1]
                                           + NewMap[pixcrd[0]-1, pixcrd[1]+1]
                                           + NewMap[pixcrd[0]+1, pixcrd[1]-1]
                                           + NewMap[pixcrd[0]+1, pixcrd[1]+1]) / 8
    return NewMap

def from_Hpx_to_Hpx(Hpx1, Hpx2):
    if Hpx1.data.ndim == 1:
        NewHpx = np.zeros(Hpx2.npix)
    if Hpx1.data.ndim == 2:
        NewHpx = np.zeros((Hpx2.npix, Hpx1.fields))

    if abs(Hpx1.resol) >= abs(Hpx2.resol):
        if Hpx1.coord == 'G':
            ipix = skycrd2ipix(Hpx1.ordering, Hpx1.nside, Hpx2.skycrd_G)
        if Hpx1.coord == 'C':
            ipix = skycrd2ipix(Hpx1.ordering, Hpx1.nside, Hpx2.skycrd_C)

        for i in range(Hpx2.npix):
            if ipix[i] < 0 or ipix[i] >= Hpx1.npix:
                continue
            NewHpx[i] = Hpx1.data[ipix[i]]

    if abs(Hpx1.resol) < abs(Hpx2.resol):
        if Hpx1.coord == 'G':
            ipix = skycrd2ipix(Hpx2.ordering, Hpx2.nside, Hpx1.skycrd_G)
        if Hpx1.coord == 'C':
            ipix = skycrd2ipix(Hpx2.ordering, Hpx2.nside, Hpx1.skycrd_C)

        Omega = np.zeros_like(NewHpx)
        for i in range(Hpx1.npix):
            if ipix[i] < 0 or ipix[i] >= Hpx2.npix:
                continue

            NewHpx[ipix[i]] += Hpx1.data[i] * Hpx1.dOmega
            Omega[ipix[i]] += Hpx1.dOmega

        NewHpx[Omega != 0] = NewHpx[Omega != 0] / Omega[Omega != 0]
    return NewHpx

def Fix_PSFMap(psfmap, skycrd0_C, ccube, exposure, logE):
    skycrd0_G = skycrdC2G(skycrd0_C)
    if ccube.coord == 'C':
        pixcrd0 = ccube.wcs.wcs_world2pix([skycrd0_C], 0)
    else:
        pixcrd0 = ccube.wcs.wcs_world2pix([skycrd0_G], 0)

    if pixcrd0[0, 0] < 0 or pixcrd0[0, 0] > ccube.nxpix or\
       pixcrd0[0, 1] < 0 or pixcrd0[0, 1] > ccube.nypix:
        return psfmap

    if exposure.getwpsf(skycrd0_C, logE, 0.5) < abs(ccube.deltax):
        psfmap[round(pixcrd0[0, 0]), round(pixcrd0[0, 1])] = 0
        resum = np.sum(psfmap * ccube.dOmega)
        psfmap[round(pixcrd0[0, 0]), round(pixcrd0[0, 1])] =\
                (1 - resum) / ccube.dOmega[round(pixcrd0[0, 0]), round(pixcrd0[0, 1])]
        return psfmap
    else:
        return psfmap


def SpatialAverage(Map, pixcrd):
    if pixcrd[0] + 1 == Map.nxpix and pixcrd[1] + 1 == Map.nypix:
        return (Map.data[pixcrd[0], pixcrd[1]]\
              + Map.data[pixcrd[0] - 1, pixcrd[1]]\
              + Map.data[pixcrd[0], pixcrd[1] - 1]\
              + Map.data[pixcrd[0] - 1, pixcrd[1] - 1]) / 4

    if pixcrd[0] + 1 == Map.nxpix:
        return (Map.data[pixcrd[0], pixcrd[1]]\
              + Map.data[pixcrd[0] - 1, pixcrd[1]]\
              + Map.data[pixcrd[0], pixcrd[1] - 1]\
              + Map.data[pixcrd[0], pixcrd[1] + 1]\
              + Map.data[pixcrd[0] - 1, pixcrd[1] - 1]\
              + Map.data[pixcrd[0] - 1, pixcrd[1] + 1]) / 6

    if pixcrd[1] + 1 == Map.nypix:
        return (Map.data[pixcrd[0], pixcrd[1]]\
              + Map.data[pixcrd[0] - 1, pixcrd[1]]\
              + Map.data[pixcrd[0] + 1, pixcrd[1]]\
              + Map.data[pixcrd[0], pixcrd[1] - 1]\
              + Map.data[pixcrd[0] - 1, pixcrd[1] - 1]\
              + Map.data[pixcrd[0] + 1, pixcrd[1] - 1]) / 6

    return (Map.data[pixcrd[0], pixcrd[1]]\
          + Map.data[pixcrd[0] - 1, pixcrd[1]]\
          + Map.data[pixcrd[0] + 1, pixcrd[1]]\
          + Map.data[pixcrd[0], pixcrd[1] - 1]\
          + Map.data[pixcrd[0], pixcrd[1] + 1]\
          + Map.data[pixcrd[0] - 1, pixcrd[1] - 1]\
          + Map.data[pixcrd[0] - 1, pixcrd[1] + 1]\
          + Map.data[pixcrd[0] + 1, pixcrd[1] - 1]\
          + Map.data[pixcrd[0] + 1, pixcrd[1] + 1]) / 9
