import os
import numpy as np
import matplotlib.pyplot as plt
from astropy.io import fits
from DmpST import DmpTime, maptools, SpaceCraft
from plotObservation import plotObservedTime, plotObservedAngle, plotLiveTimeMap, plotLiveTime_cosTheta, plotLiveTime_Phi


def FindMins(array):
    imin = np.arange(len(array))
    index = np.hstack([False, (array[1:] < array[:-1])])\
          * np.hstack([(array[:-1] < array[1:]), False])
    return imin[index], array[index]

def FindMaxs(array):
    imax = np.arange(len(array))
    index = np.hstack([False, (array[1:] > array[:-1])])\
          * np.hstack([(array[:-1] > array[1:]), False])
    return imax[index], array[index]


def FindObservedTime(scfile, srcName, GorC, xref, yref, UTCmin, UTCmax):
    if not os.path.exists(scfile):
        raise FileNotFoundError('The scfile not be found, please check it.')

    if not os.path.exists(srcName):
        os.mkdir(srcName)

    if GorC == 'G':
        skycrd_G = (xref, yref)
        skycrd_C = maptools.skycrdG2C(skycrd_G)
    if GorC == 'C':
        skycrd_C = (xref, yref)
        skycrd_G = maptools.skycrdC2G(skycrd_C)

    spacecraft = SpaceCraft.SpaceCraft(scfile)
    SCmin = spacecraft.SCmin
    SCmax = spacecraft.SCmax
    skycrd_SCX = spacecraft.skycrd_scx
    skycrd_SCY = spacecraft.skycrd_scy
    skycrd_SCZ = spacecraft.skycrd_scz

    METmin = DmpTime.UTC2MET(UTCmin)
    METmax = DmpTime.UTC2MET(UTCmax)
    tfilter = (SCmin > METmin) * (SCmax < METmax)
    SCmin = SCmin[tfilter]
    SCmax = SCmax[tfilter]
    skycrd_SCX = skycrd_SCX[tfilter]
    skycrd_SCY = skycrd_SCY[tfilter]
    skycrd_SCZ = skycrd_SCZ[tfilter]

    SCX_sep = maptools.Sep(skycrd_C, skycrd_SCX)
    SCY_sep = maptools.Sep(skycrd_C, skycrd_SCY)
    SCZ_sep = maptools.Sep(skycrd_C, skycrd_SCZ)
    Theta = 180 - SCZ_sep
    # phi = np.arccos(np.cos(np.deg2rad(SCX_sep)) / np.sin(np.deg2rad(SCZ_sep)))
    # phi[SCY_sep < 90] = np.pi + phi[SCY_sep < 90]
    # phi[SCY_sep > 90] = np.pi - phi[SCY_sep > 90]
    # Phi = np.rad2deg(phi)
    # np.savetxt(f'{srcName}/ObservedAngle.dat', np.array([SCmin, SCmax, Theta, Phi]).T, header='SCmin, SCmax, Theta, Phi')
    # plotObservedAngle(srcName)

    imins, sepMins = FindMins(Theta)
    imaxs, sepMaxs = FindMaxs(Theta)
    if imins.size < imaxs.size:
        imaxs = imaxs[0:imins.size]
        sepMaxs = sepMaxs[0:imins.size]
    if imins.size > imaxs.size:
        imins = imins[0:imaxs.size]
        sepMins = sepMins[0:imaxs.size]
    np.savetxt(f'{srcName}/ObservedTime.dat', np.array([SCmin[imins], SCmax[imaxs], sepMins, sepMaxs]).T, header='Tmins, Tmaxs, sepMins, sepMaxs')
    plotObservedTime(srcName, 60)

def FindLiveTime(scfile, srcName, GorC, xref, yref, UTCmin, UTCmax, thetabins=40, phibins=72, psibins=50):
    if not os.path.exists(scfile):
        raise FileNotFoundError('The scfile not be found, please check it.')

    if not os.path.exists(srcName):
        os.mkdir(srcName)

    if GorC == 'G':
        skycrd0_G = (xref, yref)
        skycrd0_C = maptools.skycrdG2C(skycrd0_G)
    if GorC == 'C':
        skycrd0_C = (xref, yref)
        skycrd0_G = maptools.skycrdC2G(skycrd0_C)

    cosThetabounds = np.linspace(0.5, 1.0, thetabins+1)
    cosThetamins = cosThetabounds[:-1]
    cosThetamaxs = cosThetabounds[1:]
    np.savetxt(f'{srcName}/cosThetabounds.dat', cosThetabounds.T)

    phibounds = np.linspace(0, 360, phibins+1)
    phimins = phibounds[:-1]
    phimaxs = phibounds[1:]
    np.savetxt(f'{srcName}/phibounds.dat', phibounds.T)

    spacecraft = SpaceCraft.SpaceCraft(scfile)
    SCmin = spacecraft.SCmin
    SCmax = spacecraft.SCmax
    skycrd_SCX = spacecraft.skycrd_scx
    skycrd_SCY = spacecraft.skycrd_scy
    skycrd_SCZ = spacecraft.skycrd_scz
    livetime = spacecraft.livetime

    METmin = DmpTime.UTC2MET(UTCmin)
    METmax = DmpTime.UTC2MET(UTCmax)
    tfilter = (SCmin > METmin) * (SCmax < METmax)
    SCmin = SCmin[tfilter]
    SCmax = SCmax[tfilter]
    skycrd_SCX = skycrd_SCX[tfilter]
    skycrd_SCY = skycrd_SCY[tfilter]
    skycrd_SCZ = skycrd_SCZ[tfilter]
    livetime = livetime[tfilter]

    SCX_sep = maptools.Sep(skycrd0_C, skycrd_SCX)
    SCY_sep = maptools.Sep(skycrd0_C, skycrd_SCY)
    SCZ_sep = maptools.Sep(skycrd0_C, skycrd_SCZ)

    Theta = 180 - SCZ_sep
    cosTheta = np.cos(np.deg2rad(Theta))
    phi = np.arccos(np.cos(np.deg2rad(SCX_sep)) / np.sin(np.deg2rad(SCZ_sep)))
    phi[SCY_sep < 90] = np.pi + phi[SCY_sep < 90]
    phi[SCY_sep > 90] = np.pi - phi[SCY_sep > 90]
    Phi = np.rad2deg(phi)

    ltmap = np.zeros((thetabins, phibins))
    for i, (cosThetamin, cosThetamax) in enumerate(zip(cosThetamins, cosThetamaxs)):
        for j, (phimin, phimax) in enumerate(zip(phimins, phimaxs)):
            ltmap[i, j] += np.sum(livetime[(cosTheta >= cosThetamin) * (cosTheta < cosThetamax) * (Phi >= phimin) * (Phi < phimax)])
    if phibins == 1:
        np.savetxt(f'{srcName}/Livetime_cosTheta.dat', ltmap)
        plotLiveTime_cosTheta(srcName)
        return
    else:
        np.savetxt(f'{srcName}/Livetime_cosTheta.dat', ltmap.sum(axis=1))

    if thetabins == 1:
        np.savetxt(f'{srcName}/Livetime_Phi.dat', ltmap)
        plotLiveTime_Phi(srcName)
        return
    else:
        np.savetxt(f'{srcName}/Livetime_Phi.dat', ltmap.sum(axis=0))
    np.savetxt(f'{srcName}/LivetimeMap.dat', ltmap)
    plotLiveTime_cosTheta(srcName)
    plotLiveTime_Phi(srcName)
    plotLiveTimeMap(srcName)


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('scfile', type=str, help='scfile')
    parser.add_argument('srcName', type=str, help='srcName')
    parser.add_argument('GorC', type=str, choices=['G', 'C'], help='coord')
    parser.add_argument('xref', type=float, help='RA or GLON')
    parser.add_argument('yref', type=float, help='DEC or GLAT')
    parser.add_argument('UTCmin', type=str, help='SCmin time in UTC')
    parser.add_argument('UTCmax', type=str, help='SCmax time in UTC')
    parser.add_argument('--thetabins', type=int, default=40, help='bins of theta for livetimMap')
    parser.add_argument('--phibins', type=int, default=72, help='bins of phi for livetimMap')
    args = parser.parse_args()
    FindObservedTime(args.scfile, args.srcName, args.GorC, args.xref, args.yref, args.UTCmin, args.UTCmax)
    #FindLiveTime(args.scfile, args.srcName, args.GorC, args.xref, args.yref, args.UTCmin, args.UTCmax, args.thetabins, args.phibins)