import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from scipy.interpolate import interp1d
from DmpST import DmpTime


def plotObservedAngle(srcName):
    start, stop, Theta, Phi = np.loadtxt(f'{srcName}/ObservedAngle.dat').T

    time = (start + stop) / 2
    plt.figure()
    plt.plot(time, Theta, '.')
    plt.xlabel('MET(s)')
    plt.ylabel('Theta(deg)')
    plt.ylim(0, 180)
    plt.xlim(time[0], time[-1])

    plt.savefig(f'{srcName}/ObservedTheta.png')
    plt.savefig(f'{srcName}/ObservedTheta.pdf')
    
    plt.figure()
    plt.plot(time, Phi, '.')
    plt.xlabel('MET(s)')
    plt.ylabel('Phi(deg)')
    plt.ylim(0, 360)
    plt.xlim(time[0], time[-1])
    plt.savefig(f'{srcName}/ObservedPhi.png')
    plt.savefig(f'{srcName}/ObservedPhi.pdf')

def plotLiveTime_cosTheta(srcName):
    cosThetabounds = np.loadtxt(f'{srcName}/cosThetabounds.dat')
    livetime = np.loadtxt(f'{srcName}/Livetime_cosTheta.dat')

    plt.figure()
    plt.step(cosThetabounds, np.hstack([livetime[0], livetime]))
    plt.xlabel('cosTheta')
    plt.ylabel('Livetime')
    plt.xlim(0.5, 1.0)
    plt.savefig(f'{srcName}/Livetime_cosTheta.png')

def plotLiveTime_Phi(srcName):
    phibounds = np.loadtxt(f'{srcName}/phibounds.dat')
    livetime = np.loadtxt(f'{srcName}/Livetime_Phi.dat')

    plt.figure()
    plt.step(phibounds, np.hstack([livetime[0], livetime]))
    plt.xlabel('Phi')
    plt.ylabel('Livetime')
    plt.xlim(0, 360)
    plt.savefig(f'{srcName}/Livetime_Phi.png')

def plotLiveTimeMap(srcName):
    cosThetabounds = np.loadtxt(f'{srcName}/cosThetabounds.dat')
    phibounds = np.loadtxt(f'{srcName}/phibounds.dat')
    ltmap = np.loadtxt(f'{srcName}/LivetimeMap.dat')
    print(len(cosThetabounds), len(phibounds), ltmap.shape)

    cmap = mpl.cm.gnuplot2
    norm = mpl.colors.PowerNorm(0.5, vmin=0, vmax=1e5)
    fig = plt.figure()
    ax = fig.add_subplot(111)#, projection='polar')
    img = ax.pcolor(ltmap)#, cmap=cmap)#, norm=norm)
    xbins = int((len(cosThetabounds) - 1) / 4)
    ybins = int((len(phibounds) - 1) / 6)
    ax.set_xticks(np.arange(len(phibounds) / ybins) * ybins)
    ax.set_xticklabels(phibounds[::ybins])
    ax.set_yticks(np.arange(len(cosThetabounds-1) / xbins) * xbins)
    ax.set_yticklabels(cosThetabounds[::xbins])
    ax.grid()
    ax.set_xlabel(r'$\phi$')
    ax.set_ylabel(r'cos$\theta$')
    fig.colorbar(img, orientation='horizontal')
    fig.savefig(f'{srcName}/LivetimeMap.png')
    fig.savefig(f'{srcName}/LivetimeMap.pdf')

def plotObservedTime(srcName, ObservedAngle):
    Tmins, Tmaxs, sepMins, sepMaxs = np.loadtxt(f'{srcName}/ObservedTime.dat').T
    if np.amax(sepMins) <= ObservedAngle:
        METbounds_left = Tmins[0]
        METbounds_right = Tmins[-1]
    else:
        METbounds_left = Tmins[:-1][(sepMins[:-1] >= ObservedAngle) * (sepMins[1:] <= ObservedAngle)]
        METbounds_right = Tmins[1:][(sepMins[:-1] <= ObservedAngle) * (sepMins[1:] >= ObservedAngle)]
    
    if sepMins[0] <= ObservedAngle:
        METbounds_left = np.hstack([Tmins[0], METbounds_left])
    if sepMins[-1] <= ObservedAngle:
        METbounds_right = np.hstack([METbounds_right, Tmins[-1]])

    np.savetxt(f'{srcName}/METbounds.dat', np.array([METbounds_left, METbounds_right]).T, header='METbound_left, METbound_right')
    UTCbounds_left = []
    UTCbounds_right = []
    with open(f'{srcName}/UTCbounds.dat', 'w') as f:
        f.write('# UTCbound_left, UTCbound_right\n')
        for METbound_left, METbound_right in zip(METbounds_left, METbounds_right):
            UTCbound_left = DmpTime.MET2UTC(METbound_left)
            UTCbound_right = DmpTime.MET2UTC(METbound_right)
            UTCbounds_left.append(UTCbound_left)
            UTCbounds_right.append(UTCbound_right)
            f.write(UTCbound_left)
            f.write('\t')
            f.write(UTCbound_right)
            f.write('\n')

    plt.figure()
    plt.plot(Tmins, sepMins, 'k')
    plt.plot(Tmaxs, sepMaxs, 'k')
    plt.fill_between(Tmins, sepMins, sepMaxs, where=sepMaxs > sepMins, interpolate=True)
    plt.axhline(ObservedAngle)
    plt.vlines(METbounds_left, 0, ObservedAngle, 'g')
    plt.vlines(METbounds_right, 0, ObservedAngle, 'r')
    
    ax = plt.gca()
    ax.set_position([0.125, 0.25, 0.8, 0.7])
    ax.set_xticks(list(METbounds_left) + list(METbounds_right))
    # ax.set_xticklabels([date for date in UTCbounds_left + UTCbounds_right])
    ax.set_xticklabels([date.split()[0] for date in UTCbounds_left + UTCbounds_right])
    ax.tick_params(axis='x', labelrotation=-90)
    plt.xlim(Tmins[0], Tmaxs[-1])
    plt.ylim(0, 180)
    plt.xlabel('Date')
    plt.ylabel('Observing Angle')
    plt.savefig(f'{srcName}/ObservedTime.png')
    plt.savefig(f'{srcName}/ObservedTime.pdf')
